package jmel.interp;

/**
    A specialized namespace	for Blocks, e.g. the body of a "for" statement.
	The Block acts like a child namespace but only for typed variables 
	declared within it.  Elsewhere variable assignment (including untyped
	variable usage) acts like it is part of the containing block.  
	<p>

	Note: It *must* remain possible for a BlockNameSpace to be a child of
	another BlockNameSpace and have variable propogation pass all the way
	through.  (This happens naturally and simply here). This is used in 
	BSHForStatement (see notes there).
*/
public class BlockNameSpace extends NameSpace 
{

    public BlockNameSpace( NameSpace parent ) 
		throws EvalError
	{
		super( parent, parent.name + "/BlockNameSpace" );
    }

	/**
		Override the standard namespace behavior.
		If the variables exists in our namespace assign it there,
		otherwise in the parent space.
		i.e. only allow typed var declaration to happen in this namespace.
		Typed vars are handled in the ordinary way... local scope.
	*/
    public void	setVariable(String name, Object	o) throws EvalError {
		if ( haveVar( name ) ) 
			super.setVariable( name, o );
		else
			getParent().setVariable( name, o );
    }

	boolean haveVar( String name ) {
		return super.getVariableImpl( name, false ) != null;
	}

}

