package jmel.interp;

import jmel.JMelInterpreter;
import jmel.code.*;
import jmel.script.*;
import jmel.util.*;
import jmel.tree.*;
import jmel.tree.Tree.*;
import java.io.*;
import java.util.Vector;
import java.util.Enumeration;
import java.lang.reflect.Array;

/**
 * Evaluates a Mel script.
 */
public class EvalGlobal extends GlobalVisitor  {

	private CallStack callstack;
	private JMelInterpreter interpreter;

	public EvalGlobal(CallStack callstack, JMelInterpreter interpreter) {
		this.callstack = callstack;
		this.interpreter = interpreter;
	}

	/**
	* Visitor methods
	*/
	public void _case(TopLevel tree) {
		visitStats(tree.defs);
	}
	
	public void _case(MethodDef tree) {
		visitExpr(tree.body);
	}

	public void _case(VarDef tree) {
		if ((tree.flags & Flags.PUBLIC) == 0) return;
		Class type = Reflect.getType(tree.vartype);
		NameSpace namespace = callstack.top();
		Object init = tree.init != null ? (new Eval(callstack, interpreter)).eval(tree.init) : Primitive.ANY;
	
		NameSpace.definedGlobalVariables.put(tree.name.toString(), new NameSpace.TypedVariable(type, init));
	}
	
	public void _case(GlobalRef tree) {
		Class type = Reflect.getType(tree.type);
		NameSpace namespace = callstack.top();
		NameSpace.TypedVariable tv = (NameSpace.TypedVariable)NameSpace.definedGlobalVariables.get(tree.name.toString());
		if (tv == null)
			NameSpace.definedGlobalVariables.put(tree.name.toString(), new NameSpace.TypedVariable(type, Primitive.ANY));
	}

	public void _case(Assign tree) {
		if (tree.lhs instanceof GlobalRef) {
			GlobalRef globalRef = (GlobalRef)tree.lhs;
			Class type = Reflect.getType(globalRef.type);
			NameSpace namespace = callstack.top();
			Object rhs = (new Eval(callstack, interpreter)).eval(tree.rhs);
			NameSpace.definedGlobalVariables.put(globalRef.name.toString(), new NameSpace.TypedVariable(type, rhs));
		}
	}
	
	public void _case(Block tree) {
		visitBlock(tree.stats);
	}
	
	public void _case(DoLoop tree) {
		visitStat(tree.body);
	}

	public void _case(WhileLoop tree) {
		visitStat(tree.body);
	}
	
	public void _case(ForLoop tree) {
		visitStat(tree.body);
	}
	
	public void _case(ForIn tree) {
		visitStat(tree.body);
	}
		
	public void _case(Labelled tree) {
		visitStat(tree.body);
	}
	
	public void _case(Switch tree) {
		visitStats(tree.cases);
	}

	public void _case(Case tree) {
		visitStats(tree.stats);
	}
	
	public void _case(Catch tree) {
		visitStat(tree.body);
	}

	public void _case(Conditional tree) {
		visitStat(tree.thenpart);
		visitStat(tree.elsepart);
	}

	public void _case(Exec tree) {
	}
	
	public void _case(Tree.Command tree) {
		visitExpr(tree.cmd);
	}
	
	public void _case(Break tree) {
	}
	
	public void _case(Continue tree) {
	}
	
	public void _case(Return tree) {
	}
	
	public void _case(Apply tree) {
	}

	public void _case(NewArray tree) {
	}
	
	public void _case(Assignop tree) {
	}

	public void _case(Operation tree) {
	}

	public void _case(TypeCast tree) {
	}
	
	public void _case(TypeTest tree) {
	}
	
	public void _case(Indexed tree) {
	}
	
	public void _case(Select tree) {
	}
	
	public void _case(Ident tree) {
	}
	
	public void _case(Literal tree) {
	}

	public void _case(TypeIdent tree) {
	}
	
	public void _case(TypeArray tree) {
	}
	
	public void _case(TypeMatrix tree) {
	}
	
	public void _case(NewVector tree) {
	}
	
	public void _case(NewMatrix tree) {
	}
	
	public void _case(Erroneous tree) {
	}
}
