package jmel.interp;

import jmel.script.*;
import java.lang.reflect.Field;
import java.util.Hashtable;

/**
	The left hand side in an assignment

	This is probably the most debatable design issue in bsh...
	Because of the way things started, the grammar splits on whether
	something is an LHS or not...  The alternative would be to wrap all
	objects in bsh, rather than just carrying types that might be used in
	an assignment.  Wrapping all objects, say, in a generalized BshObject 
	type, would also provide a nice place to put all the reflection stuff, 
	which is now static in bsh.Reflect

	Note: moving some stuff from Reflect to a BshObject, but not going
	as far as the above yet...
*/
public class LHS implements java.io.Serializable
{
	NameSpace nameSpace;

	static final int
		VARIABLE = 0,
		FIELD = 1,
		INDEX = 2;

	int type;

	String varName;

	Object object;
	Field field;
	boolean isMatrix = false;
	int index1, index2 = -1;

	LHS(NameSpace nameSpace, String varName)
	{
		type = VARIABLE;
		this.varName = varName;
		this.nameSpace = nameSpace;
	}

	// Object field
	LHS(Object object, Field field)
	{
		if(object == null)
			throw new NullPointerException("constructed empty LHS");

		type = FIELD;
		this.object = object;
		this.field = field;
	}

	// Array index
	LHS(Object array, int index)
	{
		if(array == null)
			throw new NullPointerException("constructed empty LHS");

		type = INDEX;
		this.object = array;
		this.index1 = index;
	}
	LHS(Object array, int index1, int index2)
	{
		if(array == null)
			throw new NullPointerException("constructed empty LHS");

		type = INDEX;
		this.object = array;
		this.index1 = index1;
		this.index2 = index2;
		isMatrix = true;
	}

	public Object getValue() throws EvalError
	{
		if(type == VARIABLE)
			return nameSpace.getVariable(varName);
		else if (type == FIELD)
			try {
				return field.get(object);
			}
			catch(IllegalAccessException e2) {
				throw new EvalError("Can't read field: " + field);
			}
		else if(type == INDEX)
			try
			{
				if (isMatrix)
					return Reflect.getIndex(object, index1, index2);
				return Reflect.getIndex(object, index1);
			}
			catch(Exception e)
			{
				throw new EvalError("Array access: " + e);
			}
			
		throw new InterpreterError("LHS type");
	}

	public Object assign( Object val ) throws EvalError
	{
		if ( type == VARIABLE )
			nameSpace.setVariable(varName, val);
			
		else if ( type == FIELD )
			try {
				if(val instanceof Primitive)
					val = ((Primitive)val).getValue();

				field.set(object, val);
				return val;
			}
			catch( NullPointerException e) {   
    			throw new EvalError(
					"LHS ("+field.getName()+") not a static field.");
			}     
   			catch( IllegalAccessException e2) {   
				throw new EvalError(
					"LHS ("+field.getName()+") can't access field.");
			}     
			catch( IllegalArgumentException e3) {
				throw new EvalError(
					"Argument type mismatch. "
					+ (val == null ? "null" : val.getClass().getName())
					+ " not assignable to field "+field.getName());
			}
			
		else if(type == INDEX)
			try {
				if (isMatrix)
					Reflect.setIndex(object, index1, index2, val);
				else
					Reflect.setIndex(object, index1, val);
			} catch(TargetError e1) { // pass along target error
				throw e1;
			} catch(Exception e) {
				throw new EvalError("Assignment: " + e.getMessage());
			}
			
		return val;
	}

	public String toString() { return "LHS"+type; }
}

