package jmel.interp;

import jmel.JMelInterpreter;
import jmel.script.*;
import jmel.tree.Tree;
import jmel.code.Type;
import jmel.tree.Tree.*;
import java.lang.reflect.*;
import java.io.*;
import java.util.Vector;

/**
    All of the reflection API code lies here.  It is in the form
	of static utilities.  See the design note about object wrappers 
	in LHS.java for lamentations regarding this.

	Note: More work to do in here to fix up the extended signature matching.
	need to work in a search along with findMostSpecificSignature...
*/
public class Reflect {

    public static Object getIndex(Object array, int index)
        throws ReflectError, TargetError
    {
        try {
        	Object val;
   	    	if (array instanceof Primitive && (((Primitive)array).getValue()) instanceof MelArray)
        		val = ((MelArray)((Primitive)array).getValue()).getValue(index);
           	else
           		val = Array.get(array, index);
            return new Primitive(val);
        }
        catch( ArrayIndexOutOfBoundsException  e1 ) {
			throw new TargetError( "Array Index", e1 );
        } catch(Exception e) {
            throw new ReflectError("Array access:" + e);
        }
    }

    public static Object getIndex(Object array, int index1, int index2)
        throws ReflectError, TargetError
    {
        try {
        	Object val;
   	    	if (array instanceof Primitive && (((Primitive)array).getValue()) instanceof MelArray)
        		val = ((MelArray)((Primitive)array).getValue()).getValue(index1, index2);
           	else
           		val = Array.get(Array.get(array, index1), index2);
            return new Primitive(val);
        }
        catch( ArrayIndexOutOfBoundsException  e1 ) {
			throw new TargetError( "Array Index", e1 );
        } catch(Exception e) {
            throw new ReflectError("Array access:" + e);
        }
    }

    public static LHS getLHSIndex(Object array, int index)
        throws ReflectError, TargetError
    {
        return new LHS(array, index);
   }

    public static LHS getLHSIndex(Object array, int index1, int index2)
        throws ReflectError, TargetError
    {
        return new LHS(array, index1, index2);
   }


    public static void setIndex(Object array, int index, Object val)
        throws ReflectError, TargetError
    {
        try {
            val = unwrapPrimitive(val);    
   	    	if (array instanceof Primitive && (((Primitive)array).getValue()) instanceof MelArray)
        		((MelArray)((Primitive)array).getValue()).setValue(index, val);
            else
            	Array.set(array, index, val);
        }
        catch( ArrayStoreException e2 ) {
			throw new TargetError( "Array store exception", e2 );
        } catch( IllegalArgumentException e1 ) {
			throw new TargetError( "Illegal Argument", 
				new ArrayStoreException( e1.toString() ) );
        } catch(Exception e) {
            throw new ReflectError("Array access:" + e);
        }
    }

    public static void setIndex(Object array, int index1, int index2, Object val)
        throws ReflectError, TargetError
    {
        try {
            val = unwrapPrimitive(val);
   	    	if (array instanceof Primitive && (((Primitive)array).getValue()) instanceof MelArray)
        		((MelArray)((Primitive)array).getValue()).setValue(index1, index2, val);
            else
            	Array.set(Array.get(array, index1), index2, val);
        }
        catch( ArrayStoreException e2 ) {
			throw new TargetError( "Array store exception", e2 );
        } catch( IllegalArgumentException e1 ) {
			throw new TargetError( "Illegal Argument", 
				new ArrayStoreException( e1.toString() ) );
        } catch(Exception e) {
            throw new ReflectError("Array access:" + e);
        }
    }

     public static Object getObjectField(Object object, String fieldName)
        throws ReflectError
    {
		if ( object instanceof This )
			return ((This)object).namespace.getVariable( fieldName );
		else {
			try {
				object = unwrapPrimitive(object);
				return getFieldValue(object.getClass(), object, fieldName);
			} catch ( ReflectError e ) {
				// no field, try property acces

				throw e;
			}
		}
    }

	/**
		Get an LHS reference to an object field.

		This method also deals with the field style property access.
		In the field does not exist we check for a property setter.
	*/
    public static LHS getLHSObjectField(Object object, String fieldName)
        throws ReflectError
    {
		if ( object instanceof This )
			return new LHS(((This)object).namespace, fieldName );

		try {
			Field f = getField(object.getClass(), fieldName);
			return new LHS(object, f);
		} catch ( ReflectError e ) {
			// not a field, try property access

			throw e;
		}
    }

    private static Object getFieldValue(
		Class clas, Object object, String fieldName) throws ReflectError
    {
        try {
            Field f = getField(clas, fieldName);

           if(f == null)
                throw new ReflectError("internal: field not found:"+fieldName);

            Object value = f.get(object);
            Class returnType = f.getType();
            return wrapPrimitive(value, returnType);

        }
        catch(NullPointerException e) {
            throw new ReflectError(
				"???" + fieldName + " is not a static field.");
        }
        catch(IllegalAccessException e) {
            throw new ReflectError("Can't access field: " + fieldName);
        }
    }

	/**
		All field lookup should come through here.
		i.e. this method owns Class getField();
	*/
    private static Field getField(Class clas, String fieldName)
        throws ReflectError
    {
       try
        {
			// this one only finds public 
			return clas.getField(fieldName);
        }
        catch(NoSuchFieldException e)
        {
			// try declaredField
            throw new ReflectError("No such field: " + fieldName );
        }
    }

	public static Class getType( Tree vartype ) {
		if (vartype instanceof TypeIdent) {
			switch (((TypeIdent)vartype).typetag) {
				case Type.INT: return int.class;
				case Type.FLOAT: return float.class;
				case Type.DOUBLE: return double.class;
				case Type.STRING: return String.class;
				case Type.VECTOR: return MelVector.class;
				case Type.VOID: return void.class;
			}
		} else if (vartype instanceof TypeArray) {
			TypeArray ta = (TypeArray)vartype;
			switch (((TypeIdent)((TypeArray)vartype).elemtype).typetag) {
				case Type.INT: return MelArrayInt.class;
				case Type.FLOAT: return MelArrayFloat.class;
				case Type.DOUBLE: return MelArrayDouble.class;
				case Type.STRING: return MelArrayString.class;
				case Type.VECTOR: return MelArrayVector.class;
			}
		} else if (vartype instanceof TypeMatrix) {
			return MelMatrix.class;
		}
		throw new EvalError("Cannot extract type from " + vartype);
	}


	/**
		Used when accessibility capability is available to locate an occurrance
		of the field in the most derived class or superclass and set its 
		accessibility flag.
		Note that this method is not needed in the simple non accessible
		case because we don't have to hunt for fields.
		Note that classes may declare overlapping private fields, so the 
		distinction about the most derived is important.  Java doesn't normally
		allow this kind of access (super won't show private variables) so 
		there is no real syntax for specifying which class scope to use...

		Need to improve this to handle interfaces.
	*/
	private static Field findAccessibleField( Class clas, String fieldName ) 
		throws NoSuchFieldException
	{
		while ( clas != null )
		{
			try {
				Field field = clas.getDeclaredField(fieldName);
				return field;
				// else fall through
			}
			catch(NoSuchFieldException e) { }

			clas = clas.getSuperclass();
		}
		throw new NoSuchFieldException( fieldName );
	}

   private static Object wrapPrimitive(Object value, Class returnType) throws ReflectError
    {
        if(value == null)
            return Primitive.NULL;

        if(returnType == Void.TYPE)
            return Primitive.VOID;

        else
            if(returnType.isPrimitive())
            {
                if(value instanceof Number)
                    return new Primitive((Number)value);

                throw new ReflectError("Something bad happened");
            }
            else
                return value;
    }

    public static Class[] getTypes( Object[] args)
    {
        if(args == null)
            return new Class[0];

        Class[] types = new Class[args.length];

        for(int i=0; i<args.length; i++)
        {
			if ( args[i] == null )
                types[i] = null;
            else
				//throw new InterpreterError("Null arg in getTypes()");

            if(args[i] instanceof Primitive)
                types[i] = ((Primitive)args[i]).getType();
            else
                types[i] = args[i].getClass();
        }

        return types;
    }

    /*
        Replace Primitive wrappers with their java.lang wrapper values

        These barf if one of the args is void...  maybe these should throw
        an exception on void arg to force the rest of the code to clean up.
        There are places where we don't check right now... (constructors, index)
    */
    public static void unwrapPrimitives(Object[] args)
    {
        for(int i=0; i<args.length; i++)
            args[i] = unwrapPrimitive(args[i]);
    }

    public static Object unwrapPrimitive(Object arg)
    {
        if(arg instanceof Primitive)
            return((Primitive)arg).getValue();
        else
            return arg;
    }

 
	/**
        Implement JLS 15.11.2
		Return the index of the most specific arguments match or -1 if no	
		match is found.
	*/
	static int findMostSpecificSignature(
		Class [] idealMatch, Class [][] candidates )
	{
		Class [] bestMatch = null;
		int bestMatchIndex = -1;

		for (int i=0; i < candidates.length; i++) {
			Class[] targetMatch = candidates[i];

            /*
                If idealMatch fits targetMatch and this is the first match 
				or targetMatch is more specific than the best match, make it 
				the new best match.
            */
			if ( isAssignable(idealMatch, targetMatch ) )
			{
				bestMatch = targetMatch;
				bestMatchIndex = i;
			}
		}

		if ( bestMatch != null ) {
			
			JMelInterpreter.debug("best match: " 
				+ StringUtil.methodString("args",bestMatch));
			
				
			return bestMatchIndex;
		}
		else {
			JMelInterpreter.debug("no match found");
			return -1;
		}
	}

	/**
		Determine if the 'from' signature is assignable to the 'to' signature
		'from' arg types, 'to' candidate types
		null value in 'to' type parameter indicates loose type.

		null value in either arg is considered empty array
	*/
    static boolean isAssignable(Class[] from, Class[] to)
    {
		if ( from == null )
			from = new Class[0];
		if ( to == null )
			to = new Class[0];

        if (from.length != to.length)
            return false;

        for(int i=0; i<from.length; i++)
        {
			// Null type indicates loose type.  Match anything.
			if ( to[i] == null )
				continue;

            // Let null arg type match any reference type
            if (from[i] == null) {

                if (!(to[i].isPrimitive()))
                    continue;
                else
                    return false;
            }

            if(!isAssignableFrom(to[i], from[i]))
                return false;
        }

        return true;
    }

    /**
		This base method is meant to address a deficiency of 
		Class.isAssignableFrom() which does not take primitive widening 
		conversions into account.

		Note that the getAssigbableForm() method in NameSpace is the primary
		bsh method for checking assignability.  It adds extended bsh
		conversions, etc.

		@param lhs assigning from rhs to lhs
		@param rhs assigning from rhs to lsh
	*/
    public static boolean isAssignableFrom(Class lhs, Class rhs)
    {
		if(lhs.isPrimitive() && rhs.isPrimitive())
        {
            if(lhs == rhs)
                return true;

			if (lhs == Integer.TYPE) {
				if (rhs == Float.TYPE) return true;
				if (rhs == Double.TYPE) return true;
			}

			if (lhs == Float.TYPE) {
				if (rhs == Integer.TYPE) return true;
				if (rhs == Double.TYPE) return true;
			}

			if (lhs == Double.TYPE) {
				if (rhs == Integer.TYPE) return true;
				if (rhs == Float.TYPE) return true;
			}
        }
        else
            if(lhs.isAssignableFrom(rhs))
                return true;

        return false;
     }

   /** 
		This method is meant to convert a JVM-array class name to the correct
    	'fully-qualified name' for the array class - JLS 6.7
	*/
    public static String normalizeClassName(Class type)
    {
        if(!type.isArray())
            return type.getName();

        StringBuffer className = new StringBuffer();
        try
        {
            className.append(getArrayBaseType(type).getName());
            for(int i = 0; i < getArrayDimensions(type); i++)
                className.append("[]");
        }
        catch(Exception e) { }

        return className.toString();
    }

	/**
		returns the dimensionality of the Class
		returns 0 if the Class is not an array class
	*/
    public static int getArrayDimensions(Class arrayClass)
    {
        if(!arrayClass.isArray())
            return 0;

        return arrayClass.getName().lastIndexOf('[') + 1;
    }

    /**

		Returns the base type of an array Class.
    	throws ReflectError if the Class is not an array class.
	*/
    public static Class getArrayBaseType(Class arrayClass) throws ReflectError
    {
        if(!arrayClass.isArray())
            throw new ReflectError("The class is not an array.");

		return arrayClass.getComponentType();

    }

	static Method findAccessibleMethod( 
		Class clas, String name, Class [] types, boolean onlyStatic ) 
	{
		Method meth = null;
		Vector classQ = new Vector();

		classQ.addElement( clas );
		Method found = null;
		while ( classQ.size() > 0 ) 
		{
			Class c = (Class)classQ.firstElement();
			classQ.removeElementAt(0);

			// Is this it?
			// Is the class public or can we use accessibility?
			if ( Modifier.isPublic( c.getModifiers() ) )
			{
				try {
					meth = c.getDeclaredMethod( name, types );

					// Is the method public or are we in accessibility mode?
					if ( Modifier.isPublic( meth.getModifiers() ) )
					{
						found = meth; // Yes, it is.
						break;
					}
				} catch ( NoSuchMethodException e ) { 
					// ignore and move on
				}
			}
			// No, it is not.
			
			// Is this a class?
			if ( !c.isInterface() ) {
				Class superclass = c.getSuperclass();
				if ( superclass != null )
					classQ.addElement((Object)superclass);
			}

			// search all of its interfaces breadth first
			Class [] intfs = c.getInterfaces();
			for( int i=0; i< intfs.length; i++ )
				classQ.addElement((Object)intfs[i]);
		}

		/* 
			If we found one and it satisfies onlyStatic return it
			
			Note: I don't believe it is necessary to check for the static
			condition in the above search because the Java compiler will not
			let dynamic and static methods hide/override one another.  So
			we simply check what is found, if any, at the end.
		*/
		if ( found != null &&
			( !onlyStatic || Modifier.isStatic( found.getModifiers() ) ) )
			return found;
		
		// Didn't find one
		/*
		Interpreter.debug(
			"Can't find publically accessible "+
			( onlyStatic ? " static " : "" )
			+" version of method: "+
			StringUtil.methodString(name, types) +
			" in interfaces or class hierarchy of class "+clas.getName() );
		*/

		return null;
	}


   public static Object invokeStaticMethod(	Class clas, String methodName, Object [] args)
	throws ReflectError, EvalError
	{
		JMelInterpreter.debug("invoke static Method");
		return invokeMethod( clas, null, methodName, args, true );
	}

	private static Method [] retainStaticMethods( Method [] methods ) {
		Vector v = new Vector();
		for(int i=0; i<methods.length; i++)
			if ( Modifier.isStatic( methods[i].getModifiers() ) )
				v.addElement( methods[i] );

		Method [] ma = new Method [ v.size() ];
		v.copyInto( ma );
		return ma;
	}

    private static Object invokeMethod(	Class clas, Object object, String name, Object[] args,	boolean onlyStatic)
        throws ReflectError, EvalError
    {
		if ( object == Primitive.NULL )
			throw new TargetError("Attempt to invoke method "
				+name+" on null value", new NullPointerException() );
		if ( object == Primitive.VOID )
			throw new EvalError("Attempt to invoke method "
				+name+" on undefined variable or class name" );

        if (args == null)
            args = new Object[] { };

        // Simple sanity check for voids
        // (maybe this should have been caught further up?)
        for(int i=0; i<args.length; i++)
            if(args[i] == Primitive.VOID)
                throw new ReflectError("Attempt to pass void argument " +
                    "(position " + i + ") to method: " + name);

        Class returnType = null;
        Object returnValue = null;

        Class[] types = getTypes(args);
        unwrapPrimitives(args);

        try
        {
			// Try the easy case: Look for an accessible version of the 
			// direct match.

			Method m = null;
			try {
				m  = findAccessibleMethod(clas, name, types, onlyStatic);
			} catch ( SecurityException e ) { }

			if ( m == null )
				JMelInterpreter.debug("Exact method " + 
					StringUtil.methodString(name, types) +
					" not found in '" + clas.getName() + "'" );

			// Next look for an assignable match
            if ( m == null ) {

				// If no args stop here
				if ( types.length == 0 )
					throw new ReflectError(
						"No args "+ ( onlyStatic ? "static " : "" )
						+"method " + StringUtil.methodString(name, types) + 
						" not found in class'" + clas.getName() + "'");

				// try to find an assignable method
				Method[] methods = clas.getMethods();
				if ( onlyStatic )
					// only try the static methods
					methods = retainStaticMethods( methods );

				m = findMostSpecificMethod(name, types, methods);

				// try to find an extended method
				methods = clas.getMethods();
				if ( m == null )
					m = findExtendedMethod(name, args, methods);

				// If we found an assignable method, make sure it's accessible
				if ( m != null ) {
					try {
						m = findAccessibleMethod( clas, m.getName(), 
							m.getParameterTypes(), onlyStatic);
					} catch ( SecurityException e ) { }
				}
            }

			// Found something?
			if (m == null )
				throw new ReflectError(
					( onlyStatic ? "Static method " : "Method " )
					+ StringUtil.methodString(name, types) + 
					" not found in class'" + clas.getName() + "'");

			// Invoke it
            returnValue =  m.invoke(object, args);
            if(returnValue == null)
                returnValue = Primitive.NULL;
            returnType = m.getReturnType();

		} catch(IllegalAccessException e) {
			throw new ReflectError( 
				"Cannot access method " + StringUtil.methodString(name, types) +
				" in '" + clas.getName() + "' :" + e);
		} catch(InvocationTargetException e) {
			throw new ReflectError( 
				"Cannot access method " + StringUtil.methodString(name, types) +
				" in '" + clas.getName() + "' :" + e.getTargetException());
		}

        return wrapPrimitive(returnValue, returnType);
    }
    static Method findMostSpecificMethod(
		String name, Class[] idealMatch, Method[] methods )
    {
		// Pull out the method signatures whos name matches
		Vector sigs = new Vector();
		Vector meths = new Vector();
		for(int i=0; i<methods.length; i++)
			// method matches name 
			if ( methods[i].getName().equals( name )  ) 
			{
				meths.addElement( methods[i] );
				sigs.addElement( methods[i].getParameterTypes() );
			}

		Class [][] candidates = new Class [ sigs.size() ][];
		sigs.copyInto( candidates );

		JMelInterpreter.debug("Looking for most specific method: "+name);
		int match = findMostSpecificSignature( idealMatch, candidates );
		if ( match == -1 )
			return null;
		else
			return (Method)meths.elementAt( match );
    }

	/**
		This uses the NameSpace.getAssignableForm() method to determine
		compatability of args.  This allows special (non standard Java) bsh 
		widening operations...

		@returns null on not found
	*/
    static Method findExtendedMethod(
		String name, Object[] args, Method[] methods)
    {
        Method bestMatch = null;
        Object[] tempArgs = new Object[args.length];

        for(int i = 0; i < methods.length; i++) {
            Method currentMethod = methods[i];
            if ( name.equals( currentMethod.getName() )) {
                Class[] parameters = currentMethod.getParameterTypes();
		
				if ( parameters.length != args.length )
					continue;
                try {
                    for(int j = 0; j < parameters.length; j++)
                        tempArgs[j] = NameSpace.getAssignableForm( 
							args[j], parameters[j]);

                    // if you get here, all the arguments were assignable
                    System.arraycopy(tempArgs, 0, args, 0, args.length);
                    return currentMethod;
                } catch(EvalError e) {
                    // do nothing (exception breaks you out of the for loop).
                }
            }
        }

        return null;
    }
}

