package jmel.script;

import jmel.interp.*;
import java.lang.reflect.Method;

public abstract class Callback extends ScriptCaller
{
	private String callbackName;
	
	public Callback(String callbackName, Class[] callbackArgs, Class callbackReturnType) {
		if (!insideScript)
			try {
				Class cl = Commands.getClass(callbackName);
				Method m = cl.getMethod("exec", callbackArgs);
				if ( !callbackReturnType.isAssignableFrom(m.getReturnType()) )
					throw new InternalError("wrong return type for callback '" + callbackName + "'.");
				setMethod(m);
			} catch (NoSuchMethodException e) {
				throw new InternalError("wrong argument types for callback '" + callbackName + "'.");
			} catch (Throwable e) {
				throw new InternalError("Unknown callback '" + callbackName + "'.");
			}
		else {
	        JMelScriptMethod meth = namespace.getMethod( callbackName, callbackArgs );
	        if ( meth == null || !meth.isPublic())
				throw new InternalError("Unknown callback or wrong argument types for '" + callbackName + "'.");
			if ( !callbackReturnType.isAssignableFrom(meth.getReturnType()) )
				throw new InternalError("wrong return type for callback '" + callbackName + "'.");
			setMethod(meth);
		}
		this.callbackName = callbackName;
	}
	
	public String toString() {
		return callbackName;
	}
}
