package jmel.script;

import jmel.tree.*;
import jmel.tree.Tree.*;
import jmel.code.Type;
import jmel.util.*;
import java.lang.reflect.*;
import java.util.Iterator;
import java.util.Vector;
import java.util.TreeMap;
import java.util.Enumeration;

class Delta {
	int delta = 0;
}
	
abstract class CommandElements 
{
	protected static TreeMaker make = new TreeMaker();
	public static final Name RETURNVALUE = Name.fromString("returnValue");
	
	abstract public Tree toTree(int pos);
	
	public Tree optimize(Tree tree) { return null; }
	public Tree optimize(Tree tree1, Tree tree2) { return null; }
	
	public Class getReturnType() { return null; }
}

class CommandProperty extends CommandElements
{
	protected Method method;
	protected int minParamCount = 0;
	private Object[] args;
	
	public CommandProperty(Method method) { 
		this.method = method; 
		Class[] args = method.getParameterTypes();
		for (int i = 0; i < args.length; i++) {
			if (!args[i].isArray()) minParamCount++;
		}
	}
	public CommandProperty(CommandProperty cp, Object[] args) { 
		this(cp.method); 
		this.args = args; 
	}

	public Object[] parseArgs(CommandInvoker me, String[] args, Delta d) {
		d.delta = args != null ? args.length : 0;
		if (minParamCount > 0 && (args == null || args.length == 0)) return null;
		return me.parseArgs(method.getParameterTypes(), args, d);
	}

	public Object[] parseArgs(CommandObjectInvoker me, Object[] args, Delta d) {
		d.delta = args != null ? args.length : 0;
		if (minParamCount > 0 && (args == null || args.length == 0)) return null;
		return me.parseArgs(method.getParameterTypes(), args, d);
	}
	
	public Class getReturnType() { return method.getReturnType(); }
	
	public Object invoke(Object obj) throws Throwable { 
		return method.invoke(obj, args);
	}
	
	public Tree toTree(int pos) { return null; }
	
	public String toString() {
		String tmp = method.getName() + "(";
		Class[] args = method.getParameterTypes();
		for (int i = 0; i < args.length; i++) {
			tmp += args[i].getName();
			if (i < args.length-1) tmp += ", ";
		}
		tmp += ") " + this.args + " "+(this.args!=null?this.args.length:0);
		return tmp;
	}
}		

class CommandPropertyGet extends CommandProperty
{
	protected String commandName = null;
	protected List argTrees;

	public CommandPropertyGet(Method method) { super(method); }
	public CommandPropertyGet(CommandPropertyGet cp, List argTrees, String commandName) { 
		super(cp.method); 
		if (argTrees == null) argTrees = new List();
		this.argTrees = argTrees; 
		this.commandName = commandName;
	}

	public List parseArgs(CommandTreeInvoker me, List args, Delta d) {
		d.delta = args != null ? args.length() : 0;
		if (minParamCount > 0 && (args == null || args.length() == 0)) return null;
		return me.parseArgs(method.getParameterTypes(), args, d);
	}
	
	public Tree toTree(int pos) { 
		// returnValue = cmdObj.get<opt>(<arg>);
		Select select = make.at(pos).Select(make.at(pos).Ident(Name.fromString(commandName != null ? commandName : "cmdObj")), Name.fromString(method.getName()));
		Apply apply = make.at(pos).Apply(select, argTrees);
		Assign assign = make.at(pos).Assign(make.at(pos).Ident(RETURNVALUE), apply);
		Exec exec = make.at(pos).Exec(assign);
		return exec;
	}
	
	public Tree optimize(Tree tree) { 
		return ((Assign)((Exec)tree).expr).rhs;
	}
	
	public Tree optimize(Tree tree1, Tree tree2) { 
		// tree1 : VarDef (CommandArgument/CommandInstance)
		// tree2 : Exec (CommandPropertyGet)
		//		<cmd> <obj> = new <cmd>(arg);
		//		returnValue = <obj>.get<opt>();
		((Select)((Apply)((Assign)((Exec)tree2).expr).rhs).meth).selected = ((VarDef)tree1).init;
		return (Apply)((Assign)((Exec)tree2).expr).rhs;
	}
}

class CommandPropertySet extends CommandProperty
{
	protected String commandName = null;
	protected List argTrees;

	public CommandPropertySet(Method method) { super(method); }
	public CommandPropertySet(CommandPropertySet cp, List argTrees, String commandName) { 
		super(cp.method); 
		if (argTrees == null) argTrees = new List();
		this.argTrees = argTrees; 
		this.commandName = commandName;
	}

	public List parseArgs(CommandTreeInvoker me, List args, Delta d) {
		d.delta = args != null ? args.length() : 0;
		if (minParamCount > 0 && (args == null || args.length() == 0)) return null;
		return me.parseArgs(method.getParameterTypes(), args, d);
	}
	
	public Tree toTree(int pos) { 
		// cmdObj.set<opt>(<arg>);
		Select select = make.at(pos).Select(make.at(pos).Ident(Name.fromString(commandName != null ? commandName : "cmdObj")), Name.fromString(method.getName()));
		Apply apply = make.at(pos).Apply(select, argTrees);
		Exec exec = make.at(pos).Exec(apply);
		return exec;
	}
	
	public Tree optimize(Tree tree) { 
		return ((Exec)tree).expr;
	}
	
	public Tree optimize(Tree tree1, Tree tree2) { 
		// tree1 : VarDef (CommandArgument/CommandInstance)
		// tree2 : Exec (CommandPropertySet)
		//		<cmd> <obj> = new <cmd>(arg);
		//		<obj>.set<opt>(arg);
		((Select)((Apply)((Exec)tree2).expr).meth).selected = ((VarDef)tree1).init;
		return ((Exec)tree2).expr;
	}
}
	
class CommandExecAsExpr extends CommandPropertyGet
{
	public CommandExecAsExpr(Method method) { super(method); }
	public CommandExecAsExpr(CommandExecAsExpr cp) { 
		super(cp.method); 
	}

	public List parseArgs(CommandTreeInvoker me, List args, Delta d) {
		d.delta = args != null ? args.length() : 0;
		if (minParamCount > 0 && (args == null || args.length() == 0)) return null;
		return me.parseArgs(method.getParameterTypes(), args, d);
	}
	
	public Tree toTree(int pos) { 
		// returnValue = Commands.execAsExpr(cmdObj);
		Select select = make.at(pos).Select(make.at(pos).Ident(Name.fromString("Commands")), Name.fromString("execAsExpr"));
		Apply apply = make.at(pos).Apply(select, Tree.emptyList.append(make.at(pos).Ident(Name.fromString("cmdObj"))) );
		Assign assign = make.at(pos).Assign(make.at(pos).Ident(RETURNVALUE), apply);
		Exec exec = make.at(pos).Exec(assign);
		return exec;
	}
	
	public Tree optimize(Tree tree) { 
		return ((Assign)((Exec)tree).expr).rhs;
	}
	
	public Tree optimize(Tree tree1, Tree tree2) { 
		// tree1 : VarDef (CommandArgument/CommandInstance)
		// tree2 : Exec (CommandExecAsExpr)
		//		<cmd> <obj> = new <cmd>(arg);
		//		returnValue = Commands.execAsExpr(<obj>);
		((Apply)((Assign)((Exec)tree2).expr).rhs).args.head = ((VarDef)tree1).init;
		return (Apply)((Assign)((Exec)tree2).expr).rhs;
	}
}
	
class CommandExec extends CommandPropertySet
{
	public CommandExec(Method method) { super(method); }
	public CommandExec(CommandExec cp) { 
		super(cp.method); 
	}

	public List parseArgs(CommandTreeInvoker me, List args, Delta d) {
		d.delta = args != null ? args.length() : 0;
		if (minParamCount > 0 && (args == null || args.length() == 0)) return null;
		return me.parseArgs(method.getParameterTypes(), args, d);
	}
	
	public Tree toTree(int pos) { 
		// Commands.exec(cmdObj);
		Select select = make.at(pos).Select(make.at(pos).Ident(Name.fromString("Commands")), Name.fromString("exec"));
		Apply apply = make.at(pos).Apply(select, Tree.emptyList.append(make.at(pos).Ident(Name.fromString("cmdObj"))) );
		Exec exec = make.at(pos).Exec(apply);
		return exec;
	}
	
	public Tree optimize(Tree tree) { 
		return ((Exec)tree).expr;
	}
	
	public Tree optimize(Tree tree1, Tree tree2) { 
		// tree1 : VarDef (CommandArgument/CommandInstance)
		// tree2 : Exec (CommandExec)
		//		<cmd> <obj> = new <cmd>(arg);
		//		Commands.exec(<obj>);
		((Apply)((Exec)tree2).expr).args.head = ((VarDef)tree1).init;
		return ((Exec)tree2).expr;
	}
}
		
class CommandStaticPropertyGet extends CommandPropertyGet
{
	public CommandStaticPropertyGet(Method method) { super(method); }
	public CommandStaticPropertyGet(CommandStaticPropertyGet cp, List argTrees, String commandName) { 
		super(cp, argTrees, commandName); 
	}
	
	public Tree toTree(int pos) { 
		// returnValue = <cmd>.get<opt>(<arg>);
		Select select = make.at(pos).Select(make.at(pos).Ident(Name.fromString(commandName)), Name.fromString(method.getName()));
		Apply apply = make.at(pos).Apply(select, argTrees);
		Assign assign = make.at(pos).Assign(make.at(pos).Ident(RETURNVALUE), apply);
		Exec exec = make.at(pos).Exec(assign);
		return exec;
	}
	
	public Tree optimize(Tree tree1, Tree tree2) { 
		// tree1 : VarDef (CommandArgument/CommandInstance)
		// tree2 : Exec (CommandStaticPropertyGet)
		//		<cmd> <obj> = new <cmd>(arg);
		//		returnValue = <cmd>.get<opt>();
		return (Apply)((Assign)((Exec)tree2).expr).rhs;
	}
}
	
class CommandStaticPropertySet extends CommandPropertySet
{
	public CommandStaticPropertySet(Method method) { super(method); }
	public CommandStaticPropertySet(CommandStaticPropertySet cp, List argTrees, String commandName) { 
		super(cp, argTrees, commandName); 
	}
	
	public Tree toTree(int pos) { 
		// <cmd>.set<opt>(<arg>);
		Select select = make.at(pos).Select(make.at(pos).Ident(Name.fromString(commandName)), Name.fromString(method.getName()));
		Apply apply = make.at(pos).Apply(select, argTrees);
		Exec exec = make.at(pos).Exec(apply);
		return exec;
	}
	
	public Tree optimize(Tree tree1, Tree tree2) { 
		// tree1 : VarDef (CommandArgument/CommandInstance)
		// tree2 : Exec (CommandPropertySet)
		//		<cmd> <obj> = new <cmd>(arg);
		//		<cmd>.set<opt>(arg);
		return ((Exec)tree2).expr;
	}
}	

class CommandStaticAsExpr extends CommandPropertyGet
{
	public CommandStaticAsExpr(Method method) { super(method); }
	public CommandStaticAsExpr(CommandStaticAsExpr cp, List argTrees, String commandName) { 
		super(cp, argTrees, commandName); 
	}
}
	
class CommandStatic extends CommandPropertySet
{
	public CommandStatic(Method method) { super(method); }
	public CommandStatic(CommandStatic cp, List argTrees, String commandName) { 
		super(cp, argTrees, commandName); 
	}
}	
	
class CommandArgument extends CommandElements
{
	private Constructor constructor;
	protected int minParamCount = 0;
	private Object[] args;
	private List argTrees;
	
	public CommandArgument(Constructor constructor) { 
		this.constructor = constructor; 
		Class[] args = constructor.getParameterTypes();
		for (int i = 0; i < args.length; i++) {
			if (!args[i].isArray()) minParamCount++;
		}
	}
	public CommandArgument(CommandArgument ca, Object[] args) { 
		this(ca.constructor); 
		this.args = args; 
	}
	public CommandArgument(CommandArgument ca, List argTrees) { 
		this(ca.constructor); 
		this.argTrees = argTrees; 
	}
	
	public Object[] parseArgs(CommandInvoker me, String[] args, Delta d) {
		d.delta = args != null ? args.length : 0;
		if (minParamCount > 0 && (args == null || args.length == 0)) return null;
		return me.parseArgs(constructor.getParameterTypes(), args, d);
	}
	
	public Object[] parseArgs(CommandObjectInvoker me, Object[] args, Delta d) {
		d.delta = args != null ? args.length : 0;
		if (minParamCount > 0 && (args == null || args.length == 0)) return null;
		return me.parseArgs(constructor.getParameterTypes(), args, d);
	}
		
	public List parseArgs(CommandTreeInvoker me, List args, Delta d) {
		d.delta = args != null ? args.length() : 0;
		if (minParamCount > 0 && (args == null || args.length() == 0)) return null;
		return me.parseArgs(constructor.getParameterTypes(), args, d);
	}
	
	public Object invoke() throws Throwable { 
		return constructor.newInstance(args);
	}

	public Tree toTree(int pos) { 
		if (argTrees == null) argTrees = new List();
		// <cmd> cmdObj = new <cmd>(<arg>);
		Name name = Name.fromString(constructor.getName());
		NewClass newClass = make.at(pos).NewClass(null, make.at(pos).Ident(name), argTrees, null);
		VarDef varDef = make.at(pos).VarDef(0, Name.fromString("cmdObj"), make.at(pos).Ident(name), newClass);
		return varDef;
	}
	
	public Tree optimize(Tree tree) { 
		// <cmd> cmdObj = new <cmd>(<arg>); => new <cmd>(<arg>) 
		return ((VarDef)tree).init;
	}
	
	public String toString() {
		String tmp = "new " + constructor.getName() + "(";
		if (args != null)
			for (int i = 0; i < args.length; i++) {
				tmp += args[i].toString();
				if (i < args.length-1) tmp += ", ";
			}
		tmp += ")";
		return tmp;
	}
}

class CommandInstance extends CommandElements
{
	private String arg, commandName;
	public Tree argTree;
	public CommandInstance() {}
	public CommandInstance(String commandName) { 
		this.commandName = commandName;
	}

	public Object[] parseArgs(CommandInvoker me, String[] args, Delta d) {
		Object[] objs = me.parseArgs(new Class[] { String.class }, args, d);
		if (objs != null && objs.length > 0) arg = (String)objs[0];
		return objs;
	}

	public Object[] parseArgs(CommandObjectInvoker me, Object[] args, Delta d) {
		Object[] objs = me.parseArgs(new Class[] { String.class }, args, d);
		if (objs != null && objs.length > 0) arg = (String)objs[0];
		return objs;
	}
	
	public List parseArgs(CommandTreeInvoker me, List args, Delta d) {
		List objs = me.parseArgs(new Class[] { String.class }, args, d);
		if (!objs.isEmpty()) {
			if (objs.head instanceof Literal)
				arg = (String)((Literal)objs.head).value;
			else {
				arg = null;
			}
		} else return null;
		return objs;
	}
	
	public Commandable invoke() throws Throwable { 
System.out.println(commandName);		
		return Commands.getInstance(arg, Commands.getClass(commandName));
	}
	
	public Tree toTree(int pos) { 
		// <cmd> cmdObj = (<cmd>)Commands.getInstance(<arg>, <cmd>.class);
		Name commandName = Name.fromString(this.commandName);
		Select select1 = make.at(pos).Select(make.at(pos).Ident(Name.fromString("Commands")), Name.fromString("getInstance"));
		Select select2 = make.at(pos).Select(make.at(pos).Ident(commandName), Name.fromString("class"));
		Apply apply = make.at(pos).Apply(select1, List.make(argTree, select2));
		TypeCast typeCast = make.at(pos).TypeCast(make.at(pos).Ident(commandName), apply);
		VarDef varDef = make.at(pos).VarDef(0, Name.fromString("cmdObj"), make.at(pos).Ident(commandName), typeCast);
		return varDef;
	}
	
	public Tree optimize(Tree tree) { 
		// <cmd> cmdObj = (<cmd>)Commands.getInstance(<arg>, <cmd>.class);
		return ((VarDef)tree).init;
	}
	
	public String toString() {
		String tmp = "(" + commandName + ")" + arg;
		return tmp;
	}
}
