package jmel.script;

import jmel.tree.*;
import jmel.tree.Tree.*;
import jmel.code.Type;
import jmel.code.Flags;
import jmel.util.*;
import java.lang.reflect.*;
import java.util.Iterator;
import java.util.Vector;
import java.util.TreeMap;
import java.util.Enumeration;
import java.io.*;

public class CommandInvoker
{
	protected String commandName;
    protected Vector constructors = new Vector();
    protected Constructor defaultConstructor;
    protected TreeMap properties = new TreeMap();
    protected TreeMap events = new TreeMap();
    protected TreeMap execs = new TreeMap();
    protected boolean isCommand, isStaticCommand;
	protected static TreeMaker make = new TreeMaker();

	// Public	
	/*
	
	A command:
	
		<type> get<Opt>()
		void set<Opt>(<args>)
		void exec()
		
	A static command:
	
		static <type> get<Opt>()
		static void set<Opt>(<args>)
		
	A script :
		static <type> exec(<args>)
	*/
	public CommandInvoker(Class c) {
		
		String commandName = c.getName();
		commandName = commandName.substring(commandName.lastIndexOf(".")+1);
		
		this.commandName = commandName;
		isCommand = Commandable.class.isAssignableFrom(c);
		isStaticCommand = isCommand && StaticCommand.class.isAssignableFrom(c);

		if (!isCommand && !Script.class.isAssignableFrom(c)) {
			throw new InternalError("Class '"+ c.getName() + "' is not a command nor a script.");
		}
		
		Constructor cs[] = c.getConstructors();
		for (int i = 0; i < cs.length; i++) {
			if (cs[i].getParameterTypes().length == 0) defaultConstructor = cs[i];
			constructors.addElement(new CommandArgument(cs[i]));		
		}
		
		boolean found = false;
		Method md[] = c.getMethods();
		int count = 0;
		for (int i = 0;i < md.length; i++) {
			Method mdi = md[i];
			String name = mdi.getName();

			if (name.startsWith("get")) {
				name = decapitalize(name.substring(3)) + "get";
				Vector v = (Vector)properties.get(name);
				if (v == null) v = new Vector();
				if ((mdi.getModifiers() & Modifier.STATIC) != 0)
					v.addElement(new CommandStaticPropertyGet(mdi));				
				else
					v.addElement(new CommandPropertyGet(mdi));				
				properties.put(name, v);
			} else if (mdi.getReturnType() == boolean.class && name.startsWith("is")) {
				name = decapitalize(name.substring(2)) + "get";
				Vector v = (Vector)properties.get(name);
				if (v == null) v = new Vector();
				if ((mdi.getModifiers() & Modifier.STATIC) != 0)
					v.addElement(new CommandStaticPropertyGet(mdi));				
				else
					v.addElement(new CommandPropertyGet(mdi));				
				properties.put(name, v);
			} else if (name.startsWith("set")) {
				name = decapitalize(name.substring(3)) + "set";
				Vector v = (Vector)properties.get(name);
				if (v == null) v = new Vector();
				if ((mdi.getModifiers() & Modifier.STATIC) != 0)
					v.addElement(new CommandStaticPropertySet(mdi));				
				else
					v.addElement(new CommandPropertySet(mdi));				
				properties.put(name, v);
			} else if ("exec".equals(name)) {
				if (isCommand) {
					execs.put(name, new CommandExec(mdi));
					// For return value
					execs.put("execAsExpr", new CommandExecAsExpr(mdi));
				} else if ((mdi.getModifiers() & Modifier.STATIC) != 0) {
					found = true;
					if (mdi.getReturnType() == void.class)
						execs.put("exec"+(count++), new CommandStatic(mdi));
					else
						execs.put("exec"+(count++), new CommandStaticAsExpr(mdi));
				}
			} else if ((mdi.getModifiers() & Modifier.STATIC) != 0) {
				// by default : options are settable
				name += "set";
				Vector v = (Vector)properties.get(name);
				if (v == null) v = new Vector();
				if (mdi.getReturnType() != void.class) {
					v.addElement(new CommandStaticPropertyGet(mdi));				
				} else {
					v.addElement(new CommandStaticPropertySet(mdi));				
				}
				properties.put(name, v);
			}
		}
		
		if (!isCommand && !found)
			throw new InternalError("No static method 'exec' found in command '"+commandName+"'.");
	}
	
	public Vector process(Vector cmdArgs) throws Throwable {
		Vector actions = new Vector();
		
		String[] args;
		Object[] objs = null;
		boolean hasConstructor = false;
		boolean isEdit = false, isQuery = false;
		boolean hasEdit = false,  hasQuery = false, hasExists = false;
		
		Delta d = new Delta();

		// Get options arguments
		Iterator i = cmdArgs.iterator();
		while (i.hasNext()) {
			String key = (String)i.next();
			args = (String[])i.next();

			objs = null;
			d.delta = 0;
			boolean hasArgs = false;
			
			if (!key.equals(Commands.ARGS)) {
				boolean oldIsQuery = isQuery;
				
				if (isCommand) {
					if ("-e".equals(key) || "-edit".equals(key)) {
						if (isEdit || isQuery) 
							throw new InternalError("Option '" + key + "' is not editable nor queryable.");
						if (properties.get(key.substring(1) + "set") == null) {
							if (args != null && args.length > 0) 
								throw new InternalError("No argument needed for option '" + key + "'.");
							isEdit = true;
							hasEdit = true;
							continue;
						}
					} 
					if ("-q".equals(key) || "-query".equals(key)) {
						if (isEdit || isQuery) 
							throw new InternalError("Option '" + key + "' is not editable nor queryable.");
						if (properties.get(key.substring(1) + "set") == null) {
							if (args != null && args.length > 0) 
								throw new InternalError("No argument needed for option '" + key + "'.");
							isQuery = oldIsQuery = true;
							hasQuery = true;
							continue;
						} 
					} 
					if ("-ex".equals(key) || "-exists".equals(key)) {
						if (isEdit || isQuery) 
							throw new InternalError("Option '" + key + "' is not editable nor queryable.");
						if (properties.get(key.substring(1) + "set") == null) {
							isQuery = oldIsQuery = true;
							hasQuery = true;
							hasExists = true;
						} 
					} 
				} 
					
				isEdit = isQuery = false;
				 
				Vector methods = (Vector)properties.get(key.substring(1) + (oldIsQuery? "get" : "set"));
				if (methods == null) {
					throw new InternalError("Option '" + key + "' unknown or not " + (oldIsQuery? "get" : "set") + "table.");
				}
				
				CommandProperty ca = null;				
				for (Enumeration e = methods.elements(); e.hasMoreElements() && objs == null; ) {
					ca = (CommandProperty)e.nextElement();
					objs = ca.parseArgs(this, args, d);
				}

				if (d.delta == 0) {
					if (!oldIsQuery && objs == null)
						throw new InternalError("Argument mismatch or missing for option '" + key + "'.");
				} else if (objs == null)
					objs = new Object[0];
				
				actions.addElement(new CommandProperty(ca, objs));

				if (d.delta > 0) {
					if (i.hasNext())
						throw new InternalError("Argument count mismatch for option '" + key + "'.");
					
					String[] cmdargs = new String[d.delta];
					System.arraycopy(args, args.length-d.delta, cmdargs, 0, d.delta);
			
					hasArgs = true;
					args = cmdargs;
				}
			} else
				hasArgs = true;
				
			objs = null;
			d.delta = 0;

			if (isCommand) {
				if (hasEdit || hasQuery) {
					if (!hasExists) {
						if (!i.hasNext()) {
							if (!hasArgs || args == null || args.length == 0) 
								;//throw new InternalError("Object name(s) required.");
							else {
								CommandInstance ci = null;
								for (int j = 0; j < args.length; j++) {
									ci = new CommandInstance(commandName);
									objs = ci.parseArgs(this, args, d);
									if (objs == null)
										throw new InternalError("Object name expected.");
									actions.add(0, ci);
									
									if (hasQuery) {
										if (args.length > 1)
											System.err.println("**Extra arguments ignored.**");
										break;
									}
								}
							}
						}
					}
													
					hasConstructor = true;
			 	} else if (hasArgs) {
					CommandArgument ca = null;
					for (Enumeration e = constructors.elements(); e.hasMoreElements() && objs == null; ) {
						ca = (CommandArgument)e.nextElement();
						objs = ca.parseArgs(this, args, d);
					}
					if (objs == null)
						throw new InternalError("Argument mismatch or missing.");
							
					if (d.delta > 0)
						System.err.println("**Extra arguments ignored.**");
	
					hasConstructor = true;
					actions.add(0, new CommandArgument(ca, objs));
				}
			} else {
				CommandProperty ce = null;
				Iterator i2 = execs.values().iterator();
				while (i2.hasNext() && objs == null) {
					ce = (CommandProperty)i2.next();
					objs = ce.parseArgs(this, args, d);
				}

				if (d.delta == 0) {
					if (objs == null)
						throw new InternalError("Argument mismatch or missing.");
				} else if (objs == null)
					objs = new Object[0];
						
				actions.add(new CommandProperty(ce, objs));
			}
		}		
		
		if (!isCommand) {
			if (cmdArgs.isEmpty()) {
				CommandProperty ce = null;
				Iterator i2 = execs.values().iterator();
				while (i2.hasNext() && objs == null) {
					ce = (CommandProperty)i2.next();
					objs = ce.parseArgs(this, (String[])null, d);
				}

				if (d.delta == 0) {
					if (objs == null)
						throw new InternalError("Argument mismatch or missing.");
				} else if (objs == null)
					objs = new Object[0];
						
				actions.add(0, new CommandProperty(ce, objs));
			}
		} else if (!hasConstructor && !isStaticCommand)
			if (defaultConstructor == null)
				throw new InternalError("Argument mismatch or missing.");
			else
				actions.add(0, new CommandArgument(defaultConstructor));
		
		return actions;
	}

	public void exec(Vector actions) throws Throwable {
		execute(actions, false);
	}

	public Object execAsExpr(Vector actions) throws Throwable {
		return execute(actions, true);
	}

	public static String decapitalize(String name) {
		if (name == null || name.length() == 0) {
			return name;
		}
		if (name.length() > 1 && Character.isUpperCase(name.charAt(1)) && Character.isUpperCase(name.charAt(0))){
			return name;
		}
		char chars[] = name.toCharArray();
		chars[0] = Character.toLowerCase(chars[0]);
		return new String(chars);
	}

	// Private
	private Object execute(Vector actions, boolean expr) throws Throwable {
		Commandable cmd = null;
		Object obj = null;
		Iterator i = actions.iterator();
		boolean special = true;
		while (i.hasNext()) {
			Object action = i.next();
			if (action instanceof CommandArgument) {
				cmd = (Commandable)((CommandArgument)action).invoke();
				special = false;
			} else if (action instanceof CommandInstance) {
				cmd = (Commandable)((CommandInstance)action).invoke();
				if (cmd == null) 
					throw new InternalError("Invalid argument.");
			} else {
				obj = ((CommandProperty)action).invoke(cmd);
			}
		}
		if (isCommand && !isStaticCommand && !special)
			if (expr)
				return Commands.execAsExpr(cmd);
			else
				Commands.exec(cmd);
		return obj;
	}

	Object[] parseArgs(Class[] parameterTypes, String[] args, Delta d) {
		if (args != null && args.length > 0 && parameterTypes.length == 0) {
			d.delta = args != null ? args.length : 0;
			return null;
		}

		d.delta = 0;
		Object objs[] = new Object[parameterTypes.length];
		int j = 0;
		try {
			for (int i = 0; i < parameterTypes.length; i++) {
				Class type = parameterTypes[i];
				
				if (type.isArray()) {
					Vector data = new Vector();
					
					type = type.getComponentType();

					if (args != null) {
						String argj = args[j];
						if (argj.startsWith("{") && argj.endsWith("}")) {
					        int k = 0;
							try {
						        Reader r = new BufferedReader(new StringReader(argj));
						        StreamTokenizer st = new StreamTokenizer(r);
						        st.resetSyntax();
						        st.wordChars(' ', 255);
						        st.whitespaceChars(0, ' ');
						        st.whitespaceChars('{', '{');
						        st.whitespaceChars('}', '}');
						        while (st.nextToken() != st.TT_EOF) {
									if (type == int.class) 
										data.addElement(isInt(st.sval));
					
									else if (type == float.class) 
										data.addElement(isFloat(st.sval));
					
									else if (type == double.class) 
										data.addElement(isDouble(st.sval));
					
									else if (type == String.class) 
										data.addElement(st.sval);
					
									else if (type == MelVector.class)
										data.addElement(isVector(st.sval));
					
									else if (type == MelMatrix.class)
										data.addElement(isMatrix(st.sval));
										
									else 
										throw new InternalError("Cannot parse type " + type);
										
									k++;
								}
						        r.close();
						    } catch (NumberFormatException e) {
						    	return null;
						    } catch (IOException e) {
						    	return null;
						    }
							j++;
						} else
							while (true) {
								if (j >= args.length) break;
								
								if (type == int.class) 
									data.addElement(isInt(args[j++]));
				
								else if (type == float.class) 
									data.addElement(isFloat(args[j++]));
				
								else if (type == double.class) 
									data.addElement(isDouble(args[j++]));
				
								else if (type == String.class) 
									data.addElement(args[j++]);
				
								else if (type == MelVector.class)
									data.addElement(isVector(args[j++]));
				
								else if (type == MelMatrix.class)
									data.addElement(isMatrix(args[j++]));
									
								else 
									throw new InternalError("Cannot parse type " + type);
							}	
					}
					
					objs[i] = Array.newInstance(type, data.size());
					int k = 0;
					for (Enumeration e = data.elements(); e.hasMoreElements(); k++)
						Array.set(objs[i], k, e.nextElement());
						
					continue;
				} 
				
				if (j >= args.length) return null;

				if (type == int.class) 
					objs[i] = isInt(args[j++]);

				else if (type == boolean.class) 
					objs[i] = isBoolean(args[j++]);

				else if (type == float.class) 
					objs[i] = isFloat(args[j++]);

				else if (type == double.class) 
					objs[i] = isDouble(args[j++]);

				else if (type == String.class) 
					objs[i] = args[j++];

				else if (type == MelVector.class)
					objs[i] = isVector(args[j++]);

				else if (type == MelMatrix.class)
					objs[i] = isMatrix(args[j++]);

				else {
					if (j >= args.length) return null;
					Constructor c;
					try {
						c = type.getConstructor(new Class[] { String.class });
					} catch (Throwable t) {
						throw new InternalError("No public constructor " + type + "(String)");
					}
					try {
						objs[i] = c.newInstance(new Object[] { args[j++] });
					} catch (InvocationTargetException t) {
						throw new InternalError(t.getTargetException().toString());
					} catch (Throwable t) {
						throw new InternalError(t.toString());
					}
				}
			}
		} catch (NumberFormatException e) {
			return null;
		}
		d.delta = args != null ? args.length - j : 0;
		return objs;
	}

	private static Integer isInt(String s) throws NumberFormatException {
		if ("true".equals(s) || 
			"yes".equals(s) || 
			"on".equals(s)) return new Integer(1);
		if ("false".equals(s) || 
			"no".equals(s) || 
			"off".equals(s)) return new Integer(0);
		return new Integer(s);
	}

	private static Boolean isBoolean(String s) throws NumberFormatException {
		if (!"true".equals(s) && 
			!"false".equals(s) && 
			!"yes".equals(s) && 
			!"no".equals(s) && 
			!"1".equals(s) && 
			!"0".equals(s) && 
			!"on".equals(s) && 
			!"off".equals(s)) throw new NumberFormatException();
		if ("true".equals(s) || 
			"yes".equals(s) || 
			"1".equals(s) || 
			"on".equals(s)) return new Boolean(true);
		return new Boolean(false);
	}

	private static Float isFloat(String s) throws NumberFormatException {
		return new Float(s);
	}
	
	private static Double isDouble(String s) throws NumberFormatException {
		return new Double(s);
	}

	private static MelVector isVector(String s) throws NumberFormatException {
		if (!s.startsWith("<<") && !s.endsWith(">>")) throw new NumberFormatException();
		return MelVector.fromString(s);
	}

	private static MelMatrix isMatrix(String s) throws NumberFormatException {
		if (!s.startsWith("<<") && !s.endsWith(">>")) throw new NumberFormatException();
		return MelMatrix.fromString(s);
	}
}