package jmel.script;

import jmel.*;
import jmel.tree.*;
import jmel.tree.Tree.*;
import jmel.code.Type;
import jmel.code.Flags;
import jmel.util.*;
import jmel.interp.*;
import java.lang.reflect.*;
import java.util.Iterator;
import java.util.Vector;
import java.util.TreeMap;
import java.util.Enumeration;

public class CommandObjectInvoker extends CommandInvoker
{
	// Public	
	public CommandObjectInvoker(Class c) {
		super(c);
	}
	
	public Vector process(Vector cmdArgs) throws Throwable {
		Vector actions = new Vector();
		
		Object[] args;
		Object[] objs = null;
		boolean hasConstructor = false;
		boolean isEdit = false, isQuery = false;
		boolean hasEdit = false,  hasQuery = false, hasExists = false;
		
		Delta d = new Delta();

		// Get options arguments
		Iterator i = cmdArgs.iterator();
		while (i.hasNext()) {
			String key = (String)i.next();
			args = (Object[])i.next();
			
			objs = null;
			d.delta = 0;
			boolean hasArgs = false;
			
			if (!key.equals(Commands.ARGS)) {
				boolean oldIsQuery = isQuery;
				
				if (isCommand) {
					if ("-e".equals(key) || "-edit".equals(key)) {
						if (isEdit || isQuery) 
							throw new InternalError("Option '" + key + "' is not editable nor queryable.");
						if (properties.get(key.substring(1) + "set") == null) {
							if (args != null && args.length > 0) 
								throw new InternalError("No argument needed for option '" + key + "'.");
							isEdit = true;
							hasEdit = true;
							continue;
						}
					} 
					if ("-q".equals(key) || "-query".equals(key)) {
						if (isEdit || isQuery) 
							throw new InternalError("Option '" + key + "' is not editable nor queryable.");
						if (properties.get(key.substring(1) + "set") == null) {
							if (args != null && args.length > 0) 
								throw new InternalError("No argument needed for option '" + key + "'.");
							isQuery = oldIsQuery = true;
							hasQuery = true;
							continue;
						} 
					} 
					if ("-ex".equals(key) || "-exists".equals(key)) {
						if (isEdit || isQuery) 
							throw new InternalError("Option '" + key + "' is not editable nor queryable.");
						if (properties.get(key.substring(1) + "set") == null) {
							isQuery = oldIsQuery = true;
							hasQuery = true;
							hasExists = true;
						} 
					} 
				}
								
				isEdit = isQuery = false;
				
				Vector methods = (Vector)properties.get(key.substring(1) + (oldIsQuery? "get" : "set"));
				if (methods == null)
					throw new InternalError("Option '" + key + "' unknown or not " + (oldIsQuery? "get" : "set") + "table.");
					
				CommandProperty ca = null;				
				for (Enumeration e = methods.elements(); e.hasMoreElements() && objs == null; ) {
					ca = (CommandProperty)e.nextElement();
					objs = ca.parseArgs(this, args, d);
				}

				if (d.delta == 0) {
					if (!oldIsQuery && objs == null)
						throw new InternalError("Argument mismatch or missing for option '" + key + "'.");
				} else if (objs == null)
					objs = new Object[0];
									
				actions.addElement(new CommandProperty(ca, objs));
				
				if (d.delta > 0) {
					if (i.hasNext())
						throw new InternalError("Argument count mismatch for option '" + key + "'.");
					
					Object[] cmdargs = new Object[d.delta];
					System.arraycopy(args, args.length-d.delta, cmdargs, 0, d.delta);
			
					hasArgs = true;
					args = cmdargs;
				}
			} else
				hasArgs = true;
				
			objs = null;
			d.delta = 0;

			if (isCommand) {
				if (hasEdit || hasQuery) {
					if (!hasExists) {
						if (!i.hasNext()) {
							if (!hasArgs || args == null || args.length == 0) 
								;//throw new InternalError("Object name(s) required.");
							else {
								CommandInstance ci = null;
								for (int j = 0; j < args.length; j++) {
									ci = new CommandInstance(commandName);
									objs = ci.parseArgs(this, args, d);
									if (objs == null)
										throw new InternalError("Object name expected.");
									actions.add(0, ci);
									
									if (hasQuery) {
										if (args.length > 1)
											System.err.println("**Extra arguments ignored.**");
										break;
									}
								}
							}
						}
					}
													
					hasConstructor = true;
			 	} else if (hasArgs) {
					CommandArgument ca = null;
					for (Enumeration e = constructors.elements(); e.hasMoreElements() && objs == null; ) {
						ca = (CommandArgument)e.nextElement();
						objs = ca.parseArgs(this, args, d);
					}

					if (objs == null)
						throw new InternalError("Argument mismatch or missing.");
							
					if (d.delta > 0)
						System.err.println("**Extra arguments ignored.**");
	
					hasConstructor = true;
					actions.add(0, new CommandArgument(ca, objs));
				}
			} else {
				CommandProperty ce = null;
				Iterator i2 = execs.values().iterator();
				while (i2.hasNext() && objs == null) {
					ce = (CommandProperty)i2.next();
					objs = ce.parseArgs(this, args, d);
				}

				if (d.delta == 0) {
					if (objs == null)
						throw new InternalError("Argument mismatch or missing.");
				} else if (objs == null)
					objs = new Object[0];

				if (d.delta > 0)
					System.err.println("**Extra arguments ignored.**");
	
				actions.add(new CommandProperty(ce, objs));
			}
		}		
		
		if (!isCommand) {
			if (cmdArgs.isEmpty()) {
				CommandProperty ce = null;
				Iterator i2 = execs.values().iterator();
				while (i2.hasNext() && objs == null) {
					ce = (CommandProperty)i2.next();
					objs = ce.parseArgs(this, (Object[])null, d);
				}

				if (d.delta == 0) {
					if (objs == null)
						throw new InternalError("Argument mismatch or missing.");
				} else if (objs == null)
					objs = new Object[0];
						
				if (d.delta > 0)
					System.err.println("**Extra arguments ignored.**");
	
				actions.add(0, new CommandProperty(ce, objs));
			}
		} else if (!hasConstructor && !isStaticCommand)
			if (defaultConstructor == null)
				throw new InternalError("Argument mismatch or missing.");
			else
				actions.add(0, new CommandArgument(defaultConstructor));
		
		return actions;
	}

   	public void setScriptVars( NameSpace namespace, JMelInterpreter interpreter, CallStack callstack, Tree callerInfo ) {
   		ScriptCaller.namespace = namespace;
   		ScriptCaller.interpreter = interpreter;
   		ScriptCaller.callstack = callstack;
   		ScriptCaller.callerInfo = callerInfo;
		ScriptCaller.insideScript = true;
   	}
   	

	// Private
	Object[] parseArgs(Class[] parameterTypes, Object[] args, Delta d) {
		if (args != null && args.length > 0 && parameterTypes.length == 0) {
			d.delta = args != null ? args.length : 0;
			return null;
		}
		
		d.delta = 0;
		Object objs[] = new Object[parameterTypes.length];
		int j = 0;
		try {
			for (int i = 0; i < parameterTypes.length; i++) {
				Class type = parameterTypes[i];
				
				if (type.isArray()) {
					Vector data = new Vector();
					
					type = type.getComponentType();
					
					if (args != null) {
						if (args.length == 1) {
							if (type == int.class) 
								if (!(args[0] instanceof MelArrayInt)) return null;
			
							else if (type == float.class) 
								if (!(args[0] instanceof MelArrayFloat)) return null;
			
							else if (type == double.class) 
								if (!(args[0] instanceof MelArrayDouble)) return null;
			
							else if (type == String.class) 
								if (!(args[0] instanceof MelArrayString)) return null;
			
							else if (type == MelVector.class)
								if (!(args[0] instanceof MelArrayVector)) return null;

							else if (type == MelMatrix.class)
								if (!(args[0] instanceof MelArrayMatrix)) return null;
								
							else 
								throw new InternalError("Cannot parse type " + type);
								
							objs[i] = ((MelArray)args[j++]).getValue();
						} else {
							while (true) {
								if (j >= args.length) break;
								
								if (type == int.class) 
									data.addElement(isInt(args[j++]));
				
								else if (type == float.class) 
									data.addElement(isFloat(args[j++]));
				
								else if (type == double.class) 
									data.addElement(isDouble(args[j++]));
				
								else if (type == String.class) 
									data.addElement(isString(args[j++]));
				
								else if (type == MelVector.class)
									data.addElement(isVector(args[j++]));
	
								else if (type == MelMatrix.class)
									data.addElement(isMatrix(args[j++]));
									
								else 
									throw new InternalError("Cannot parse type " + type);
							}	

							objs[i] = Array.newInstance(type, data.size());
							int k = 0;
							for (Enumeration e = data.elements(); e.hasMoreElements(); k++)
								Array.set(objs[i], k, e.nextElement());
						}	
					}
						
					continue;
				} 

				if (j >= args.length) return null;

				if (type == int.class) 
					objs[i] = isInt(args[j++]);

				else if (type == boolean.class) 
					objs[i] = isBoolean(args[j++]);

				else if (type == float.class) 
					objs[i] = isFloat(args[j++]);

				else if (type == double.class) 
					objs[i] = isDouble(args[j++]);

				else if (type == String.class) 
					objs[i] = isString(args[j++]);

				else if (type == MelVector.class)
					objs[i] = isVector(args[j++]);

				else if (type == MelMatrix.class)
					objs[i] = isMatrix(args[j++]);

				else {
					if (j >= args.length) return null;
					Constructor c;
					try {
						c = type.getConstructor(new Class[] { String.class });
					} catch (Throwable t) {
						throw new InternalError("No public constructor " + type + "(String)");
					}
					try {
						objs[i] = c.newInstance(new Object[] { args[j++] });
					} catch (InvocationTargetException t) {
						throw new InternalError(t.getTargetException().toString());
					} catch (Throwable t) {
						throw new InternalError(t.toString());
					}
				}
			}
		} catch (NumberFormatException e) {
			return null;
		}
		
		d.delta = args != null ? args.length - j : 0;
		
		return objs;
	}

	private static Integer isInt(Object o) throws NumberFormatException {
		if (o instanceof Number) return new Integer(((Number)o).intValue());
		if (o instanceof Boolean)
			return new Integer(((Boolean)o).booleanValue() ? 1 : 0);
		if (o instanceof String) {
			String s = (String)o;
			if ("true".equals(s) || 
				"yes".equals(s) || 
				"on".equals(s)) return new Integer(1);
			if ("false".equals(s) || 
				"no".equals(s) || 
				"off".equals(s)) return new Integer(0);
		}
		throw new NumberFormatException();
	}

	private static Boolean isBoolean(Object o) throws NumberFormatException {
		if (o instanceof Boolean)
			return (Boolean)o;

		if (o instanceof Number)
			return new Boolean(((Number)o).intValue()!=0);

		if (o instanceof String) {
			String name = (String)o;
			if (!"true".equals(name) && 
				!"false".equals(name) && 
				!"yes".equals(name) && 
				!"no".equals(name) && 
				!"on".equals(name) && 
				!"off".equals(name)) throw new NumberFormatException();
			if ("true".equals(name) || 
				"yes".equals(name) || 
				"on".equals(name)) return new Boolean(true);
			return new Boolean(false);
		}
		throw new NumberFormatException();
	}
	
	private static Float isFloat(Object o) throws NumberFormatException {
		if (o instanceof Number) return new Float(((Number)o).floatValue());
		if (o instanceof String) return new Float((String)o);
		throw new NumberFormatException();
	}

	private static Double isDouble(Object o) throws NumberFormatException {
		if (o instanceof Number) return new Double(((Number)o).doubleValue());
		if (o instanceof String) return new Double((String)o);
		throw new NumberFormatException();
	}
	
	private static String isString(Object o) throws NumberFormatException {
		if (!(o instanceof String)) throw new NumberFormatException();
		return (String)o;
	}
	
	private static MelVector isVector(Object o) throws NumberFormatException {
		if (o instanceof MelVector) return (MelVector)o;
		if (!(o instanceof String)) throw new NumberFormatException();
		String s = (String)o;
		if (!s.startsWith("<<") && !s.endsWith(">>")) throw new NumberFormatException();
		return MelVector.fromString(s);
	}
	
	private static MelMatrix isMatrix(Object o) throws NumberFormatException {
		if (o instanceof MelMatrix) return (MelMatrix)o;
		if (!(o instanceof String)) throw new NumberFormatException();
		String s = (String)o;
		if (!s.startsWith("<<") && !s.endsWith(">>")) throw new NumberFormatException();
		return MelMatrix.fromString(s);
	}
}