package jmel.script;

import jmel.tree.*;
import jmel.tree.Tree.*;
import jmel.code.Type;
import jmel.code.Flags;
import jmel.util.*;
import java.lang.reflect.*;
import java.util.Iterator;
import java.util.Vector;
import java.util.TreeMap;
import java.util.Enumeration;

public class CommandTreeInvoker extends CommandInvoker
{
	// Public	
	public CommandTreeInvoker( Class c) {
		super(c);
	}
	
	public Vector process(Vector cmdArgs, boolean expr) {
		Vector actions = new Vector();
		
		List args;
		List objs = null;
		boolean hasConstructor = false;
		boolean isEdit = false, isQuery = false;
		boolean hasEdit = false,  hasQuery = false, hasExists = false;
		
		Delta d = new Delta();

		// Get options arguments
		Iterator i = cmdArgs.iterator();
		while (i.hasNext()) {
			String key = (String)i.next();
			args = (List)i.next();
			
			objs = null;
			d.delta = 0;
			boolean hasArgs = false;
			
			if (!key.equals(Commands.ARGS)) {
				boolean oldIsQuery = isQuery;
				
				if (isCommand) {
					if ("-e".equals(key) || "-edit".equals(key)) {
						if (isEdit || isQuery) 
							throw new InternalError("Option '" + key + "' is not editable nor queryable.");
						if (properties.get(key.substring(1) + "set") == null) {
							if (!args.isEmpty()) 
								throw new InternalError("No argument needed for option '" + key + "'.");
							isEdit = true;
							hasEdit = true;
							continue;
						}
					} 
					if ("-q".equals(key) || "-query".equals(key)) {
						if (isEdit || isQuery) 
							throw new InternalError("Option '" + key + "' is not editable nor queryable.");
						if (properties.get(key.substring(1) + "set") == null) {
							if (!args.isEmpty()) 
								throw new InternalError("No argument needed for option '" + key + "'.");
							isQuery = oldIsQuery = true;
							hasQuery = true;
							continue;
						} 
					} 
					if ("-ex".equals(key) || "-exists".equals(key)) {
						if (isEdit || isQuery) 
							throw new InternalError("Option '" + key + "' is not editable nor queryable.");
						if (properties.get(key.substring(1) + "set") == null) {
							isQuery = oldIsQuery = true;
							hasQuery = true;
							hasExists = true;
						} 
					} 
				}
				
				isEdit = isQuery = false;
				 
				Vector methods = (Vector)properties.get(key.substring(1) + (oldIsQuery? "get" : "set"));
				if (methods == null)
					throw new InternalError("Option '" + key + "' unknown or not " + (oldIsQuery? "get" : "set") + "table.");
					
				CommandProperty ca = null;				
				for (Enumeration e = methods.elements(); e.hasMoreElements() && objs == null; ) {
					ca = (CommandProperty)e.nextElement();
					if (ca instanceof CommandStaticPropertyGet)
						objs = ((CommandStaticPropertyGet)ca).parseArgs(this, args, d);
					else if (ca instanceof CommandStaticPropertySet)
						objs = ((CommandStaticPropertySet)ca).parseArgs(this, args, d);
					else if (ca instanceof CommandPropertyGet)
						objs = ((CommandPropertyGet)ca).parseArgs(this, args, d);
					else
						objs = ((CommandPropertySet)ca).parseArgs(this, args, d);
				}
	
				if (d.delta == 0) {
					if ((!oldIsQuery || hasExists) && objs == null)
						throw new InternalError("Argument mismatch or missing for option '" + key + "'.");
				} else if (objs == null)
					objs = new List();
					
				if (!isCommand || isStaticCommand) {
					if (ca instanceof CommandStaticPropertyGet)
						actions.addElement(new CommandStaticPropertyGet((CommandStaticPropertyGet)ca, objs, commandName));
					else if (ca instanceof CommandStaticPropertySet)
						actions.addElement(new CommandStaticPropertySet((CommandStaticPropertySet)ca, objs, commandName));
					else if (ca instanceof CommandPropertyGet)
						actions.addElement(new CommandPropertyGet((CommandPropertyGet)ca, objs, commandName));
					else
						actions.addElement(new CommandPropertySet((CommandPropertySet)ca, objs, commandName));
				} else {
					if (oldIsQuery)
						actions.addElement(new CommandPropertyGet((CommandPropertyGet)ca, objs, hasExists ? commandName : null));
					else
						actions.addElement(new CommandPropertySet((CommandPropertySet)ca, objs, null));
				}

				if (d.delta > 0) {
					if (i.hasNext())
						throw new InternalError("Argument count mismatch for option '" + key + "'.");

					Object[] cmdargs = new Object[d.delta];
					Object[] _args_ = new Object[args.length()];
					_args_ = args.toArray(_args_);
					System.arraycopy(_args_, _args_.length-d.delta, cmdargs, 0, d.delta);

					hasArgs = true;
					args = List.make(cmdargs);
				}
			} else
				hasArgs = true;
				
			objs = null;
			d.delta = 0;

			if (isCommand) {
				if (hasEdit || hasQuery) {
					if (!hasExists) {
						if (!i.hasNext()) {
							if (!hasArgs || args.isEmpty()) 
								;//throw new InternalError("Object name(s) required.");
							else {
								CommandInstance ci = null;
								for (; args.nonEmpty(); args = args.tail) {
									ci = new CommandInstance(commandName);
									objs = ci.parseArgs(this, args, d);
									if (objs == null)
										throw new InternalError("Object name expected.");
									ci.argTree = (Tree)objs.head;
									actions.add(0, ci);
									
									if (hasQuery) {
										if (args.nonEmpty() && args.tail.nonEmpty())
											System.err.println("**Extra arguments ignored.**");
										break;
									}
								}
							}
						}
					} else {
						if (args.nonEmpty() && args.tail.nonEmpty())
							System.err.println("**Extra arguments ignored.**");
					}
											
					hasConstructor = true;
			 	} else if (hasArgs) {
					CommandArgument ca = null;
					for (Enumeration e = constructors.elements(); e.hasMoreElements() && objs == null; ) {
						ca = (CommandArgument)e.nextElement();
						objs = ca.parseArgs(this, args, d);
					}
					if (objs == null)
						throw new InternalError("Argument mismatch or missing.");
							
					if (d.delta > 0)
						System.err.println("**Extra arguments ignored.**");
	
					hasConstructor = true;
					actions.add(0, new CommandArgument(ca, objs));
				}
			} else {
				CommandProperty ce = null;
				Iterator i2 = execs.values().iterator();
				while (i2.hasNext() && objs == null) {
					ce = (CommandProperty)i2.next();
					if (ce instanceof CommandStaticAsExpr)
						objs = ((CommandStaticAsExpr)ce).parseArgs(this, args, d);
					else
						objs = ((CommandStatic)ce).parseArgs(this, args, d);
				}

				if (d.delta == 0) {
					if (objs == null)
						throw new InternalError("Argument mismatch or missing.");
				} else if (objs == null)
					objs = new List();
						
				if (d.delta > 0)
					System.err.println("**Extra arguments ignored.**");
	
				if (ce instanceof CommandStaticAsExpr)
					actions.add(new CommandStaticAsExpr((CommandStaticAsExpr)ce, objs, commandName));
				else
					actions.add(new CommandStatic((CommandStatic)ce, objs, commandName));
			}
		}		
		
		if (!isCommand) {
			if (cmdArgs.isEmpty()) {
				CommandProperty ce = null;
				List l = new List();
				Iterator i2 = execs.values().iterator();
				while (i2.hasNext() && objs == null) {
					ce = (CommandProperty)i2.next();
					if (ce instanceof CommandStaticAsExpr)
						objs = ((CommandStaticAsExpr)ce).parseArgs(this, l, d);
					else
						objs = ((CommandStatic)ce).parseArgs(this, l, d);
				}

				if (d.delta == 0) {
					if (objs == null)
						throw new InternalError("Argument mismatch or missing.");
				} else if (objs == null)
					objs = new List();
						
				if (d.delta > 0)
					System.err.println("**Extra arguments ignored.**");
	
				if (expr)
					actions.add(0, new CommandStaticAsExpr((CommandStaticAsExpr)ce, objs, commandName));
				else
					actions.add(0, new CommandStatic((CommandStatic)ce, objs, commandName));
			}
		} else if (!isStaticCommand) {
			if (!hasEdit && !hasQuery && !hasExists) {
				if (expr)
					actions.addElement(new CommandExecAsExpr((CommandExecAsExpr)execs.get("execAsExpr")));
				else
					actions.addElement(new CommandExec((CommandExec)execs.get("exec")));
			}
			
			if (!hasConstructor)
				if (defaultConstructor == null)
					throw new InternalError("Argument mismatch or missing.");
				else
					actions.add(0, new CommandArgument(defaultConstructor));
		}
		
		return actions;
	}
	
	public Tree exec(int pos, Vector actions, boolean expr, int methCnt) {

		Tree type, tree = null, tree2 = null;
		ListBuffer defs = new ListBuffer();
		CommandElements action = null, action2 = null;
		Class returnType = null, rt;
		
		Iterator i = actions.iterator();
		if (i.hasNext())
			while (true) {
				if (action2 != null) {
					if (action == action2) break;
					action = action2;
					tree = tree2;
				} else {
					if (!i.hasNext()) break;
					action = (CommandElements)i.next();
					tree = action.toTree(pos);
				}
				if (tree != null) defs.append(tree);
				rt = action.getReturnType();
				if (rt != null) {
					if (rt != void.class) 
						returnType = rt;
					else if (action instanceof CommandExecAsExpr)
						returnType = Object.class;
				}

				if (i.hasNext()) {
					action2 = (CommandElements)i.next();
					tree2 = action2.toTree(pos);

					// Rule2: ARG+EXEC alone => returns optimized tree ARG%EXEC
					// Rule3: INST+EXEC alone => returns optimized tree INST%EXEC
					if (!i.hasNext())
						if ((action instanceof CommandArgument ||
							action instanceof CommandInstance) &&
							(action2 instanceof CommandPropertyGet || 
							action2 instanceof CommandPropertySet)) 
							return action2.optimize(tree, tree2);
				} else {
					// Rule1: STATIC alone => returns optimized tree
					// Rule1: GET alone => returns optimized tree
					if (action2 == null) {
						if (action instanceof CommandStaticAsExpr || 
							action instanceof CommandStatic) 
							return action.optimize(tree);
						if (action instanceof CommandPropertyGet || 
							action instanceof CommandPropertySet) 
							return action.optimize(tree);
						if (action instanceof CommandArgument || 
							action instanceof CommandInstance) 
							return action.optimize(tree);
					}
				}
			}

		if (!expr || returnType == null)
			type = make.at(pos).TypeIdent( Type.VOID );
		else {
			if (returnType == int.class)
				type = make.at(pos).TypeIdent( Type.INT );
			else if (returnType == boolean.class)
				type = make.at(pos).TypeIdent( Type.BOOLEAN );
			else if (returnType == float.class)
				type = make.at(pos).TypeIdent( Type.FLOAT );
			else if (returnType == double.class)
				type = make.at(pos).TypeIdent( Type.DOUBLE );
			else if (returnType == MelVector.class)
				type = make.at(pos).Ident( Name.fromString("MelVector") );
			else if (returnType == MelMatrix.class)
				type = make.at(pos).Ident( Name.fromString("MelMatrix") );
			else if (MelArrayInt.class.isAssignableFrom(returnType))
				type = make.at(pos).Ident( Name.fromString("MelArrayInt") );
			else if (MelArrayFloat.class.isAssignableFrom(returnType))
				type = make.at(pos).Ident( Name.fromString("MelArrayFloat") );
			else if (MelArrayDouble.class.isAssignableFrom(returnType))
				type = make.at(pos).Ident( Name.fromString("MelArrayDouble") );
			else if (MelArrayVector.class.isAssignableFrom(returnType))
				type = make.at(pos).Ident( Name.fromString("MelArrayVector") );
			else if (MelArrayString.class.isAssignableFrom(returnType))
				type = make.at(pos).Ident( Name.fromString("MelArrayString") );
			else
				type = toName(pos, returnType.getName() );
		}
		
		if (!expr) return make.at(pos).Block(0, defs.toList());
	
		if (expr)
			// return returnValue;
			defs.append(make.at(pos).Return( make.at(pos).Ident( CommandElements.RETURNVALUE ) ));
		List l = defs.toList();
		if (expr)
			// <type> returnValue;
			l = l.prepend(make.at(pos).VarDef(0, CommandElements.RETURNVALUE, type, null));
		Name n = Name.fromString("__cmd" + methCnt + "_" + commandName);
		MethodDef methodDef = make.at(pos).MethodDef( Flags.PRIVATE | Flags.STATIC, 
				n, type, Tree.emptyList, Tree.emptyList, Tree.emptyList, make.at(pos).Block(0, l));													
		return methodDef;
	}
		
	// Private
	private String toJavaLang(String name) {
		if (name.indexOf('.') != -1) return name;
		if 		(name.equals("char") 	|| name.equals("C")) name = "Character";
		else if (name.equals("byte") 	|| name.equals("B")) name = "Byte";
		else if (name.equals("short") 	|| name.equals("S")) name = "Short";
		else if (name.equals("int") 	|| name.equals("I")) name = "Integer";
		else if (name.equals("long") 	|| name.equals("J")) name = "Long";
		else if (name.equals("float") 	|| name.equals("F")) name = "Float";
		else if (name.equals("double") 	|| name.equals("D")) name = "Double";
		else if (name.equals("boolean") || name.equals("Z")) name = "Boolean";
		return name;
	}

	private Tree toName(int pos, String name) {
		String n = "";
		if (name.startsWith("[")) {
			name = toJavaLang(name.substring(1));
			Tree t = toName(pos, name);
			return make.at(pos).TypeArray(t, 0);
		}
		int p = name.indexOf('.');
		if (p == -1) return make.at(pos).Ident(Name.fromString(name));
		Tree selected = null;
		while (p != -1) {
			n = name.substring(0,p);
			if (selected == null) 
				selected = make.at(pos).Ident(Name.fromString(n));
			else
				selected = make.at(pos).Select(selected, Name.fromString(n));

			name = name.substring(p+1);
			p = name.indexOf('.');
		}
		return make.at(pos).Select(selected, Name.fromString(name));
	}

	List parseArgs(Class[] parameterTypes, List args, Delta d) {
		if (args.length() > 0 && parameterTypes.length == 0) {
			d.delta = args.length();
			return null;
		}
		
		d.delta = 0;
		ListBuffer objs = new ListBuffer();
		int j = 0, length = args.length();
		try {
			for (int i = 0; i < parameterTypes.length; i++) {
				Class type = parameterTypes[i];

				if (type.isArray()) {
					ListBuffer data = new ListBuffer();
					Tree datatype = null;
					Tree tree = (Tree)args.head;
					type = type.getComponentType();

					int pos = tree != null ? tree.pos : 0;
							
					if (type == int.class) 
						datatype = make.at(pos).TypeIdent(Type.INT);
					else if (type == float.class) 
						datatype = make.at(pos).TypeIdent(Type.FLOAT);
					else if (type == double.class) 
						datatype = make.at(pos).TypeIdent(Type.DOUBLE);
					else if (type == String.class) 
						datatype = make.at(pos).Ident(Name.fromString("String"));
					else if (type == MelVector.class) 
						datatype = make.at(pos).Ident(Name.fromString("MelVector"));
					else if (type == MelMatrix.class) 
						datatype = make.at(pos).Ident(Name.fromString("MelMatrix"));

					if (j < length) {
						if (args.length() == 1) {
							if (type == int.class) 
								isArrayFromType((Tree)args.head, "MelArrayInt");
			
							else if (type == float.class) 
								isArrayFromType((Tree)args.head, "MelArrayFloat");
			
							else if (type == double.class) 
								isArrayFromType((Tree)args.head, "MelArrayDouble");
			
							else if (type == String.class) 
								isArrayFromType((Tree)args.head, "MelArrayString");
			
							else if (type == MelVector.class)
								isArrayFromType((Tree)args.head, "MelArrayVector");

							else if (type == MelMatrix.class)
								isArrayFromType((Tree)args.head, "MelArrayMatrix");
								
							else 
								throw new InternalError("Cannot parse type " + type);
								
							j++;
								
							Select select = make.at(pos).Select((Tree)args.head, Name.fromString("getValue"));
							Apply apply = make.at(pos).Apply(select, Tree.emptyList);
							objs.append( make.at(pos).TypeCast( make.at(pos).TypeArray(datatype, 0), apply) );
						} else {
							while (true) {
								if (j >= length) break;
								
								if (type == int.class) 
								{
									data.append(isInt((Tree)args.head));
									args = args.tail;
									j++;
								}
				
								else if (type == float.class) 
								{
									data.append(isFloat((Tree)args.head));
									args = args.tail;
									j++;
								}
				
								else if (type == double.class) 
								{
									data.append(isDouble((Tree)args.head));
									args = args.tail;
									j++;
								}
				
								else if (type == String.class) 
								{
									data.append(isString((Tree)args.head));
									args = args.tail;
									j++;
								}
				
								else if (type == MelVector.class)
								{
									data.append(isVector((Tree)args.head));
									args = args.tail;
									j++;
								}
				
								else if (type == MelMatrix.class)
								{
									data.append(isMatrix((Tree)args.head));
									args = args.tail;
									j++;
								}
									
								else 
									throw new InternalError("Cannot parse type " + type);
							}
							
							objs.append(  make.at(pos).NewArray(datatype, Tree.emptyList, data.toList()) );
						}	
					}	
										
					continue;
				} 
				
				if (j >= length) return null;
				if (type == int.class) 
				{
					objs.append(isInt((Tree)args.head));
					args = args.tail;
					j++;
				}

				else if (type == boolean.class) 
				{
					objs.append(isBoolean((Tree)args.head));
					args = args.tail;
					j++;
				}

				else if (type == float.class) 
				{
					objs.append(isFloat((Tree)args.head));
					args = args.tail;
					j++;
				}

				else if (type == double.class) 
				{
					objs.append(isDouble((Tree)args.head));
					args = args.tail;
					j++;
				}

				else if (type == String.class) 
				{
					objs.append(isString((Tree)args.head));
					args = args.tail;
					j++;
				}

				else if (type == MelVector.class)
				{
					objs.append(isVector((Tree)args.head));
					args = args.tail;
					j++;
				}
									
				else if (type == MelMatrix.class) {
					objs.append(isMatrix((Tree)args.head));
					args = args.tail;
					j++;
				}

				else {
					if (j >= length) return null;
					Constructor c;
					Tree newClass;
					try {
						c = type.getConstructor(new Class[] { String.class });
					} catch (Throwable t) {
						throw new InternalError("No public constructor " + type + "(String)");
					}
					try {
						Tree tree = (Tree)args.head, tree2;
						if (!(tree instanceof Literal) || ((Literal)tree).typetag != Type.CLASS) {
							ListBuffer args2 = new ListBuffer();
							args2.append(make.at(tree.pos).Literal(Type.CLASS, ""));
							args2.append(tree);
							tree2 = make.at(tree.pos).Operation(Tree.PLUS, args2.toList());
						} else
							tree2 = tree;
						newClass = make.at(tree.pos).NewClass(null, make.at(tree.pos).Ident(Name.fromString(type.getName())), Tree.emptyList.append(tree2), null);
						objs.append( newClass );
						args = args.tail;
						j++;
					} catch (Throwable t) {
						throw new InternalError(t.toString());
					}
				}
			}
		} catch (NumberFormatException e) {
			return null;
		}
		d.delta = length - j;

		return objs.toList();
	}

	private static Tree isArrayFromType(Tree s, String type) throws NumberFormatException {
		if (s instanceof NewClass) {
			if (((NewClass)s).clazz instanceof Ident) {
				String name = ((Ident)((NewClass)s).clazz).name.toString();
				if (name.equals("jmel.script."+type) || 
					name.equals(type)) return s;
			}
		} else 
		if (s instanceof Ident ||
			s instanceof Apply ||
			s instanceof Assign
			) return s;
		throw new NumberFormatException();
	}

	private static Tree isFromType(Tree s, int type) throws NumberFormatException {
		if (s instanceof Literal) {
			if (((Literal)s).typetag == type) return s;
		} else 
		if (s instanceof Operation) {
			if (((Operation)s).tag == Tree.NEG ||
				((Operation)s).tag == Tree.PLUS) {
				isFromType((Tree) ((Operation)s).args.head, type);
				return s;
			}
		} else
		if (s instanceof TypeCast) {
			isFromType(((TypeCast)s).clazz, type);
			return s;
		} else
		if (s instanceof Ident ||
			s instanceof Apply ||
			s instanceof Assign ||
			s instanceof Assignop ||
			s instanceof Indexed
			) return s;
		throw new NumberFormatException();
	}
	
	private static Tree isInt(Tree s) throws NumberFormatException {
		if (s instanceof Literal) {
			if (((Literal)s).typetag == Type.CLASS) {
				String st = (String)((Literal)s).value;
				if ("true".equals(st) || 
					"yes".equals(st) || 
					"on".equals(st)) return make.at(s.pos).Literal(Type.INT, new Integer(1));
				if ("false".equals(st) || 
					"no".equals(st) || 
					"off".equals(st)) return make.at(s.pos).Literal(Type.INT, new Integer(0));
			}
		}
		return isFromType(s, Type.INT);
	}

	private static Tree isBoolean(Tree s) throws NumberFormatException {
		if (s instanceof Ident) {
			String name = ((Ident)s).name.toString();
			if (!"true".equals(name) && 
				!"false".equals(name) && 
				!"yes".equals(name) && 
				!"no".equals(name) && 
				!"on".equals(name) && 
				!"off".equals(name)) throw new NumberFormatException();
			if ("true".equals(name) || 
				"yes".equals(name) || 
				"on".equals(name)) return make.at(s.pos).Ident(Name.fromString("true"));
			return make.at(s.pos).Ident(Name.fromString("false"));
		}
		if (s instanceof Literal) {
			Object value = ((Literal)s).value;
			if (value instanceof Number)
				return make.at(s.pos).Ident(Name.fromString(((Number)value).intValue()!=0? "true":"false"));
			
			String name = (String)value;
			if (!"true".equals(name) && 
				!"false".equals(name) && 
				!"yes".equals(name) && 
				!"no".equals(name) && 
				!"on".equals(name) && 
				!"off".equals(name)) throw new NumberFormatException();
			if ("true".equals(name) || 
				"yes".equals(name) || 
				"on".equals(name)) return make.at(s.pos).Ident(Name.fromString("true"));
			return make.at(s.pos).Ident(Name.fromString("false"));
		}
		isFromType(s, Type.INT);
		return make.at(s.pos).Operation(Tree.NE, List.make(s, make.at(s.pos).Literal(Type.INT, new Integer(0))));
	}

	private static Tree isFloat(Tree s) throws NumberFormatException {
		if (s instanceof Literal) {
			if (((Literal)s).typetag == Type.INT) return s;
			if (((Literal)s).typetag == Type.FLOAT) return s;
		}
		return isFromType(s, Type.FLOAT);
	}
	
	private static Tree isDouble(Tree s) throws NumberFormatException {
		if (s instanceof Literal) {
			if (((Literal)s).typetag == Type.INT) return s;
			if (((Literal)s).typetag == Type.FLOAT) return s;
			if (((Literal)s).typetag == Type.DOUBLE) return s;
		}
		return isFromType(s, Type.DOUBLE);
	}
	
	private static Tree isString(Tree s) throws NumberFormatException {
		if (s instanceof NewClass) {
			if (!(((NewClass)s).clazz instanceof Ident)) throw new NumberFormatException();
			Ident ident = (Ident)((NewClass)s).clazz;
			String name = ident.name.toString();
			if (!name.equals("String") && !name.equals("jmel.script.MelMatrix")) throw new NumberFormatException();
			return s;
		}
		return isFromType(s, Type.CLASS);
	}
	
	private static Tree isVector(Tree s) throws NumberFormatException {
		if (s instanceof NewClass) {
			if (!(((NewClass)s).clazz instanceof Ident)) throw new NumberFormatException();
			Ident ident = (Ident)((NewClass)s).clazz;
			String name = ident.name.toString();
			if (!name.equals("MelVector") && !name.equals("jmel.script.MelVector")) throw new NumberFormatException();
			return s;
		}
		return isFromType(s, -1);
	}
	
	private static Tree isMatrix(Tree s) throws NumberFormatException {
		if (s instanceof NewClass) {
			if (!(((NewClass)s).clazz instanceof Ident)) throw new NumberFormatException();
			Ident ident = (Ident)((NewClass)s).clazz;
			String name = ident.name.toString();
			if (!name.equals("MelMatrix") && !name.equals("jmel.script.MelMatrix")) throw new NumberFormatException();
			return s;
		}
		return isFromType(s, -1);
	}
}
