package jmel.script;

import java.lang.reflect.*;
import java.io.*;
import java.util.*;
import java.net.URLClassLoader;
import jmel.util.*;
import jmel.tree.Tree;
import jmel.code.Type;

public class Commands
{
	public static final String ARGS = "<args>";
	public static PathClassLoader pathClassLoader = new PathClassLoader();
	protected static TreeMap cmds = new TreeMap();
	private static HashMap templates = new HashMap();	
	private static java.util.Hashtable cmdInstances = new java.util.Hashtable();
	private static Stack objs = new Stack();

	private CommandInvoker sc;
	private Vector actions = new Vector();
	private int cnt = 0;
	private String arg[];
	private Object obj[];
	private jmel.util.List tree;
	
	// Public
	public static Commands get(String name) {
		CommandInvoker sc = (CommandInvoker)cmds.get(name);
		if (sc == null) {
			sc = new CommandInvoker(getClass(name));
			cmds.put(name, sc);
		}
		return new Commands(sc);
	}
	
	public void exec(String[] args) {
		this.arg = args;
		process(false);
	}
	
	public void exec(String arg) {
		this.arg = new String[] { arg };
		process(false);
	}
	
	public void exec() {
		this.arg = new String[0];
		process(false);
	}
	
	public Object execAsExpr(String[] args) {
		this.arg = args;
		return process(true);
	}
	
	public Object execAsExpr(String arg) {
		this.arg = new String[] { arg };
		return process(true);
	}
	
	public Object execAsExpr() {
		this.arg = new String[0];
		return process(true);
	}
	
	public static void exec(Commandable cmd) {
		if (!withCatch) cmd.exec();
		else
			try {
				cmd.exec();
			} catch (Throwable t) {
				catchThrowed = true;
			}
	}
	
	public static Object execAsExpr(Commandable cmd) {
		exec(cmd);
		return catchThrowed ? null : cmd.returnValue.get();
	}
	
	public static String getAlias(String propName) {
		return System.getProperty("alias."+propName);
	}

	//
	// Instances
	//
	public static Commandable getInstance(String name, Class _class) {
		if (name == null) return null;
		TreeMap map = (TreeMap)cmdInstances.get(_class);
		if (map == null) return null;
		Commandable cmd = (Commandable)map.get(name);
		if (cmd == null)
			for (Iterator i = map.keySet().iterator(); i.hasNext(); ) {
				String cmdName = (String)i.next();
				if (cmdName.endsWith("|" + name)) 
					if (cmd == null)
						cmd = (Commandable)map.get(cmdName);
					else
						throw new InternalError("Several objects exist width name '"+name+"'");
			}
		return cmd;
	}
	
	public static Vector getInstances(Class _class) {
		Vector v = new Vector();
		TreeMap map = (TreeMap)cmdInstances.get(_class);
		if (map == null) return v;
		Iterator it = map.values().iterator();
		while (it.hasNext()) {
			Commandable cmd = (Commandable)it.next();
			v.addElement(cmd);
		}
		return v;
	}
	
	public static Vector getInstancesAssignableFrom(String name, Class _class) {
		Vector v = new Vector();
		for (java.util.Enumeration e = cmdInstances.keys(); e.hasMoreElements(); ) {
			Class myClass = (Class)e.nextElement();
			if (_class.isAssignableFrom(myClass)) {
				TreeMap map = (TreeMap)cmdInstances.get(myClass);
				Commandable cmd = (Commandable)map.get(name);
				if (cmd == null)
					for (Iterator i = map.keySet().iterator(); i.hasNext(); ) {
						String cmdName = (String)i.next();
						if (cmdName.endsWith("|" + name)) {
							cmd = (Commandable)map.get(cmdName);
							v.addElement(cmd);
						}
					}
				else
					v.addElement(cmd);
			}
		}
		return v;
	}
	
	public static Vector getInstancesAssignableFrom(Class _class) {
		Vector v = new Vector();
		for (java.util.Enumeration e = cmdInstances.keys(); e.hasMoreElements(); ) {
			Class myClass = (Class)e.nextElement();
			if (_class.isAssignableFrom(myClass)) {
				TreeMap map = (TreeMap)cmdInstances.get(myClass);
				Iterator it = map.values().iterator();
				while (it.hasNext()) {
					Commandable cmd = (Commandable)it.next();
					v.addElement(cmd);
				}
			}
		}
		return v;
	}
	
	public static Vector getInstances(String name) {
		Vector v = new Vector();
		if (name == null) return v;
		for (java.util.Enumeration e = cmdInstances.keys(); e.hasMoreElements(); ) {
			TreeMap map = (TreeMap)cmdInstances.get((Class)e.nextElement());
			Commandable cmd = (Commandable)map.get(name);
			if (cmd == null)
				for (Iterator i = map.keySet().iterator(); i.hasNext(); ) {
					String cmdName = (String)i.next();
					if (cmdName.endsWith("|" + name)) {
						cmd = (Commandable)map.get(cmdName);
						v.addElement(cmd);
					}
				}
			else
				v.addElement(cmd);
		}
		return v;
	}
	
	public static void addInstance(String name, Commandable cmd) {
		Class _class = cmd.getClass();
		TreeMap map = (TreeMap)cmdInstances.get(_class);
		if (map == null) {
			map = new TreeMap();
			map.put(name, cmd);
			cmdInstances.put(_class, map);
		} else {
			if (map.get(name) != null)
				throw new InternalError("Object '" + name + "' already exists.");
			map.put(name, cmd);			
		}
		cmd.returnValue.set(name);
	}
	
	public static void removeInstance(String name, Class _class) {
		TreeMap map = (TreeMap)cmdInstances.get(_class);
		if (map != null) map.remove(name);
	}

	public static boolean existsInstance(String name, Class _class) {
		return getInstance(name, _class) != null;
	}

	//
	// Template
	//
	public static void createTemplate(String name) { 
		TreeMap mytemplates = (TreeMap)templates.get(name);
		if (mytemplates != null)
			throw new InternalError("Template '" + name +"' already exists.");
		templates.put(name, new TreeMap());
	}
	
	public static void removeTemplate(String name) { 
		templates.remove(name);
	}

	public static boolean existsTemplate(String name) { 
		return templates.get(name) != null;
	}

	public static void defineTemplate(String name, Templatable cmd) { 
		String className = cmd.getClass().getName();
		TreeMap mytemplates = (TreeMap)templates.get(name);
		if (mytemplates == null)
			throw new InternalError("Unknown template '" + name +"'.");
		mytemplates.put(className, cmd);
	}

	public static void affectTemplate(String name, Templatable cmd) { 
		String className = cmd.getClass().getName();
		TreeMap mytemplates = (TreeMap)templates.get(name);
		if (mytemplates == null)
			throw new InternalError("Unknown template '" + name +"'.");
		Templatable templ = (Templatable)mytemplates.get(className);
		if (templ == null)
			throw new InternalError("Unknown template '" + name +"' for command '" + className + "'");
		if (!templ.getClass().equals(cmd.getClass()))
			throw new InternalError("Invalid template '" + name +"': '" + className + "' type expected.");
		cmd.init(templ);
	}

	//
	// General functions
	//
	public static int size(Object o) { 
		if (o instanceof String) return size((String)o);
		if (o instanceof MelArray) return size((MelArray)o);
		throw new InternalError("Bad argument: "+o.getClass().getName()); 
	}
	public static int size(String s) { return s.length(); }
	public static int size(MelArray a) { return a.getSize(); }
	
	public static void clear(Object o) { 
		if (o instanceof MelArray) 
			clear((MelArray)o);
		else
		throw new InternalError("Bad argument: "+o.getClass().getName()); 
	}
	public static void clear(MelArray a) { a.clear(); }
	
	public static PrintStream out = System.out;
	
	public static void warning(String opt, String s, int line) {
		if (!opt.equals("-sl") && !opt.equals("-showLineNumber"))
			throw new InternalError("Unknown option '" + opt + "' for command 'warning'.");
		out.println("// Warning: line " + line + ": " + s + " //");
	}
	
	public static void warning(String s, int line) {
		out.println("// Warning: " + s + " //");
	}
	
	public static void print(Object s) { 
		if (s instanceof Boolean)
			print(((Boolean)s).booleanValue());
		else if (s instanceof Integer)
			print(((Integer)s).intValue());
		else if (s instanceof Float)
			print(((Float)s).floatValue());
		else if (s instanceof Double)
			print(((Double)s).doubleValue());
		else if (s instanceof String)
			print((String)s);
		else if (s == null)
			print("");
		else if (s.getClass().isArray()) {
			printArray(s);
		} else
			out.print(s.toString()); 
	}
	public static void print(int s) { out.print(s); }
	public static void print(float s) { out.print(s); }
	public static void print(double s) { out.print(s); }
	public static void print(String s) { out.print(s); }
	public static void print(boolean s) { out.print(s); }
	public static void printArray(Object s) { 
		if (s == null) return;
		int length = Array.getLength(s);
		for (int i = 0; i < length; i++) {
			out.print(Array.get(s,i));
			if (i < length-1) out.print(" ");
		}
	}
	public static void print(int[] s) { printArray(s); }
	public static void print(float[] s) { printArray(s); }
	public static void print(double[] s) { printArray(s); }
	public static void print(String[] s) { printArray(s); }
	public static void print() { out.print(""); }

	
	private static TreeMap globalVars = new TreeMap();
	public static void setGlobalVar(String name, int value) { globalVars.put(name, new Integer(value)); }
	public static void setGlobalVar(String name, float value) { globalVars.put(name, new Float(value)); }
	public static void setGlobalVar(String name, double value) { globalVars.put(name, new Double(value)); }
	public static void setGlobalVar(String name, String value) { globalVars.put(name, value); }
	public static void setGlobalVar(String name, MelVector value) { globalVars.put(name, value); }
	public static void setGlobalVar(String name, MelMatrix value) { globalVars.put(name, value); }
	public static void setGlobalVar(String name, MelArray value) { globalVars.put(name, value); }
	public static void setGlobalVar(String name, Object value) { globalVars.put(name, value); }
	
	public static int getGlobalVar(String name, int value) {
		return ((Integer)globalVars.get(name)).intValue();
	}
	public static float getGlobalVar(String name, float value) {
		return ((Float)globalVars.get(name)).floatValue();
	}
	public static double getGlobalVar(String name, double value) {
		return ((Double)globalVars.get(name)).doubleValue();
	}
	public static String getGlobalVar(String name, final String value) {
		return (String)globalVars.get(name);
	}
	public static MelVector getGlobalVar(String name, final MelVector value) {
		return (MelVector)globalVars.get(name);
	}
	public static MelMatrix getGlobalVar(String name, final MelMatrix value) {
		return (MelMatrix)globalVars.get(name);
	}
	
	public static MelArrayInt getGlobalVar(String name, final MelArrayInt value) {
		return (MelArrayInt)globalVars.get(name);
	}
	public static MelArrayFloat getGlobalVar(String name, final MelArrayFloat value) {
		return (MelArrayFloat)globalVars.get(name);
	}
	public static MelArrayString getGlobalVar(String name, final MelArrayString value) {
		return (MelArrayString)globalVars.get(name);
	}
	public static MelArrayDouble getGlobalVar(String name, final MelArrayDouble value) {
		return (MelArrayDouble)globalVars.get(name);
	}
	public static MelArrayVector getGlobalVar(String name, final MelArrayVector value) {
		return (MelArrayVector)globalVars.get(name);
	}
	
	public static boolean isGlobalVar(String name) {
		return globalVars.containsKey(name);
	}

	
	static boolean catchThrowed = false;
	static boolean withCatch = false;
	public static Commands doCatch() { withCatch = true; return new Commands(); }
	public static boolean tryCatch(int o) { return isCatchThrowed(); }
	public static boolean tryCatch(boolean o) { return isCatchThrowed(); }
	public static boolean tryCatch(float o) { return isCatchThrowed(); }
	public static boolean tryCatch(double o) { return isCatchThrowed(); }
	public static boolean tryCatch(Object o) { return isCatchThrowed(); }
	static boolean isCatchThrowed() { 
		boolean answer = catchThrowed;
		catchThrowed = withCatch = false;
		return answer;
	}
	
	private static boolean usePathLoader = Boolean.getBoolean("pathloader");
	private static Class loadClass(String name) throws ClassNotFoundException, NoClassDefFoundError {	
		String alias = getAlias(name);
		return usePathLoader ? pathClassLoader.loadClass(alias != null ? alias : name) : Class.forName(alias != null ? alias : name);
	}
	
	public static Class getClass(String name) {
		try {
			return loadClass(name);
		} catch (ClassNotFoundException e) {
		} catch (NoClassDefFoundError e) {
		}
		try {
			if (name.lastIndexOf(".") != -1) {
				String other = name.substring(0, name.lastIndexOf("."));
				String n = name.substring(name.lastIndexOf(".")+1);
				char c = n.charAt(0);
				return loadClass(other + "." + Character.toUpperCase(c) + n.substring(1));
			} else {
				char c = name.charAt(0);
				return loadClass(Character.toUpperCase(c) + name.substring(1));
			}
		} catch (ClassNotFoundException e) {
			throw new InternalError("Command not found: '" + name + "'.");
		} catch (NoClassDefFoundError e) {
			throw new InternalError("Command not found: '" + name + "'.");
		}
	}
	
	// Protected
	protected Commands() {}
	
	// Private
	private Commands(CommandInvoker sc) { this.sc = sc; }
	
	private Object process(boolean expr) 	{

		Vector cmdArgs = new Vector();
		while (cnt < arg.length) {
			
			if (arg[cnt].charAt(0) == '-' && (arg[cnt].charAt(1) >= 'A' && arg[cnt].charAt(1) <= 'z')) {
				String optName = arg[cnt++];
				if (optName.charAt(1) == '"')
					throw new InternalError("Syntax error: unexpected '\"' after '-'.");
				
				String[] optArgs = null;
				String optArg = null;
				
				while (cnt < arg.length) {
					optArg = next();
					if (optArg == null) break;
					optArgs = append(optArgs, optArg);
				}
				
				cmdArgs.add(optName);
				cmdArgs.add(optArgs);
			} else {
				String[] args = null;
				String optArg = null;
				while (cnt < arg.length) {
					optArg = next();
					if (optArg == null) 
						throw new InternalError("Syntax error: misplaced option.");
					args = append(args, optArg);
				}

				cmdArgs.add(ARGS);
				cmdArgs.add(args);
			}				
		}
		
		try {
			actions = sc.process(cmdArgs);
			Object value = sc.execAsExpr(actions);
			if (expr) return value;
		} catch (Throwable t) {
			if (Boolean.getBoolean("debug")) t.printStackTrace();			
			if (withCatch) catchThrowed = true;
			else if (t instanceof java.lang.reflect.InvocationTargetException)
				throw new InternalError(((java.lang.reflect.InvocationTargetException)t).getTargetException().toString());			
			else
				throw new InternalError(t.toString());			
		}
	
		return null;
	}
	
	private String next() {

		String s = "", argcnt = arg[cnt];
		if (argcnt.length() > 0 && argcnt.charAt(0) == '"') {
			cnt++;
			while (cnt < arg.length && (argcnt.charAt(argcnt.length()-1) != '"' || argcnt.charAt(argcnt.length()-2) != '\\')) s += (argcnt=arg[cnt++]) + " ";
			if (cnt >= arg.length)
				throw new InternalError("Syntax error: missing '\"'.");
		} else if (argcnt.length() > 0 && argcnt.charAt(0) == '-' && (argcnt.charAt(1) >= 'A' && argcnt.charAt(1) <= 'z')) 
			return null;
		else
			s = argcnt;
		cnt++;
		return s;
	}

	// Static
	static String[] append(String[] s, String a) {
		int l;
		String tmp[];
		if (s != null) {
			l = s.length;
			tmp = new String[l+1];
			System.arraycopy(s, 0, tmp, 0, l);
		} else {
			l = 0;
			tmp = new String[1];
		}
		tmp[l] = a;
		return tmp;
	}
}
