package jmel.script;

import java.io.File;
import java.io.FileInputStream;

/**
 * A simple class loader capable of loading from a local directory. 
 *
 */
public class FileClassLoader extends MyClassLoader 
{
	private String mPrefix;

	public FileClassLoader (String dir) 
	{
		mPrefix = dir;
		if ( !mPrefix.endsWith(File.separator) ) mPrefix += File.separator;
	}


	// Protected
	protected byte[] loadClassBytes (String class_name) 
	{
		class_name = formatClassName(class_name);
		
		byte result[] = null;
		File file = new File( mPrefix + class_name );
		if ( file.length() != 0 ) {
			FileInputStream in = null;
			try 
			{
				in = new FileInputStream(file);
				result = new byte[(int)file.length()];
				in.read(result);
				in.close();
				return result;
			} catch (Exception e) {
				// If we caught an exception, either the class
				// wasn't found or it was unreadable by our process.
			} finally {
				if ( in != null ) {
					try { 
						in.close(); 
					} 
					catch (Exception e) { 
						// ignore 
					}
				}
			}
		}
		
		// if no class file is found
		return null;
	}

	protected String formatClassName (String class_name) 
	{
		// replace with the default separator character for files on this system
		return class_name.replace('.',File.separatorChar) + ".class";
	}
}
