package jmel.script;

import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import java.util.jar.JarInputStream;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Hashtable;

/**
 * A simple class loader capable of loading from
 * local jar files. 
 */
public class JarClassLoader extends MyClassLoader 
{
	private String mJarFileName;
	private Hashtable mSizes;
	private Hashtable mClassData;

	public JarClassLoader (String jarfile) throws IOException
	{
		mJarFileName = jarfile;
		mSizes = new Hashtable();
		mClassData = new Hashtable();
		
		init();
	}
	
	private void init () throws IOException
	{
		// initialize the mSizes table
		JarFile jar_file = new JarFile( mJarFileName );
		Enumeration entries = jar_file.entries();
		JarEntry entry;
		while ( entries.hasMoreElements() )
		{
			entry = (JarEntry) entries.nextElement();
			mSizes.put( entry.getName(), new Integer( (int)entry.getSize() ) );
		}
		jar_file.close();		
		
		// initialize the mClassData table
		FileInputStream fin = null;
		JarInputStream jin = null;
		try
		{
			fin = new FileInputStream( mJarFileName );
			jin = new JarInputStream( fin );
			
			while ( (entry = jin.getNextJarEntry()) != null )
			{
				if ( entry.isDirectory() ) continue;
				
				// figure the size of the buffer
				int size = (int) entry.getSize();
				if ( size == -1 ) size = ((Integer)mSizes.get(entry.getName())).intValue();
				
				// create and fill the buffer
				byte buf[] = new byte[ size ];
				int count = 0;
				while ( count < size ) count += jin.read( buf, count, size-count );
				
				// store the buffer in the table
				mClassData.put( entry.getName(), buf );
			}
		}
		finally 
		{
			try { 
				fin.close(); 
			}
			catch (Exception e) {
			// ignore
			}
			
			try { 
				jin.close(); 
			}
			catch (Exception e) {
			// ignore
			}
		}
	}

	// Protected 
	protected byte[] loadClassBytes (String class_name) 
	{
		String entry_name = formatClassName(class_name);
		byte result[] = (byte[]) mClassData.get( entry_name );
		return result;
	}
} 
