package jmel.script;

import java.io.*;
import java.util.*;

public class MelMatrix extends MelArrayFloat
{
	public float[][] a;
	public int rows, cols;
	
	public MelMatrix() {
		this(new float[][] { {} });
	}
	public MelMatrix(float[][] a) {
		super(a);
		this.a = a;
		if (a != null)
		rows = a.length;
		cols = a[0].length;
	}
	
	public void setElement(int i, int j, Float o) { a[i][j] = o.floatValue(); }
	public void setElement(int i, int j, float o) { a[i][j] = o; }

	public float length() {
		if (rows != 1) return 0F;
		float l = 0;
		for (int i = 0; i < cols; i++) l += a[0][i]*a[0][i];
		return (float)Math.sqrt(l);
	}
		
	public static MelMatrix fromString(String s) {
		float a[][] = new float[][] { {} };
		try {
			Reader r = new BufferedReader(new StringReader(s));
			StreamTokenizer st = new StreamTokenizer(r);
			st.resetSyntax();
			st.wordChars(' ', 255);
	        st.wordChars('<', '<');
	        st.wordChars('>', '>');
			st.whitespaceChars(0, ' ');
			st.whitespaceChars('\"', '\"');
			st.whitespaceChars(',', ',');
			int i = 0, j = 0;
			Vector rows = new Vector();
			Vector cols = new Vector();
			while (st.nextToken() != st.TT_EOF) {
	        	if ("<<".equals(st.sval)) continue;
	        	if (">>".equals(st.sval)) break;
				if (st.ttype == ';') {
					rows.addElement(cols);
					cols = new Vector();
				}
				else
					cols.addElement(new Float((new Double(st.sval)).floatValue()));
			}
			r.close();
			int ro = rows.size(), co = cols.size();
			i = j = 0;
			a = new float[ro][co];
			for (Enumeration e = rows.elements(); e.hasMoreElements(); i++) {
				cols = (Vector)e.nextElement();
				for (Enumeration e2 = cols.elements(); e2.hasMoreElements(); j++) {
					a[i][j] = ((Float)e2.nextElement()).floatValue();
				}
			}
	    } catch (NumberFormatException e) {
	    } catch (IOException e) {}
		return new MelMatrix(a);
	}

	public static MelMatrix fromString(int rows, int cols, String s) {
		float a[][] = new float[][] { {} };
		try {
			Reader r = new BufferedReader(new StringReader(s));
			StreamTokenizer st = new StreamTokenizer(r);
			st.resetSyntax();
			st.wordChars(' ', 255);
			st.whitespaceChars(0, ' ');
			st.whitespaceChars('\"', '\"');
			st.whitespaceChars('<', '<');
			st.whitespaceChars('>', '>');
			int i = 0, j = 0;
			a = new float[rows][cols];
			while (st.nextToken() != st.TT_EOF) {
				a[i][j] = (float)st.nval;
				j++;
				if (j >= cols) {
					i++;
					j = 0;
				}
			}
			r.close();
	    } catch (IOException e) {}
		return new MelMatrix(a);
	}
	
	public String toString() {
		if (a == null) return "";
		String s = "<<";
		for (int i = 0; i < rows; i++) {
			for (int j = 0; j < cols; j++) {
				s += a[i][j];
				if (j < cols-1) s += ", ";
			}
			if (i < rows-1) s += "; ";
		}
		return s + ">>";
	}
	
	public MelMatrix neg() {
		float tmp[][] = new float[rows][cols];
		for (int i = 0; i < rows; i++)
			for (int j = 0; j < cols; j++) tmp[i][j] = -a[i][j];
		return new MelMatrix(tmp);
	}
	
	public MelMatrix add(MelMatrix m) {
		if (m.rows != rows || m.cols != cols) return this;
		float tmp[][] = new float[rows][cols];
		for (int i = 0; i < rows; i++)
			for (int j = 0; j < cols; j++) tmp[i][j] = a[i][j]+m.a[i][j];
		return new MelMatrix(tmp);
	}
	
	public MelMatrix sub(MelMatrix m) {
		if (m.rows != rows || m.cols != cols) return this;
		float tmp[][] = new float[rows][cols];
		for (int i = 0; i < rows; i++)
			for (int j = 0; j < cols; j++) tmp[i][j] = a[i][j]-m.a[i][j];
		return new MelMatrix(tmp);
	}
	
	public MelMatrix mul(float f) {
		float tmp[][] = new float[rows][cols];
		for (int i = 0; i < rows; i++)
			for (int j = 0; j < cols; j++) tmp[i][j] = a[i][j]*f;
		return new MelMatrix(tmp);
	}
	
	public MelMatrix div(float f) {
		float tmp[][] = new float[rows][cols];
		for (int i = 0; i < rows; i++)
			for (int j = 0; j < cols; j++) tmp[i][j] = a[i][j]/f;
		return new MelMatrix(tmp);
	}
	
	public MelMatrix mod(float f) {
		float tmp[][] = new float[rows][cols];
		for (int i = 0; i < rows; i++)
			for (int j = 0; j < cols; j++) tmp[i][j] = a[i][j]%f;
		return new MelMatrix(tmp);
	}
	
	public MelMatrix mul(MelMatrix m) {
		if (m.rows != cols) return this;
		float tmp[][] = new float[rows][m.cols], s;
		for (int i = 0; i < rows; i++) {
			for (int j = 0; j < m.cols; j++) {
				s = 0;
				for (int k = 0; k < cols; k++) s += a[i][k]*m.a[k][j];
				tmp[i][j] = s;
			}
		}
		return new MelMatrix(tmp);
	}
	
	public MelMatrix div(MelMatrix m) {
		if (m.rows != rows || m.cols != cols) return this;
		float tmp[][] = new float[rows][cols];
		for (int i = 0; i < rows; i++)
			for (int j = 0; j < cols; j++) tmp[i][j] = a[i][j]/m.a[i][j];
		return new MelMatrix(tmp);
	}
	
	public MelMatrix mod(MelMatrix m) {
		if (m.rows != rows || m.cols != cols) return this;
		float tmp[][] = new float[rows][cols];
		for (int i = 0; i < rows; i++)
			for (int j = 0; j < cols; j++) tmp[i][j] = a[i][j]%m.a[i][j];
		return new MelMatrix(tmp);
	}
	
	public MelVector mul(MelVector v) {
		if (rows != 3 || cols != 3) return v;
		float x = 0, y = 0, z = 0;
		for (int i = 0; i < rows; i++) {
			x = a[0][0]*v.x+a[0][1]*v.y+a[0][2]*v.z;
			y = a[1][0]*v.x+a[1][1]*v.y+a[1][2]*v.z;
			z = a[2][0]*v.x+a[2][1]*v.y+a[2][2]*v.z;
		}
		return new MelVector(x, y, z);
	}
}