package jmel.script;

import java.io.*;

public class MelVector
{
	public float x, y, z;
	
	public MelVector() {
		this(0, 0, 0);
	}
	public MelVector(float x) {
		this(x, 0, 0);
	}
	public MelVector(float x, float y) {
		this(x, y, 0);
	}
	public MelVector(float x, float y, float z) {
		this.x = x;
		this.y = y;
		this.z = z;
	}
	public MelVector(Float x, Float y, Float z) {
		this.x = x.floatValue();
		this.y = y.floatValue();
		this.z = z.floatValue();
	}
	public MelVector(Object x) {
		if (!(x instanceof Number))
			this.x = this.y = this.z = 0;
		else {
			this.x = ((Number)x).floatValue();
		}
	}
	public MelVector(Object x, Object y) {
		if (!(x instanceof Number) ||
			!(y instanceof Number))
			this.x = this.y = this.z = 0;
		else {
			this.x = ((Number)x).floatValue();
			this.y = ((Number)y).floatValue();
		}
	}
	public MelVector(Object x, Object y, Object z) {
		if (!(x instanceof Number) ||
			!(y instanceof Number) ||
			!(z instanceof Number))
			this.x = this.y = this.z = 0;
		else {
			this.x = ((Number)x).floatValue();
			this.y = ((Number)y).floatValue();
			this.z = ((Number)z).floatValue();
		}
	}
	
	public float length() {
		return (float)Math.sqrt(x*x+y*y+z*z);
	}
	
	public String toString() {
		return "<<" + x + " " + y + " " + z + ">>";
	}

	public static MelVector fromString(String s) {
        float x = 0, y = 0, z = 0;
		try {
	        Reader r = new BufferedReader(new StringReader(s));
	        StreamTokenizer st = new StreamTokenizer(r);
	        st.resetSyntax();
	        st.wordChars(' ', 255);
	        st.whitespaceChars(0, ' ');
	        st.whitespaceChars('\"', '\"');
	        st.whitespaceChars(',', ',');
	        st.whitespaceChars('<', '<');
	        st.whitespaceChars('>', '>');
	        int i = 0;
	        while (st.nextToken() != st.TT_EOF) {
	        	if (i == 0) x = (new Double(st.sval)).floatValue();
	        	else if (i == 1) y = (new Double(st.sval)).floatValue();
	        	else if (i == 2) {
	        		z = (new Double(st.sval)).floatValue();
	        		break;
	        	}
	            i++;
	        }
	        r.close();
	    } catch (NumberFormatException e) {
	    } catch (IOException e) {}
		return new MelVector(x, y, z);
	}
			
	public MelVector neg() {
		return new MelVector(-x, -y, -z);
	}
	
	public MelVector add(MelVector v) {
		return new MelVector(x+v.x, y+v.y, z+v.z);
	}
	
	public MelVector add(float f) {
		return new MelVector(x+f, y+f, z+f);
	}
	
	public MelVector sub(MelVector v) {
		return new MelVector(x-v.x, y-v.y, z-v.z);
	}
	
	public MelVector sub(float f) {
		return new MelVector(x-f, y-f, z-f);
	}
	
	public MelVector subinv(float f) {
		return new MelVector(f-x, f-y, f-z);
	}
	
	public float mul(MelVector v) {
		return x*v.x+y*v.y+z*v.z;
	}
	
	public MelVector mul(float f) {
		return new MelVector(x*f, y*f, z*f);
	}
	
	public MelVector div(MelVector v) {
		return new MelVector(x/v.x, y/v.y, z/v.z);
	}

	public MelVector div(float f) {
		return new MelVector(x/f, y/f, z/f);
	}	

	public MelVector invdiv(float f) {
		return new MelVector(f/x, f/y, f/z);
	}	
	
	public MelVector mod(float f) {
		return new MelVector(x%f, y%f, z%f);
	}
	
	public MelVector mod(MelVector v) {
		return new MelVector(x%v.x, y%v.y, z%v.z);
	}
	
	public MelVector invmod(float f) {
		return new MelVector(f%x, f%y, f%z);
	}

	public MelVector pow(MelVector v) {
		return new MelVector(y*v.z-z*v.y, z*v.x-x*v.z, x*v.y-y*v.x);
	}

	public boolean equals(Object o) {
		if (o instanceof MelVector) return equals((MelVector)o);
		return false;
	}

	public boolean equals(MelVector v) {
		return x==v.x && y==v.y && z==v.z;
	}
}
