package jmel.script;

import java.util.Hashtable;

public abstract class MyClassLoader extends ClassLoader 
{
	private Hashtable mClasses;

	public MyClassLoader () 
	{
		mClasses = new Hashtable();
	}
	

	public Class loadClass (String className) throws ClassNotFoundException 
	{
	    return (loadClass(className, true));
	}

	public synchronized Class loadClass (String class_name, boolean resolve) throws ClassNotFoundException 
	{
		Class   result;
		byte[]  class_bytes;
		
		// Check if we've loaded the class already
		result = (Class)mClasses.get(class_name);
		if (result != null) return result;
		
		// pass the buck to the system class loader
		try {
			return super.findSystemClass(class_name);
		} 
		catch (ClassNotFoundException e) {}
		
		// security measure: only system can load Java API classes
		if ( class_name.startsWith( "java." ) )
			throw new ClassNotFoundException( class_name );
		
		
		// load the class bytes from the file
		class_bytes = loadClassBytes(class_name);
		if (class_bytes == null) 
			throw new ClassNotFoundException( class_name );
		
		// have ClassLoader define this class (parse the class file)
		result = defineClass(class_name, class_bytes, 0, class_bytes.length);
		if (result == null) 
			throw new ClassFormatError();
		
		// resolve the class, if requested
		if (resolve) resolveClass(result);
		
		// save this class in our cache in case requested again
		mClasses.put(class_name, result);
		return result;
	}
	
	// Protected
	protected abstract byte[] loadClassBytes (String class_name);
	
	protected String formatClassName (String class_name) 
	{
		// replace with the default separator character for Jar files
		return class_name.replace('.',System.getProperty("file.separator").charAt(0)) + ".class";
	}

}
