package jmel.script;

import java.lang.reflect.*;
import java.io.*;
import java.util.*;
import jmel.util.*;
import jmel.tree.Tree;
import jmel.code.Type;

public class ObjectCommands extends Commands
{
	private CommandObjectInvoker sc;
	private Vector actions = new Vector();
	private int cnt = 0;
	private Object obj[];
	
	// Public
	public static ObjectCommands get(Object obj) {
		String name = (String)obj;
		CommandObjectInvoker sc = (CommandObjectInvoker)cmds.get(name);
		if (sc == null) {
			sc = new CommandObjectInvoker(getClass(name));
			cmds.put(name, sc);
		}
		return new ObjectCommands(sc);
	}
	
	public void exec(Object[] obj) {
		this.obj = obj;
		process(false);
	}
	
	public void exec(Object obj) {
		this.obj = new Object[] { obj };
		process(false);
	}
	
	public Object execAsExpr(Object[] obj) {
		this.obj = obj;
		return process(true);
	}
	
	public Object execAsExpr(Object obj) {
		this.obj = new Object[] { obj };
		return process(true);
	}
	
	public CommandObjectInvoker getInvoker() { return sc; }
	
	// Private
	private ObjectCommands(CommandObjectInvoker sc) { this.sc = sc; }

	private Object process(boolean expr) {

		Vector cmdArgs = new Vector();
		while (cnt < obj.length) {
			
			String objcnt = null;
			if (obj[cnt] instanceof String)
				objcnt = (String)obj[cnt];
				
			if (objcnt != null && objcnt.charAt(0) == '-' && (objcnt.charAt(1) >= 'A' && objcnt.charAt(1) <= 'z')) {
				String optName = objcnt; cnt++;
				if (optName.charAt(1) == '"')
					throw new InternalError("Syntax error: unexpected '\"' after '-'.");
				
				Object[] optArgs = null;
				Object optArg = null;
				
				while (cnt < obj.length) {
					optArg = next();
					if (optArg == null) break;
					optArgs = append(optArgs, optArg);
				}
				
				cmdArgs.add(optName);
				cmdArgs.add(optArgs);
			} else {
				Object[] args = null;
				Object optArg = null;
				while (cnt < obj.length) {
					optArg = next();
					if (optArg == null) 
						throw new InternalError("Syntax error: misplaced option.");
					args = append(args, optArg);
				}

				cmdArgs.add(ARGS);
				cmdArgs.add(args);
			}				
		}
		
		try {
			actions = sc.process(cmdArgs);
			return sc.execAsExpr(actions);
		} catch (Throwable t) {
			if(Boolean.getBoolean("debug"))
				t.printStackTrace();
			if (withCatch) catchThrowed = true;
			else if (t instanceof java.lang.reflect.InvocationTargetException)
				throw new InternalError(((java.lang.reflect.InvocationTargetException)t).getTargetException().toString());			
			else
				throw new InternalError(t.toString());			
		}
	
		return null;
	}

	private Object next() {

		Object objcnt = obj[cnt];
		String arg = null;
		if (objcnt instanceof String)
			arg = (String)objcnt;
		if (arg != null && arg.length() > 0 && arg.charAt(0) == '-' && (arg.charAt(1) >= 'A' && arg.charAt(1) <= 'z'))
			return null;
		cnt++;
		return objcnt;
	}


	// Static
	static Object[] append(Object[] s, Object a) {
		int l;
		Object tmp[];
		if (s != null) {
			l = s.length;
			tmp = new Object[l+1];
			System.arraycopy(s, 0, tmp, 0, l);
		} else {
			l = 0;
			tmp = new Object[1];
		}
		tmp[l] = a;
		return tmp;
	}
}
