package jmel.script;

import java.util.StringTokenizer;
import java.io.IOException;

/**
 * A class loader capable of loading from local files and jar archives. 
 *
 */
public class PathClassLoader extends MyClassLoader 
{
	private MyClassLoader mLoaders[];
	private int mCount;

	public PathClassLoader () 
	{
		this( System.getProperty("java.class.path"), System.getProperty("path.separator").charAt(0) );
	}
	public PathClassLoader (String path) 
	{
		this( path, System.getProperty("path.separator").charAt(0) );
	}
	public PathClassLoader (String path, char separator) {
		StringTokenizer st = new StringTokenizer( path, String.valueOf(separator) );
		mLoaders = new MyClassLoader[ st.countTokens() ];
		String entry;
		int i = 0;
		while ( st.hasMoreTokens() )
		{
			entry = st.nextToken();
			if ( entry.endsWith( ".jar" ) )
			{
				try
				{
				mLoaders[i] = new JarClassLoader( entry );
				++i;
			}
			catch (IOException e)
			{}
			}
			else
			{
				mLoaders[i]= new FileClassLoader( entry );
				++i;
			}
		}
		
		mCount = i;
	}


	//---------- Protected Methods ---------------------------
	protected byte[] loadClassBytes (String class_name) {
		byte result[] = null;
		for (int i=0; i < mCount; ++i)
		{
			result = mLoaders[i].loadClassBytes( class_name );
			if ( result != null) return result;
		}
		
		// if no class file is found
		return null;
	}
}
