package jmel.script;

import jmel.*;
import jmel.interp.*;
import jmel.tree.Tree;
import jmel.tree.Tree.TopLevel;
import java.lang.reflect.Method;

public class ScriptCaller
{
	public static boolean insideScript = false;
	private static int classId = 0;
	private String script;
	private Method method;
	private JMelScriptMethod scriptMethod;
	
	// Public
	public ScriptCaller(String script) { 
		if (!insideScript)
			try {
				if (!script.endsWith(";")) script += ";";
				jmel.util.Hashtable options = new jmel.util.Hashtable();
				options.put("-O2", "");
				JMelCompiler comp = JMelCompiler.make(options);
				String className = "_script" + classId++;
				TopLevel topLevel = comp.parse(className, script);
				if (comp.errorCount() != 0)
					throw new InternalError("Syntax error.");
				comp.generate(topLevel);
	
				Class cl = Commands.getClass(className);
				Method m = cl.getMethod("exec", null);
				setMethod(m);
			} catch (Throwable t) {
				throw new InternalError(t.toString());
			}
		this.script = script;
	}
	
	public void invoke() {
		invoke(null);
	}
	
	public String toString() {
		return script == null ? super.toString() : script;
	}
	
   	public static NameSpace namespace;
   	public static JMelInterpreter interpreter;
   	public static CallStack callstack;
   	public static Tree callerInfo;

	// Proctected
	protected ScriptCaller() {}
	
	protected void setMethod(Method method) { this.method = method; }
	protected void setMethod(JMelScriptMethod scriptMethod) { this.scriptMethod = scriptMethod; }
	
	protected Object invoke(Object[] args) {
		try {
			if (method != null) return method.invoke(null, args);
			if (scriptMethod != null) {
				Object[] tmp;
				if (args == null) tmp = null;
				else {
					tmp = new Object[args.length];
					for (int i = 0; i < args.length; i++)
						if (args[i] instanceof Primitive) tmp[i] = args[i];
						else tmp[i] = new Primitive(args[i]);
				}
				return scriptMethod.invokeDeclaredMethod( args, interpreter, callstack, callerInfo );
			}
			return interpreter.eval(script);
 		} catch (Throwable t) {
 			String name;
 			if (method != null) name = method.getName();
 			else if (scriptMethod != null) name = scriptMethod.toString();
 			else name = script;
			throw new InternalError("Error while invoking callback '" + name + "': " + t);
		}
	}
}