package jmel.script;

import java.lang.reflect.*;
import java.io.*;
import java.util.*;
import jmel.util.*;
import jmel.tree.Tree;
import jmel.code.Type;

public class TreeCommands extends Commands
{
	private CommandTreeInvoker sc;
	private Vector actions = new Vector();
	private int cnt = 0;
	private jmel.util.List tree;
	
	// Public
	public static TreeCommands get(Tree.Literal tree) {
		String name = (String)tree.value;
		CommandTreeInvoker sc = (CommandTreeInvoker)cmds.get(name);
		if (sc == null) {
			sc = new CommandTreeInvoker(getClass(name));
			cmds.put(name, sc);
		}
		return new TreeCommands(sc);
	}
	
	public Tree exec(int pos, jmel.util.List trees, int methCnt) {
		this.tree = trees;
		return process(pos, false, methCnt);
	}
	
	public Tree execAsExpr(int pos, jmel.util.List trees, int methCnt) {
		this.tree = trees;
		return process(pos, true, methCnt);
	}
	
	// Private
	private TreeCommands(CommandTreeInvoker sc) { this.sc = sc; }

	private Tree process(int pos, boolean expr, int methCnt) {

		Vector cmdArgTrees = new Vector();
		Tree curtree;

		while (true) {
			if (tree.isEmpty()) break;
			
			curtree = (Tree)tree.head;
			String argcnt = null;
			if (curtree instanceof Tree.Literal && ((Tree.Literal)curtree).typetag == Type.CLASS)
				argcnt = (String)((Tree.Literal)curtree).value;
			
			if (argcnt != null && argcnt.charAt(0) == '-' && (argcnt.charAt(1) >= 'A' && argcnt.charAt(1) <= 'z')) {
				String optName = argcnt;
				tree = tree.tail;
				if (optName.charAt(1) == '"')
					throw new InternalError("Syntax error: unexpected '\"' after '-'.");
				
				jmel.util.ListBuffer optArgs = new jmel.util.ListBuffer();
				Tree optArg = null;
				
				for (; tree.nonEmpty(); tree = tree.tail) {
					optArg = next();
					if (optArg == null) break;
					optArgs.append(optArg);
				}
				
				cmdArgTrees.add(optName);
				cmdArgTrees.add(optArgs.toList());
			} else {
				jmel.util.ListBuffer args = new jmel.util.ListBuffer();
				Tree optArg = null;
				for (; tree.nonEmpty(); tree = tree.tail) {
					optArg = next();
					if (optArg == null) 
						throw new InternalError("Syntax error: misplaced option.");
					args.append(optArg);
				}

				cmdArgTrees.add(ARGS);
				cmdArgTrees.add(args.toList());
			}				
		}
		
		Vector actions = sc.process(cmdArgTrees, expr);
		return sc.exec(pos, actions, expr, methCnt);
	}
	
	private Tree next() {

		Tree argcnt = (Tree)tree.head;
		String arg = null;
		if (argcnt instanceof Tree.Literal && ((Tree.Literal)argcnt).typetag == Type.CLASS)
			arg = (String)((Tree.Literal)argcnt).value;
		if (arg != null && arg.length() > 0 && arg.charAt(0) == '-' && (arg.charAt(1) >= 'A' && arg.charAt(1) <= 'z'))
			return null;
		return argcnt;
	}
}
