package jmel.tree;

import jmel.code.*;
import jmel.util.*;
import jmel.tree.Tree;
import jmel.tree.Tree.*;

/**
 * Return the Java type.
 */
public class GlobalVisitor extends Visitor implements Flags {
	/**
	* Visitor method: visits expression tree.
	*/
	public void visitExpr(Tree tree) throws Error {
		if (tree != null) tree.visit(this);
	}
	
	/**
	* Derived visitor method: visit statement tree.
	*/
	public void visitStat(Tree tree) throws Error {
		visitExpr(tree);
	}
	
	/**
	* Derived visitor method: visit list of expression trees, separated by commas.
	*/
	public void visitExprs(List trees) throws Error {
		for (List l = trees; l.nonEmpty(); l = l.tail) {
			visitExpr((Tree) l.head);
		}
	}
	
	/**
	* Derived visitor method: visit list of statements, each on a separate line.
	*/
	public void visitStats(List trees) throws Error {
		visitExprs(trees);
	}

	/**
	* visit a block.
	*/
	public void visitBlock(List stats) throws Error {
		visitStats(stats);
	}
}
