package jmel.tree;

import java.util.TreeMap;
import java.util.Iterator;
import java.util.Vector;
import java.util.Enumeration;
import java.util.HashMap;
import jmel.util.*;
import jmel.code.*;
import jmel.tree.*;
import jmel.tree.Tree.*;
import jmel.script.*;

/**
 * Translate a Mel Script to a source program.
 */
public class MelTranslator extends GlobalVisitor {
	
	private static String className = "_script_";
	private TreeMaker make = new TreeMaker();
	
	private TreeMap procs = new TreeMap();
	private TreeMap globalprocs = new TreeMap();
	
	private static final String _core = "<core>";
	private static final String Cmds_ = "__Cmds";
	private static final String Globals_ = "__Globals";
	private static final String GlobalDefs_ = "__GlobalDefs";
	private HashMap procVals = new HashMap();
	
	private Tree _tree_;
	private boolean fullOpt = false;
	public ListBuffer toplevels = new ListBuffer();
	public ListBuffer extras = new ListBuffer();

	public MelTranslator(boolean fullOpt) {
		this.fullOpt = fullOpt;
	}
	
	public void visitExpr(Tree tree) throws Error {
		_tree_ = null;
		if (tree != null) tree.visit(this);
	}
	
	private class Namespace
	{
		public Namespace(Namespace parent) { this.parent = parent; }
		
		Namespace parent;
		private TreeMap vars = new TreeMap();
		private TreeMap aliases = new TreeMap();
		
		VarDef getVar(String name) {
			VarDef v = (VarDef)vars.get(name);
			if (v == null && parent != null) return parent.getVar(name);
			return v;
		}
		
		void putVar(String name, VarDef vdef) {
			vars.put(name, vdef);
		}
		

		public Object getAlias( String name ) {
			Object args = aliases.get(name);
			if (args == null && parent != null) return parent.getAlias(name);
			return args;
		}
		
		public void	setAlias(String	name, Tree cmd) 
		{
			aliases.put(name, cmd);
		}
		
		public void	setAlias(String	name, List args) 
		{
			aliases.put(name, args);
		}
	}
	
	private Namespace currentNamespace = new Namespace(null);;
	
	/**
	* Visitor methods
	*/
	public void _case(TopLevel tree) {
		// List of commands called from main 
		procVals.put(_core + Cmds_, new ListBuffer());
		// List of global vars used in main 
		procVals.put(_core + Globals_, new ListBuffer());
		// List of global var defined in main 
		procVals.put(_core + GlobalDefs_, new TreeMap());
		
		String dir = "";
		
		className = tree.sourcefile.toString();
		int idx = className.lastIndexOf(System.getProperty("file.separator"))+1;
		dir = className.substring(0, idx);
		className = className.substring(idx);
		idx = className.indexOf(".");
		if (idx != -1) className = className.substring(0, idx);
		
		ListBuffer l1 = new ListBuffer();
		ListBuffer l2 = new ListBuffer();

		for (List l = tree.defs; l.nonEmpty(); l = l.tail) {
			visitExpr((Tree) l.head);
			if (_tree_ != null)
				if (((Tree) l.head).tag == Tree.METHODDEF)
					l1.append(_tree_);
				else
					l2.append(_tree_);
		}

		TopLevel topLevel;
		
		Iterator i = globalprocs.values().iterator();
		String name = "";
		boolean found = false;
		
		while (i.hasNext()) {
			Vector v = (Vector) i.next();
			ListBuffer l4 = new ListBuffer();
			
			name = "";
			
			for (Enumeration e = v.elements(); e.hasMoreElements(); ) {
				MethodDef md = (MethodDef) e.nextElement();
				name = md.name.toString();
				md.name = Name.fromString("exec");
				
				l4.append(md);
			}
	
			l4.appendList(extras.toList());
			
			ListBuffer lb;
			lb = (ListBuffer)procVals.get(name + Globals_);
			if (lb != null) l4.appendList(lb.toList());								
			
			lb = (ListBuffer)procVals.get(name + Cmds_);
			if (lb != null) l4.appendList(lb.toList());								

			if (name.equals(className)) {
				found = true;
			}

			ClassDef classdef = make.at(tree.pos).ClassDef(
										Flags.PUBLIC | Flags.FINAL, 
										Name.fromString(name), 
										Tree.emptyList, 
										null,
										Tree.emptyList.append(toName(tree.pos, "Script")), 
										l4.toList());


			l4 = new ListBuffer();
			l4.append(make.at(tree.pos).Import(toName(tree.pos, "jmel.script.*")));
			l4.append(classdef);
			
			topLevel = make.at(tree.pos).TopLevel(null, l4.toList());
			topLevel.sourcefile = Name.fromString(dir + name + ".mel");
			append(topLevel);
		}
		
		if (!found) {
	
			// Create exec method
			Block b2 = make.at(tree.pos).Block( 0, l2.toList());
			
			MethodDef execDef = make.at(tree.pos).MethodDef(
										Flags.PUBLIC | Flags.STATIC, 
										Name.fromString("exec"), 
										make.at(tree.pos).TypeIdent( Type.VOID ), 
										Tree.emptyList, 
										Tree.emptyList, 
										Tree.emptyList, 
										b2);	
			l1.append(execDef);		
			l1.appendList(extras.toList());
	
			ListBuffer lb;
			lb = (ListBuffer)procVals.get(_core + Globals_);
			if (lb != null) l1.appendList(lb.toList());								
			
			lb = (ListBuffer)procVals.get(_core + Cmds_);
			if (lb != null) l1.appendList(lb.toList());								
			
			ClassDef classdef = make.at(tree.pos).ClassDef(
										Flags.PUBLIC | Flags.FINAL, 
										Name.fromString(className), 
										Tree.emptyList, 
										null,
										Tree.emptyList.append(toName(tree.pos, "Script")), 
										l1.toList());
	
			ListBuffer l5 = new ListBuffer();
			l5.append(make.at(tree.pos).Import(toName(tree.pos, "jmel.script.*")));
			l5.append(classdef);
			
			topLevel = make.at(tree.pos).TopLevel(null, l5.toList());
			topLevel.sourcefile = tree.sourcefile;
			append(topLevel);
		}
	}

	void append(TopLevel tree) {
		if (Boolean.getBoolean("debug")) {
			tree.visit(new Pretty());		
			tree.visit(new Core());		
		}
		toplevels.append(tree);
	}
	
	private boolean insideProc = false;
	private String insideProcName = null;
	
	// If needed, build a second method declaration with Java arrays as arguments
	private MethodDef checkMethod(MethodDef tree) {
		/*
			initial method:
				method (MelArrayInt arg1, MelArrayFloat arg2, ...)
				
			generated method:
				m1:
				method (int[] arg1, float[] arg2, ...) {
					_method (new MelArrayInt(arg1), new MelArrayFloat(arg2), ...)
				}
			
				m2:
				_method (MelArrayInt arg1, MelArrayFloat arg2, ...)
		*/
		ListBuffer params = new ListBuffer();
		ListBuffer params2 = new ListBuffer();
		boolean found = false;
		
		Tree tree1 = null, tree2 = null;
		for (List l = tree.params; l.nonEmpty(); l = l.tail) {
			VarDef vdef = (VarDef) l.head;
			if (vdef.vartype instanceof Ident) {
				Ident ident = (Ident) vdef.vartype;
				String name = ident.name.toString();
				if (name.startsWith("MelArray")) {
					if (name.equals("MelArrayInt")) {
						tree1 = make.at(ident.pos).TypeArray(make.at(ident.pos).TypeIdent(Type.INT), 0);
					} else if (name.equals("MelArrayFloat")) {
						tree1 = make.at(ident.pos).TypeArray(make.at(ident.pos).TypeIdent(Type.FLOAT), 0);
					} else if (name.equals("MelArrayDouble")) {
						tree1 = make.at(ident.pos).TypeArray(make.at(ident.pos).TypeIdent(Type.DOUBLE), 0);
					} else if (name.equals("MelArrayVector")) {
						tree1 = make.at(ident.pos).TypeArray(toName(ident.pos, "MelVector"), 0);
					} else if (name.equals("MelArrayMatrix")) {
						tree1 = make.at(ident.pos).TypeArray(toName(ident.pos, "MelMatrix"), 0);
					} else if (name.equals("MelArrayString")) {
						tree1 = make.at(ident.pos).TypeArray(toName(ident.pos, "String"), 0);
					}
					tree1 = make.at(ident.pos).VarDef(0, vdef.name, tree1, null);
					tree2 = make.at(ident.pos).NewClass( null, toName(ident.pos, name), Tree.emptyList.append( make.at(ident.pos).Ident(vdef.name)), null );
					
					if (!found) found = true;
				} else {
					tree1 = make.at(ident.pos).VarDef(0, vdef.name, vdef.vartype, null);
					tree2 = make.at(ident.pos).Ident(vdef.name);
				}
			} else {
				tree1 = make.at(vdef.pos).VarDef(0, vdef.name, vdef.vartype, null);
				tree2 = make.at(vdef.pos).Ident(vdef.name);
			}
			
			params.append(tree1);
			params2.append(tree2);
		}
		
		if (found) {
			// Do we have to convert the return value?
			Tree restype = tree.restype;
			
			found = false;
			
			if (tree.restype instanceof Ident) {
				Ident ident = (Ident) tree.restype;
				String name = ident.name.toString();
				if (name.startsWith("MelArray")) {
					if (name.equals("MelArrayInt")) {
						restype = make.at(ident.pos).TypeArray(make.at(ident.pos).TypeIdent(Type.INT), 0);
					} else if (name.equals("MelArrayFloat")) {
						restype = make.at(ident.pos).TypeArray(make.at(ident.pos).TypeIdent(Type.FLOAT), 0);
					} else if (name.equals("MelArrayDouble")) {
						restype = make.at(ident.pos).TypeArray(make.at(ident.pos).TypeIdent(Type.DOUBLE), 0);
					} else if (name.equals("MelArrayVector")) {
						restype = make.at(ident.pos).TypeArray(toName(ident.pos, "MelVector"), 0);
					} else if (name.equals("MelArrayMatrix")) {
						restype = make.at(ident.pos).TypeArray(toName(ident.pos, "MelMatrix"), 0);
					} else if (name.equals("MelArrayString")) {
						restype = make.at(ident.pos).TypeArray(toName(ident.pos, "String"), 0);
					}
					
					found = true;
				} 
			}

			Name nm = Name.fromString( "_"+tree.name.toString() );
			Apply apply =  make.at(tree.pos).Apply( make.at(tree.pos).Ident(nm), params2.toList());
			Tree t;
			if (tree.restype instanceof TypeIdent && ((TypeIdent) tree.restype).typetag == Type.VOID)
				t =  make.at(tree.pos).Exec( apply );
 			else {
 				if (found) {
 					// return <...>.get();
 					Select select = make.at(tree.pos).Select(apply, Name.fromString("get")); 		
 					apply = make.at(tree.pos).Apply( select, Tree.emptyList);
				} 
				t = make.at(tree.pos).Return(apply);
			}
			Block body = make.at(tree.pos).Block( 0, Tree.emptyList.append(t) );
			MethodDef m1 = make.at(tree.pos).MethodDef(tree.flags, tree.name, restype, Tree.emptyList, params.toList(), Tree.emptyList, body);
			
			MethodDef m2 = tree;
			m2.name = nm;
			m2.flags = tree.flags ^ Flags.PUBLIC;
			extras.append(m2);
			
			return m1;
		}
		
		return tree;
	}
	
	public void _case(MethodDef tree) {
		String name = tree.name.toString();

		insideProc = true;
		insideProcName = name;
		
		// List of commands called from proc 
		procVals.put(name + Cmds_, new ListBuffer());
		// List of global vars defined in proc 
		procVals.put(name + Globals_, new ListBuffer());
		// List of global var defined in proc 
		procVals.put(name + GlobalDefs_, new TreeMap());

		Vector v = (Vector)procs.get(name);
		if (v == null) procs.put(tree.name.toString(), v = new Vector());
		v.add(tree);
		ListBuffer params = new ListBuffer();
		for (List l = tree.params; l.nonEmpty(); l = l.tail) {
			visitExpr((Tree) l.head);
			((VarDef)_tree_).init = null;
			params.append(_tree_);
		}
		visitExpr(tree.body);
		Block body = (Block)_tree_;
		visitExpr(tree.restype);
		Tree restype = _tree_;
		int flags = tree.flags;
		flags |= Flags.STATIC;
		_tree_ = make.at(tree.pos).MethodDef(flags, tree.name, restype, Tree.emptyList, params.toList(), Tree.emptyList, body);
		_tree_ = checkMethod( (MethodDef) _tree_ );
		
		if ((flags & Flags.PUBLIC) != 0) {
			v = (Vector)globalprocs.get(tree.name.toString());
			if (v == null) globalprocs.put(tree.name.toString(), v = new Vector());
			v.add(_tree_);
			_tree_ = null;
		}
		
		insideProc = false;
		insideProcName = null;
	}

	public void _case(VarDef tree) {
		visitExpr(tree.vartype);
		Tree vartype = _tree_;
		Tree init = null;
		ListBuffer elems = new ListBuffer();
		
		if (tree.init != null) {
			visitExpr(tree.init);
			if (tree.vartype instanceof TypeIdent) {
				String type = "";
				switch (((TypeIdent)tree.vartype).typetag) {
					case Type.INT: type = "toInt"; break;
					case Type.FLOAT: type = "toFloat"; break;
					case Type.DOUBLE: type = "toDouble"; break;
					case Type.STRING: type = "toString"; break;
					case Type.VECTOR: type = "toVector"; break;
					case Type.MATRIX: type = "toMatrix"; break;
				}
				Select select = make.at(tree.init.pos).Select(toName(tree.init.pos, "ConvUtils"), Name.fromString(type));
				_tree_ = make.at(tree.init.pos).Apply(select, Tree.emptyList.append(_tree_));
			} else if (tree.vartype instanceof TypeArray) {
				TypeArray ta = (TypeArray)tree.vartype;
				Tree elemtype = ta.elemtype;
				String name = null;
				TypeIdent ti = null;
				if (elemtype.tag == Tree.TYPEIDENT) {
					ti = (TypeIdent)elemtype;
					switch (ti.typetag) {
						case Type.STRING:	name = "String"; ti = null; break;
						case Type.VECTOR: 	name = "MelVector"; ti = null; break;
						case Type.INT: 		name = "int"; break;
						case Type.FLOAT: 	name = "float"; break;
						case Type.DOUBLE: 	name = "double"; break;
					}
				}
				if (name == null)
					throw new InternalError("Unknown or illegal array type: " + elemtype.tag);
				if (_tree_ instanceof NewArray) {
					((NewArray)_tree_).elemtype = ti != null ? ti : toName(elemtype.pos, name);
				}	
			} else {
				TypeMatrix tm = (TypeMatrix)tree.vartype;
				ListBuffer args = new ListBuffer();
				args.append(_tree_);
				Select select = make.at(tree.init.pos).Select(toName(tree.init.pos, "ConvUtils"), Name.fromString("toMatrix"));
				_tree_ = make.at(tree.init.pos).Apply(select, args.toList());
			}
			init = _tree_;
		} else if (tree.vartype.tag == Tree.TYPEARRAY) {
			TypeArray ta = (TypeArray)tree.vartype;
			Tree elemtype = ta.elemtype;
			if (elemtype.tag == Tree.TYPEIDENT) {
				TypeIdent ti = (TypeIdent)elemtype;
				switch (ti.typetag) {
					case Type.STRING:
						for (int i = 0; i < ta.size; i++)
							elems.append(make.at(tree.pos).Literal(Type.CLASS, ""));
						init = make.at(tree.pos).NewArray(toName(tree.pos, "String"), Tree.emptyList, elems.toList());
						break;
							
					case Type.VECTOR:
						for (int i = 0; i < ta.size; i++) {
							ListBuffer elems2 = new ListBuffer();
							elems2.append(make.at(tree.pos).Literal(Type.INT, new Integer(0)));
							elems2.append(make.at(tree.pos).Literal(Type.INT, new Integer(0)));
							elems2.append(make.at(tree.pos).Literal(Type.INT, new Integer(0)));
							Tree array = make.at(tree.pos).NewArray(make.at(tree.pos).TypeIdent(Type.FLOAT), Tree.emptyList, elems2.toList());
							elems.append( make.at(tree.pos).NewClass(null, toName(tree.pos, "MelVector"), elems2.toList(), null) );
						}
						init = make.at(tree.pos).NewArray(toName(tree.pos, "MelVector"), Tree.emptyList, elems.toList());
						break;
									
					case Type.INT:
						for (int i = 0; i < ta.size; i++)
							elems.append(make.at(tree.pos).Literal(Type.INT, new Integer(0)));
						init = make.at(tree.pos).NewArray(make.at(tree.pos).TypeIdent(Type.INT), Tree.emptyList, elems.toList());
						break;
						
					case Type.FLOAT:
						for (int i = 0; i < ta.size; i++)
							elems.append(make.at(tree.pos).Literal(Type.FLOAT, new Float(0)));
						init = make.at(tree.pos).NewArray(make.at(tree.pos).TypeIdent(Type.FLOAT), Tree.emptyList, elems.toList());
						break;
						
						
					case Type.DOUBLE:
						for (int i = 0; i < ta.size; i++)
							elems.append(make.at(tree.pos).Literal(Type.DOUBLE, new Double(0)));
						init = make.at(tree.pos).NewArray(make.at(tree.pos).TypeIdent(Type.DOUBLE), Tree.emptyList, elems.toList());
						break;

					default:
						throw new InternalError("Unknown or illegal array type: " + ti.typetag);
				}
			} else
				throw new InternalError("Unknown or illegal array type: " + elemtype.tag);
		} else if (tree.vartype.tag == Tree.TYPEMATRIX) {
			TypeMatrix tm = (TypeMatrix)tree.vartype;
			for (int i = 0; i < tm.rows; i++) {
				ListBuffer elems2 = new ListBuffer();
				for (int j = 0; j < tm.cols; j++) 
					elems2.append(make.at(tree.pos).Literal(Type.FLOAT, new Float(0)));
				elems.append( make.at(tree.pos).NewArray(make.at(tree.pos).TypeIdent(Type.FLOAT), Tree.emptyList, elems2.toList()) );
			}
			TypeArray ta1 = make.at(tree.pos).TypeArray( make.at(tree.pos).TypeIdent(Type.FLOAT), 0 );
			TypeArray ta2 = make.at(tree.pos).TypeArray( ta1, 0 );
			ListBuffer args = new ListBuffer();
			args.append( make.at(tree.pos).NewArray(ta1, Tree.emptyList, elems.toList()) );
			init = make.at(tree.pos).NewClass(null, toName(tree.pos, "MelMatrix"), args.toList(), null);
		} else if (tree.vartype instanceof TypeIdent) {
			switch (((TypeIdent)tree.vartype).typetag) {
				case Type.INT: 
					init = make.at(tree.pos).Literal(Type.INT, new Integer(0)); 
					break;
				case Type.FLOAT: 
					init = make.at(tree.pos).Literal(Type.FLOAT, new Float(0)); 
					break;
				case Type.DOUBLE: 
					init = make.at(tree.pos).Literal(Type.DOUBLE, new Double(0)); 
					break;
				case Type.STRING: 
					init = make.at(tree.pos).Literal(Type.CLASS, ""); 
					break;
				case Type.VECTOR: 
					init = make.at(tree.pos).NewClass(null, toName(tree.pos, "MelVector"), Tree.emptyList, null);
					break;
			}
		} 
		
		if (tree.vartype instanceof TypeArray && init != null && !(init instanceof NewClass)) {
			elems = new ListBuffer();
			TypeArray ta = (TypeArray)tree.vartype;
			Tree elemtype = ta.elemtype;
			String type = "";
			
			if (elemtype.tag == Tree.TYPEIDENT) {
				TypeIdent ti = (TypeIdent)elemtype;
				switch (ti.typetag) {
					case Type.STRING:
						type = "MelArrayString";
						break;
							
					case Type.VECTOR:
						type = "MelArrayVector";
						break;
									
					case Type.INT:
						type = "MelArrayInt";
						break;
						
					case Type.FLOAT:
						type = "MelArrayFloat";
						break;			
						
					case Type.DOUBLE:
						type = "MelArrayDouble";
						break;

					default:
						throw new InternalError("Unknown or illegal array type: " + ti.typetag);
				}
			} else
				throw new InternalError("Unknown or illegal array type: " + elemtype.tag);
				
			elems.append(init);
			vartype = toName(tree.vartype.pos, type);
			init = make.at(tree.pos).NewClass(null, toName(tree.pos, type), elems.toList(), null);
		}

		if ((tree.flags & Flags.PUBLIC) != 0) {
			tree.flags = 0;
			
			TreeMap map;
			if (insideProc)
				map = (TreeMap)procVals.get(insideProcName + GlobalDefs_);
			else
				map = (TreeMap)procVals.get(_core + GlobalDefs_);
			boolean isDefined = map.get(tree.name.toString()) != null;
			if (!isDefined) {
				// Commands.defineGlobalVar(<name>, <init>);	
				map.put(tree.name.toString(), tree);
								
				ListBuffer lb;
				if (insideProc)
					lb = (ListBuffer)procVals.get(insideProcName + Globals_);
				else
					lb = (ListBuffer)procVals.get(_core + Globals_);
				//lb.append( make.at(tree.pos).VarDef(Flags.STATIC, tree.name, vartype, init) );

				Apply a1 = make.at(tree.pos).Apply(toName(tree.pos, "Commands.setGlobalVar"), List.make(make.at(tree.pos).Literal(Type.CLASS, tree.name.toString()), init));
				lb.append(make.at(tree.pos).Block(Flags.STATIC, Tree.emptyList.append(make.at(tree.pos).Exec(a1))));								
			} 
			if (isAssigned) {
				_tree_ = null;
				return;
			}
		} 
		
		_tree_ = make.at(tree.pos).VarDef(tree.flags, tree.name, vartype, init);

		if ((tree.flags & Flags.PUBLIC) == 0 && currentNamespace != null)
			currentNamespace.putVar(tree.name.toString(), (VarDef)_tree_); 
	}
	
	public void _case(Block tree) {
		currentNamespace = new Namespace(currentNamespace);
		ListBuffer stats = new ListBuffer();
		for (List l = tree.stats; l.nonEmpty(); l = l.tail) {
			visitExpr((Tree) l.head);
			if (_tree_ != null) stats.append(_tree_);
		}
		_tree_ = make.at(tree.pos).Block(0, stats.toList());
		currentNamespace = currentNamespace.parent;
	}
	
	public void _case(DoLoop tree) {
		visitStat(tree.body);
		Tree body = _tree_;
		visitExpr(tree.cond);
		Tree cond = _tree_;
		Select select = make.at(tree.pos).Select(toName(tree.pos, "ConvUtils"), Name.fromString("toBool"));
		cond = make.at(tree.pos).Apply(select, Tree.emptyList.append(cond));
		_tree_ = make.at(tree.pos).DoLoop(body, cond);
	}

	public void _case(WhileLoop tree) {
		visitStat(tree.body);
		Tree body = _tree_;
		visitExpr(tree.cond);
		Tree cond = _tree_;
		Select select = make.at(tree.pos).Select(toName(tree.pos, "ConvUtils"), Name.fromString("toBool"));
		cond = make.at(tree.pos).Apply(select, Tree.emptyList.append(_tree_));
		_tree_ = make.at(tree.pos).WhileLoop(cond, body);
	}
	
	public void _case(ForLoop tree) {
		ListBuffer init = new ListBuffer();
		for (List l = tree.init; l.nonEmpty(); l = l.tail) {
			visitExpr((Tree) l.head);
			init.append(_tree_);
		}
		Tree cond = null;
		if (tree.cond != null) {
			visitExpr(tree.cond);
			Select select = make.at(tree.pos).Select(toName(tree.pos, "ConvUtils"), Name.fromString("toBool"));
			cond = make.at(tree.pos).Apply(select, Tree.emptyList.append(_tree_));
		}
		ListBuffer step = new ListBuffer();
		for (List l = tree.step; l.nonEmpty(); l = l.tail) {
			visitExpr((Tree) l.head);
			step.append(_tree_);
		}
		visitStat(tree.body);
		Tree body = _tree_;
		_tree_ = make.at(tree.pos).ForLoop(init.toList(), cond, step.toList(), body);
	}
	
	
	public void _case(ForIn tree) {
		// {
		// 		ForUtils coll = new ForUtils(<col>);
		// 		for (
		//			;
		//			coll.hasMoreElement();
		//			
		// 		)
		//		<var> = AssignUtils.assign(<var>, coll.nextElement());
		// 		...
		// }
		Operation op = (Operation)tree.coll;
		Ident var = (Ident)op.args.head;
		Tree coll = (Tree)op.args.tail.head;
		String name = "__idx" + var.name.toString();
		ListBuffer stats = new ListBuffer();
		NewClass newClass = make.at(tree.pos).NewClass(null, toName(tree.pos, "ForUtils"), Tree.emptyList.append(coll), null);
		stats.append( make.at(tree.pos).VarDef(0, Name.fromString(name), toName(tree.pos, "ForUtils"), newClass) );
		
		ListBuffer args = new ListBuffer();
		args.append(var);
		args.append(make.at(tree.pos).Apply(toName(tree.pos, name+".nextElement"), Tree.emptyList));
		Select select = make.at(tree.pos).Select(toName(tree.pos, "AssignUtils"), Name.fromString("assign"));
		Tree step = make.at(tree.pos).Assign(var, make.at(tree.pos).Apply(select, args.toList()));
		step = make.at(tree.pos).Exec(step);
		
		Tree cond = make.at(tree.pos).Apply(toName(tree.pos, name+".hasMoreElement"), Tree.emptyList);
		
		tree.body.visit(this);

		Block block;
		if (tree.body instanceof Block) {
			block = (Block)_tree_;
			block.stats = block.stats.prepend(step);
		} else {
			block = make.at(tree.pos).Block( 0, List.make(step, _tree_) );
		}
	
		stats.append( make.at(tree.pos).ForLoop(Tree.emptyList, cond, Tree.emptyList, block) );
		_tree_ = make.at(tree.pos).Block(0, stats.toList());
	}
		
	public void _case(Labelled tree) {
		visitStat(tree.body);
		Tree body = _tree_;
		_tree_ = make.at(tree.pos).Labelled(tree.label, body);
	}
	
	public void _case(Switch tree) {
		// switch (
		//	SwitchUtils.test(
		//		new SwitchUtils.Switch(<selector>), 
		//		new SwitchUtils.Case[] { 
		//			new SwitchUtils.Case(<case0>), ... 
		//		}
		//	)
		// )
		// case 0:
		// case 1: ...
		
		visitExpr(tree.selector);
		Tree selector = _tree_;
		ListBuffer args = new ListBuffer();
		args.append( make.at(tree.pos).NewClass(null, toName(tree.pos, "SwitchUtils.Switch"), Tree.emptyList.append(selector), null));
		ListBuffer argcases = new ListBuffer();
		ListBuffer cases = new ListBuffer();
		int i = 0;
		for (List l = tree.cases; l.nonEmpty(); l = l.tail) {
			visitExpr((Tree)l.head);
			Case t = (Case)_tree_;
			if (t.pat == null) {
				argcases.append( make.at(tree.pos).NewClass(null, toName(tree.pos, "SwitchUtils.Default"), Tree.emptyList, null));
				t.pat = make.at(t.pos).Literal(Type.INT, new Integer(-1));
			} else {
				argcases.append( make.at(tree.pos).NewClass(null, toName(tree.pos, "SwitchUtils.Case"), Tree.emptyList.append(t.pat), null));
				t.pat = make.at(t.pos).Literal(Type.INT, new Integer(i++));
			}
			cases.append(t);
		}
		args.append( make.at(tree.pos).NewArray(toName(tree.pos, "SwitchUtils.Case"), Tree.emptyList, argcases.toList()) );
		selector = make.at(tree.pos).Apply(toName(tree.pos, "SwitchUtils.test"), args.toList());
		_tree_ = make.at(tree.pos).Switch(selector, cases.toList());
	}

	public void _case(Case tree) {
		visitExpr(tree.pat);
		Tree pat = _tree_;
		ListBuffer stats = new ListBuffer();
		for (List l = tree.stats; l.nonEmpty(); l = l.tail) {
			visitExpr((Tree) l.head);
			if (_tree_ != null)	stats.append(_tree_);
		}
		_tree_ = make.at(tree.pos).Case(pat, stats.toList());
	}
	
	public void _case(Catch tree) {
		// returns Commands.doCatch().tryCatch(<expr>)
		visitExpr(tree.body);
		Tree body = _tree_;
		Apply apply = make.at(tree.pos).Apply(toName(tree.pos, "Commands.doCatch"), Tree.emptyList);
		Select select  = make.at(tree.pos).Select(apply, Name.fromString("tryCatch"));
		_tree_ = make.at(tree.pos).Apply(select, Tree.emptyList.append(body));
	}

	public void _case(Conditional tree) {
		visitExpr(tree.cond);
		Tree cond = _tree_;
		Select select = make.at(tree.pos).Select(toName(tree.pos, "ConvUtils"), Name.fromString("toBool"));
		cond = make.at(tree.pos).Apply(select, Tree.emptyList.append(_tree_));
		visitStat(tree.thenpart);
		Tree thenpart = _tree_;
		visitExpr(tree.elsepart);
		Tree elsepart = _tree_;
		_tree_ = make.at(tree.pos).Conditional(tree.tag, cond, thenpart, elsepart);
	}

	boolean isExpr = false;
	public void _case(Exec tree) {
		isExpr = true;
		visitExpr(tree.expr);
		if (_tree_ == null) return;
		isExpr = false;
		Tree expr = _tree_;
		_tree_ = make.at(tree.pos).Exec(expr);
	}
	
	boolean isCommand = false, insideCommand = false;
	public void _case(Tree.Command tree) {
		isCommand = true;
		if (tree.statement) insideCommand = true;
		visitExpr(tree.cmd);
		if (_tree_ == null) return;
		isCommand = false;
		if (tree.statement && !(_tree_ instanceof Block) &&
			!(tree.cmd instanceof GlobalRef) && 
			(!(tree.cmd instanceof Assign) || !(((Assign)tree.cmd).lhs instanceof GlobalRef))) {
			insideCommand = false;
			_tree_ = make.at(tree.pos).Exec(_tree_);
		}
	}
	
	public void _case(Alias tree) {
		if (tree.cmd != null) currentNamespace.setAlias( tree.name.toString(), tree.cmd );
		else if (tree.args != null) currentNamespace.setAlias( tree.name.toString(), tree.args );
	}

	public void _case(Break tree) {
		_tree_ = make.at(tree.pos).Break(tree.label);
	}
	
	public void _case(Continue tree) {
		_tree_ = make.at(tree.pos).Continue(tree.label);
	}
	
	public void _case(Return tree) {
		visitExpr(tree.expr);
		Tree expr = _tree_;
		_tree_ = make.at(tree.pos).Return(expr);
	}
	

	private String cmdName;
	private ListBuffer cmdArgs;
	
	public boolean processNestedAliases(Apply tree) {
		cmdName = "";
		
		if (tree.meth instanceof Ident)
			cmdName = ((Ident)tree.meth).name.toString();
		else {
			Select t = (Select)tree.meth;
			Tree selected = t.selected;
			cmdName = t.name.toString();
			while (true) {
				if (selected instanceof Ident) {
					cmdName = ((Ident)selected).name.toString() + "." + cmdName;
					break;
				} else if (selected instanceof Select) {
					cmdName = ((Select)selected).name.toString() + "." + cmdName;
					selected = ((Select)selected).selected;
				}
			}
		}

		ListBuffer args = new ListBuffer();
		for (List l = tree.args; l.nonEmpty(); l = l.tail) {
			visitExpr((Tree) l.head);
			ListBuffer args2 = new ListBuffer();
			if (!fullOpt) {
				if (_tree_.tag != Tree.LITERAL) {
					Select select = make.at(tree.pos).Select(toName(tree.pos, "ConvUtils"), Name.fromString("toString"));
					_tree_ = make.at(tree.pos).Apply(select, Tree.emptyList.append(_tree_));
				} else if (((Literal)_tree_).typetag != Type.CLASS) {
					args2.append(make.at(_tree_.pos).Literal(Type.CLASS, ""));
					args2.append(_tree_);
					_tree_ = make.at(tree.pos).Operation(Tree.PLUS, args2.toList());
				}
			}
			args.append(_tree_);
		}

		// check for aliases
		Object listArgs = currentNamespace.getAlias(cmdName);
		if (listArgs != null) {
			if (listArgs instanceof List) {
				if ( ! tree.args.isEmpty() )
					throw new InternalError("No argument expected for alias '" + cmdName + "'.");
				ListBuffer bargs = new ListBuffer();
				for (List l = (List)listArgs; l.nonEmpty(); l = l.tail) {
					visitExpr((Tree) l.head);
					bargs.append(_tree_);
				}
				_tree_ = make.at(tree.pos).Block(0, bargs.toList());
				return false;
			}

			Apply apply = (Apply) listArgs;
			if (!processNestedAliases(apply)) return false;
		} 
		
		cmdArgs.appendList( args.toList() );
		return true;
	}
	
	private int methCnt = 0;
	public void _case(Apply tree) {
		boolean isProc = false, isGlobalProc = false;
		String name = "";
		if (tree.meth instanceof Ident) {
			name = ((Ident)tree.meth).name.toString();
			isProc = procs.get(name) != null;
			isGlobalProc = globalprocs.get(name) != null;
		}
		
		if (isProc) {
			visitExpr(tree.meth);
			Tree meth = _tree_;
			ListBuffer args = new ListBuffer();
			for (List l = tree.args; l.nonEmpty(); l = l.tail) {
				visitExpr((Tree) l.head);
				args.append(_tree_);
			}
			if (isGlobalProc)
				_tree_ = make.at(tree.pos).Apply(toName(tree.pos, name + ".exec"), args.toList());
			else
				_tree_ = make.at(tree.pos).Apply(meth, args.toList());
			return;
		}
		
		// <cmd> -<opt> <arg>;
		//	=> Commands.get(<cmd>).exec(new String[] { "-<opt>","<arg>" });
		visitExpr(tree.meth);
		Tree meth = _tree_;
		
		if (meth instanceof Literal) {
			name = (String)((Literal)meth).value;
			if (name.equals("size") ||
				name.equals("clear") ||
				name.equals("warning") ||
				name.equals("print")) {
				ListBuffer args = new ListBuffer();
				for (List l = tree.args; l.nonEmpty(); l = l.tail) {
					visitExpr((Tree) l.head);
					args.append(_tree_);
				}
				if (name.equals("warning")) {
					// Add the line position
					args.append(make.at(tree.pos).Literal(Type.INT, new Integer(Position.line(tree.pos))));
				}
				meth = make.at(tree.pos).Select(toName(tree.pos, "Commands"), Name.fromString(name));
				_tree_ = make.at(tree.pos).Apply(meth, args.toList());
				return;
			}
		}

		cmdArgs = new ListBuffer();
		if (!processNestedAliases(tree)) return;
		meth = make.at(tree.meth.pos).Literal(Type.CLASS, cmdName);
		
		if (!fullOpt) {
			Select select = make.at(tree.pos).Select(toName(tree.pos, "Commands"), Name.fromString("get"));
			Apply apply = make.at(tree.pos).Apply(select, Tree.emptyList.append(meth));
			meth = make.at(tree.pos).Select(apply, Name.fromString(tree.statement ? "exec" :  "execAsExpr"));
		}		
	
		if (fullOpt) {
			Tree instrs;
			if (tree.statement)
				instrs = TreeCommands.get((Literal)meth).exec(tree.pos, cmdArgs.toList(), methCnt);
			else
				instrs = TreeCommands.get((Literal)meth).execAsExpr(tree.pos, cmdArgs.toList(), methCnt);
			if (!(instrs instanceof MethodDef)) _tree_ = instrs;
			else {
				methCnt++;
				MethodDef methodDef = (MethodDef)instrs;
				methodDef = checkMethod(methodDef);
				
				ListBuffer params = new ListBuffer();
				ListBuffer args2 = new ListBuffer();
				MelVarExtractor extractor = new MelVarExtractor();
				methodDef.visit(extractor);
				for (List l = extractor.vars.toList(); l.nonEmpty(); l = l.tail) {
					VarDef vdef = currentNamespace.getVar((String) l.head);
					if (vdef != null) {
						VarDef vdef2 = make.at(tree.pos).VarDef(0, vdef.name, vdef.vartype, null);
						params.append(vdef2);
						args2.append(make.at(tree.pos).Ident(vdef.name));
					}
				}
				methodDef.params = params.toList();
				
				ListBuffer lb;
				if (insideProc)
					lb = (ListBuffer)procVals.get(insideProcName + Cmds_);
				else
					lb = (ListBuffer)procVals.get(_core + Cmds_);
				lb.append( instrs );

				_tree_ = make.at(tree.pos).Apply(make.at(tree.pos).Ident(methodDef.name), args2.toList());
			}
		} else {
			if (cmdArgs.length() == 0) {
				_tree_ = make.at(tree.pos).Apply(meth, Tree.emptyList);
			} else if (cmdArgs.length() == 1) {
				_tree_ = make.at(tree.pos).Apply(meth, cmdArgs.toList());
			} else {
				NewArray newArray = make.at(tree.pos).NewArray(toName(tree.pos, "String"), Tree.emptyList, cmdArgs.toList());
				_tree_ = make.at(tree.pos).Apply(meth, Tree.emptyList.append(newArray));
			}
		}
	}

	public void _case(NewArray tree) {
		visitExpr(tree.elemtype);
		Tree elemtype = _tree_;
		ListBuffer dims = new ListBuffer();
		if (tree.dims != null)
			for (List l = tree.dims; l.nonEmpty(); l = l.tail) {
				visitExpr((Tree) l.head);
				dims.append(_tree_);
			}
		ListBuffer elems = new ListBuffer();
		Tree firstType = null;
		if (tree.elems != null) {
			for (List l = tree.elems; l.nonEmpty(); l = l.tail) {
				visitExpr((Tree) l.head);
				elems.append(_tree_);
				if (firstType == null) firstType = (Tree) l.head;
			}
		}
		int typetag = -1;
		if (firstType instanceof Literal) typetag = ((Literal)firstType).typetag;
		else if (firstType instanceof TypeCast) typetag = ((TypeIdent)((TypeCast)firstType).clazz).typetag;
		String type = null;
		switch (typetag) {
			case Type.CLASS:
				elemtype = toName(tree.pos, "String");
				type = "String";
				break;
					
			case Type.VECTOR:
				type = "Vector";
				elemtype = toName(tree.pos, "Melvector");
				break;
							
			case Type.INT:
				type = "Int";
				elemtype = make.at(tree.pos).TypeIdent(Type.INT);
				break;
				
			case Type.FLOAT:
				type = "Float";
				elemtype = make.at(tree.pos).TypeIdent(Type.FLOAT);
				break;
								
			case Type.DOUBLE:
				type = "Double";
				elemtype = make.at(tree.pos).TypeIdent(Type.DOUBLE);
				break;		
		}
	
		NewArray newArray = make.at(tree.pos).NewArray(elemtype, dims.toList(), elems.toList());
		if (type == null) _tree_ = newArray;
		else {
			_tree_ = make.at(tree.pos).NewClass(null, toName(tree.pos, "MelArray"+type), Tree.emptyList.append(newArray), null);
		}
	}
	
	public void _case(Assign tree) {
		ListBuffer args = new ListBuffer();
		if (tree.lhs instanceof Indexed) {
			visitExpr(tree.rhs);
			Tree rhs = _tree_;
			Indexed indexed = (Indexed)tree.lhs;
			visitExpr(indexed.indexed);
			Tree indexedindexed = _tree_;
			visitExpr(indexed.index);
			Tree index = _tree_;
			args.append(index);
			if (indexed.indexed.tag == Tree.INDEXED) {
				visitExpr(((Indexed)indexed.indexed).index);
				args.append(_tree_);
			}
			args.append(rhs);
			Select select = make.at(tree.pos).Select(indexedindexed, Name.fromString("setElement"));
			_tree_ = make.at(tree.pos).Apply(select, args.toList());
		} else {
			isAssigned = true;
			visitExpr(tree.lhs);
			isAssigned = false;
			Tree lhs = _tree_;
			
			if (tree.lhs instanceof GlobalRef) {
				GlobalRef gr = (GlobalRef)tree.lhs;
				String name = gr.name.toString();

				TreeMap map;
				if (insideProc)
					map = (TreeMap)procVals.get(insideProcName + GlobalDefs_);
				else
					map = (TreeMap)procVals.get(_core + GlobalDefs_);
				boolean isDefined = map.get(name) != null;
				Tree rhs;
				if (!isDefined) {
					VarDef vdef = make.at(gr.pos).VarDef(Flags.PUBLIC, gr.name, gr.type, tree.rhs);
					vdef.visit(this);
					return;
				} else {
					visitExpr(tree.rhs);
					rhs = _tree_;
				}
				// <var> = Commands.setGlobalVar(<name>, <rhs>)
				_tree_ = make.at(tree.pos).Assign(lhs, make.at(tree.pos).Apply(toName(gr.pos, "Commands.getGlobalVar"), List.make(make.at(tree.pos).Literal(Type.CLASS, name), lhs)));			
				return;
			} else if (tree.lhs instanceof Ident && lhs == null) {
				visitExpr(tree.rhs);
				Tree rhs = _tree_;
				_tree_ = make.at(tree.pos).Apply(toName(tree.pos, "Commands.setGlobalVar"), List.make(make.at(tree.pos).Literal(Type.CLASS, ((Ident)tree.lhs).name.toString()), rhs));			
				return;
			}				
			
			visitExpr(tree.rhs);
			Tree rhs = _tree_;
			args.append(lhs);
			args.append(rhs);
			Select select = make.at(tree.pos).Select(toName(tree.pos, "AssignUtils"), Name.fromString("assign"));
			rhs = make.at(tree.pos).Apply(select, args.toList());
			_tree_ = make.at(tree.pos).Assign(lhs, rhs);
		}
	}
	
	private boolean isAssigned = false;
	public void _case(GlobalRef tree) {
		TreeMap map;
		if (insideProc)
			map = (TreeMap)procVals.get(insideProcName + GlobalDefs_);
		else
			map = (TreeMap)procVals.get(_core + GlobalDefs_);
		boolean isDefined = map.get(tree.name.toString()) != null;
		if (!isAssigned && !isDefined) {
			VarDef vdef = make.at(tree.pos).VarDef(Flags.PUBLIC, tree.name, tree.type, null);
			vdef.visit(this);
		} else
		if (!isAssigned) 
			_tree_ = null;
		else 
			_tree_ = make.at(tree.pos).Ident(tree.name);
	}

	public void _case(Assignop tree) {
		Tree lhs = tree.lhs;
		Tree rhs = tree.rhs;

		switch (tree.tag) {
			case Tree.PLUS_ASG:		rhs = make.at(tree.pos).Operation(Tree.PLUS, List.make(lhs, rhs)); break;
			case Tree.MINUS_ASG:	rhs = make.at(tree.pos).Operation(Tree.MINUS, List.make(lhs, rhs)); break;
			case Tree.MUL_ASG:		rhs = make.at(tree.pos).Operation(Tree.MUL, List.make(lhs, rhs)); break;
			case Tree.DIV_ASG:		rhs = make.at(tree.pos).Operation(Tree.DIV, List.make(lhs, rhs)); break;
			case Tree.MOD_ASG:		rhs = make.at(tree.pos).Operation(Tree.MOD, List.make(lhs, rhs)); break;
			case Tree.BITAND_ASG:	rhs = make.at(tree.pos).Operation(Tree.BITAND, List.make(lhs, rhs)); break;
			case Tree.BITOR_ASG:	rhs = make.at(tree.pos).Operation(Tree.BITOR, List.make(lhs, rhs)); break;
			case Tree.BITXOR_ASG:	rhs = make.at(tree.pos).Operation(Tree.BITXOR, List.make(lhs, rhs)); break;
			case Tree.SL_ASG:		rhs = make.at(tree.pos).Operation(Tree.SL, List.make(lhs, rhs)); break;
			case Tree.SR_ASG:		rhs = make.at(tree.pos).Operation(Tree.SR, List.make(lhs, rhs)); break;
			case Tree.USR_ASG:		rhs = make.at(tree.pos).Operation(Tree.USR, List.make(lhs, rhs)); break;
		}
		
		rhs = make.at(tree.pos).Assign(lhs, rhs);
		rhs.visit(this);
	}

	public void _case(Operation tree) {
		ListBuffer args = new ListBuffer();
		if (tree.tag <= Tree.POSTDEC) {
			if (tree.tag <= Tree.COMPL) {
				visitExpr((Tree) tree.args.head);
				if (tree.tag == Tree.NOT) {
					Select select = make.at(tree.pos).Select(toName(tree.pos, "ConvUtils"), Name.fromString("toBool"));
					_tree_ = make.at(tree.pos).Apply(select, Tree.emptyList.append(_tree_));
				} else if (tree.tag == Tree.NEG) {
					Select select = make.at(tree.pos).Select(toName(tree.pos, "ConvUtils"), Name.fromString("neg"));
					_tree_ = make.at(tree.pos).Apply(select, Tree.emptyList.append(_tree_));
					return;
				}
				args.append(_tree_);
			} else {
				Tree lhs = (Tree)tree.args.head;
				Literal literal = make.at(tree.pos).Literal(Type.INT, new Integer(1));
				int tag = -1;
				switch (tree.tag) {
					case Tree.POSTINC: tag = Tree.PLUS_ASG; break;
					case Tree.POSTDEC: tag = Tree.MINUS_ASG; break;
				}
				if (tag != -1) {
					lhs = make.at(tree.pos).Assignop(tag, lhs, literal);
					lhs.visit(this);
					return;
				} else
					args.append(lhs);
			}
		} else {
			visitExpr((Tree) tree.args.head);
			Tree lhs = _tree_;
			args.append(lhs);
			visitExpr((Tree) tree.args.tail.head);
			Tree rhs = _tree_;
			args.append(rhs);
			String op0 = "";
			String op = null, op2 = null;
			switch (tree.tag) {
				case Tree.PLUS: op0 = "Add"; op = "add"; break;
				case Tree.MINUS: op0 = "Sub"; op = "sub"; break;
				case Tree.MUL: op0 = "Mul"; op = "mul"; break;
				case Tree.DIV: op0 = "Div"; op = "div"; break;
				case Tree.MOD: op0 = "Mod"; op = "mod"; break;
				case Tree.BITXOR: op0 = "Pow"; op = "pow"; break;
				case Tree.EQ: op0 = "Eq"; op = "eq"; break;
				case Tree.NE: op0 = "Eq"; op = "eq"; op2 = "ne"; break;
				
				case Tree.LT:
				case Tree.GT:
				case Tree.LE:
				case Tree.GE: op2 = "toInt"; break;
			}
			if (op != null) {
				Select select = make.at(tree.pos).Select(toName(tree.pos, op0+"Utils"), Name.fromString(op));
				_tree_ = make.at(tree.pos).Apply(select, args.toList());
				if (op2 != null) _tree_ = make.at(tree.pos).Operation(Tree.NOT, Tree.emptyList.append(_tree_));
				return;
			} else if (op2 != null) {
				args = new ListBuffer();
				Select select = make.at(tree.pos).Select(toName(tree.pos, "ConvUtils"), Name.fromString(op2));
				lhs = make.at(tree.pos).Apply(select, Tree.emptyList.append(lhs));
				args.append(lhs);
				select = make.at(tree.pos).Select(toName(tree.pos, "ConvUtils"), Name.fromString(op2));
				rhs = make.at(tree.pos).Apply(select, Tree.emptyList.append(rhs));
				args.append(rhs);
			} else {
				args = new ListBuffer();
				Select select = make.at(tree.pos).Select(toName(tree.pos, "ConvUtils"), Name.fromString("toBool"));
				lhs = make.at(tree.pos).Apply(select, Tree.emptyList.append(lhs));
				args.append(lhs);
				select = make.at(tree.pos).Select(toName(tree.pos, "ConvUtils"), Name.fromString("toBool"));
				rhs = make.at(tree.pos).Apply(select, Tree.emptyList.append(rhs));
				args.append(rhs);
			}
		}
		_tree_ = make.at(tree.pos).Operation(tree.tag, args.toList());
	}

	public void _case(TypeCast tree) {
		visitExpr(tree.clazz);
		Tree clazz = _tree_;
		visitExpr(tree.expr);
		Tree expr = _tree_;
		String type = "";
		switch (((TypeIdent)tree.clazz).typetag) {
			case Type.INT: type = "toInt"; break;
			case Type.FLOAT: type = "toFloat"; break;
			case Type.DOUBLE: type = "toDouble"; break;
			case Type.STRING: type = "toString"; break;
			case Type.VECTOR: type = "toVector"; break;
			case Type.MATRIX: type = "toMatrix"; break;
		}
		Select select = make.at(tree.pos).Select(toName(tree.pos, "ConvUtils"), Name.fromString(type));
		_tree_ = make.at(tree.pos).Apply(select, Tree.emptyList.append(expr));
	}
	
	public void _case(TypeTest tree) {
		visitExpr(tree.clazz);
		Tree clazz = _tree_;
		visitExpr(tree.expr);
		Tree expr = _tree_;
		_tree_ = make.at(tree.pos).TypeCast(clazz, expr);
	}
	
	public void _case(Indexed tree) {
		ListBuffer args = new ListBuffer();
		visitExpr(tree.indexed);
		Tree indexed = _tree_;
		visitExpr(tree.index);
		Tree index = _tree_;
		if ((isCommand || !isExpr) && indexed instanceof Literal) {
			_tree_ = make.at(tree.pos).Literal(Type.CLASS, ((Literal)indexed).value + "[" + ((Literal)index).value + "]");
			return;
		}
		args.append(index);
		if (tree.indexed.tag == Tree.INDEXED) {
			visitExpr(((Indexed)tree.indexed).index);
			args.append(_tree_);
		}
		Select select = make.at(tree.indexed.pos).Select(indexed, Name.fromString("getElement"));
		_tree_ = make.at(tree.pos).Apply(select, args.toList());
	}
	
	public void _case(Select tree) {
		Tree selected = tree.selected;
		boolean isFullName = true;
		String s = tree.name.toString(), name = "";
		while (true) {
			if (selected instanceof Ident) {
				s = (name = ((Ident)selected).name.toString()) + "." + s;
				break;
			} else if (selected instanceof Select) {
				s = ((Select)selected).name.toString() + "." + s;
				selected = ((Select)selected).selected;
			} else {
				isFullName = false;
				break;
			}
		}
		if (isFullName && (isCommand || !isExpr) && (name.charAt(0) == '-' || name.charAt(0) != '$')) {
			_tree_ = make.at(tree.pos).Literal(Type.CLASS, s);
			return;
		}
		visitExpr(tree.selected);
		selected = _tree_;
		_tree_ = make.at(tree.pos).Select(selected, tree.name);
	}
	
	public void _case(Ident tree) {
		String name = tree.name.toString();
		boolean isProc = procs.get(name) != null;
		if (!isProc && (name.charAt(0) == '-' || name.charAt(0) != '$'))
			_tree_ = make.at(tree.pos).Literal(Type.CLASS, name);
		else {
			TreeMap map;
			if (insideProc)
				map = (TreeMap)procVals.get(insideProcName + GlobalDefs_);
			else
				map = (TreeMap)procVals.get(_core + GlobalDefs_);
			boolean isDefined = map.get(tree.name.toString()) != null;
			if (!isDefined)
				_tree_ = make.at(tree.pos).Ident(tree.name);
			else if (isAssigned)
				_tree_ = null;
			else {
				_tree_ = make.at(tree.pos).Apply(toName(tree.pos, "Commands.getGlobalVar"), List.make(make.at(tree.pos).Literal(Type.CLASS, name), tree));			
			}
		}
	}
	
	public void _case(Literal tree) {
		_tree_ = make.at(tree.pos).Literal(tree.typetag, tree.value);
	}

	public void _case(TypeIdent tree) {
		switch (tree.typetag) {
			case Type.STRING:
				_tree_ = toName(tree.pos, "String");
				break;

			case Type.VECTOR:
				_tree_ = toName(tree.pos, "MelVector");
				break;

			case Type.MATRIX:
				_tree_ = toName(tree.pos, "MelMatrix");
				break;
			
			default:
				_tree_ = make.at(tree.pos).TypeIdent(tree.typetag);
				break;
		}
	}
	
	public void _case(TypeArray tree) {
		visitExpr(tree.elemtype);
		Tree elemtype = _tree_;
		String type = null;
		if (elemtype instanceof Ident) {
			type = ((Ident)elemtype).name.toString();
		} else if (elemtype instanceof TypeIdent) {
			int typetag = -1;
			typetag = ((TypeIdent)elemtype).typetag;
			switch (typetag) {
				case Type.CLASS:
					type = "String";
					break;
						
				case Type.VECTOR:
					type = "Vector";
					break;
								
				case Type.INT:
					type = "Int";
					break;
					
				case Type.FLOAT:
					type = "Float";
					break;
									
				case Type.DOUBLE:
					type = "Double";
					break;		
			}
		}	
		if (type == null)
			_tree_ = make.at(tree.pos).TypeArray(elemtype, 0);
		else
			_tree_ = toName(tree.pos, "MelArray"+type);
	}
	
	public void _case(TypeMatrix tree) {
		_tree_ = toName(tree.pos, "MelMatrix");
	}
	
	public void _case(NewVector tree) {
		ListBuffer elems = new ListBuffer();
		ListBuffer args = new ListBuffer();
		for (List l = tree.elems; l.nonEmpty(); l = l.tail) {
			visitExpr((Tree) l.head);
			args.append(_tree_);
		}
		if (tree.elems.length() == 3) {
			_tree_ = make.at(tree.pos).NewClass(null, toName(tree.pos, "MelVector"), args.toList(), null);
		} else {
			TypeArray ta1 = make.at(tree.pos).TypeArray( make.at(tree.pos).TypeIdent(Type.FLOAT), 0 );
			elems.append( make.at(tree.pos).NewArray(make.at(tree.pos).TypeIdent(Type.FLOAT), Tree.emptyList, args.toList()) );
			args = new ListBuffer();
			args.append( make.at(tree.pos).NewArray(ta1, Tree.emptyList, elems.toList()) );
			_tree_ = make.at(tree.pos).NewClass(null, toName(tree.pos, "MelMatrix"), args.toList(), null);
		}
	}
	
	public void _case(NewMatrix tree) {
		ListBuffer args = new ListBuffer();
		ListBuffer dims = new ListBuffer();
		ListBuffer elems = new ListBuffer();
		for (int i = 0; i < tree.rows; i++) {
			ListBuffer elems2 = new ListBuffer();
			for (int j = 0; j < tree.cols; j++) 
				elems2.append(tree.elements[i][j]);
			elems.append( make.at(tree.pos).NewArray(make.at(tree.pos).TypeIdent(Type.FLOAT), Tree.emptyList, elems2.toList()) );
		}
		TypeArray ta1 = make.at(tree.pos).TypeArray( make.at(tree.pos).TypeIdent(Type.FLOAT), 0 );
		TypeArray ta2 = make.at(tree.pos).TypeArray( ta1, 0 );
		args.append( make.at(tree.pos).NewArray(ta1, Tree.emptyList, elems.toList()) );
		_tree_ = make.at(tree.pos).NewClass(null, toName(tree.pos, "MelMatrix"), args.toList(), null);
	}
	
	public void _case(Erroneous tree) {
		_tree_ = make.at(tree.pos).Erroneous();
	}

	/////////////////////////////////////
	private String toJavaLang(String name) {
		if (name.indexOf('.') != -1) return name;
		if 		(name.equals("char") 	|| name.equals("C")) name = "Character";
		else if (name.equals("byte") 	|| name.equals("B")) name = "Byte";
		else if (name.equals("short") 	|| name.equals("S")) name = "Short";
		else if (name.equals("int") 	|| name.equals("I")) name = "Integer";
		else if (name.equals("long") 	|| name.equals("J")) name = "Long";
		else if (name.equals("float") 	|| name.equals("F")) name = "Float";
		else if (name.equals("double") 	|| name.equals("D")) name = "Double";
		else if (name.equals("boolean") || name.equals("Z")) name = "Boolean";
		return name;
	}

	// Split a class name to create an Ident or a Select tree node
	protected Tree toName(int pos, String name) {
		String n = "";
		if (name.startsWith("[")) {
			name = toJavaLang(name.substring(1));
			Tree t = toName(pos, name);
			return make.at(pos).TypeArray(t, 0);
		}
		int p = name.indexOf('.');
		if (p == -1) return make.at(pos).Ident(Name.fromString(name));
		Tree selected = null;
		while (p != -1) {
			n = name.substring(0,p);
			if (selected == null) 
				selected = make.at(pos).Ident(Name.fromString(n));
			else
				selected = make.at(pos).Select(selected, Name.fromString(n));

			name = name.substring(p+1);
			p = name.indexOf('.');
		}
		return make.at(pos).Select(selected, Name.fromString(name));
	}
}
