package jmel.tree;

import jmel.util.*;
import jmel.tree.Tree.*;

/**
 * Extract var list from a command.
 */
public class MelVarExtractor extends GlobalVisitor {
	
	public ListBuffer vars = new ListBuffer();

	/**
	* Visitor methods
	*/
	public void _case(MethodDef tree) {
		visitExpr(tree.body);
	}
	
	public void _case(VarDef tree) {
		visitExpr(tree.init);
	}		

	public void _case(Block tree) {
		ListBuffer stats = new ListBuffer();
		for (List l = tree.stats; l.nonEmpty(); l = l.tail) {
			visitExpr((Tree) l.head);
		}
	}
	
	public void _case(Exec tree) {
		visitExpr(tree.expr);
	}
	
	public void _case(NewArray tree) {
		if (tree.dims != null)
			for (List l = tree.dims; l.nonEmpty(); l = l.tail) {
				visitExpr((Tree) l.head);
			}
		if (tree.elems != null) {
			for (List l = tree.elems; l.nonEmpty(); l = l.tail) {
				visitExpr((Tree) l.head);
			}
		}
	}
	
	public void _case(Return tree) {
		visitExpr(tree.expr);
	}
	
	public void _case(Apply tree) {
		visitExpr(tree.meth);
		for (List l = tree.args; l.nonEmpty(); l = l.tail) {
			visitExpr((Tree) l.head);
		}
	}

	public void _case(NewClass tree) {
		visitExpr(tree.encl);
		visitExprs(tree.args);
	}
	
	public void _case(Assign tree) {
		visitExpr(tree.lhs);
		visitExpr(tree.rhs);
	}
	
	public void _case(Assignop tree) {
		visitExpr(tree.lhs);
		visitExpr(tree.rhs);
	}

	public void _case(Operation tree) {
		ListBuffer args = new ListBuffer();
		if (tree.tag <= Tree.POSTDEC) {
			visitExpr((Tree) tree.args.head);
		} else {
			visitExpr((Tree) tree.args.head);
			visitExpr((Tree) tree.args.tail.head);
		}
	}

	public void _case(TypeCast tree) {
		visitExpr(tree.clazz);
		visitExpr(tree.expr);
	}
	
	public void _case(TypeTest tree) {
		visitExpr(tree.clazz);
		visitExpr(tree.expr);
	}
	
	public void _case(Indexed tree) {
		visitExpr(tree.indexed);
		visitExpr(tree.index);
	}
	
	public void _case(Select tree) {
		visitExpr(tree.selected);
	}
	
	public void _case(Ident tree) {
		vars.append(tree.name.toString());
	}
	
	public void _case(Literal tree) {
	}

	public void _case(TypeIdent tree) {
	}
	
	public void _case(TypeArray tree) {
	}
	
	public void _case(TypeMatrix tree) {
	}

	public void _case(NewVector tree) {
		for (List l = tree.elems; l.nonEmpty(); l = l.tail) {
			visitExpr((Tree) l.head);
		}
	}
	
	public void _case(NewMatrix tree) {
		for (int i = 0; i < tree.rows; i++) {
			for (int j = 0; j < tree.cols; j++) 
				visitExpr(tree.elements[i][j]);
		}
	}
	
	public void _case(Tree tree) {
		throw new InternalError("Unknwon tree '"+tree+"' into command invocation.");
	}
}
