package jmel.tree;

public interface Nodes {
	/**
	* Toplevel nodes, of type TopLevel, representing entire source files.
	*/
	public static final int TOPLEVEL = 1;
	
	/**
	* Import clauses, of type Import.
	*/
	public static final int IMPORT = TOPLEVEL + 1;
	
	/**
	* Class definitions, of type ClassDef.
	*/
	public static final int CLASSDEF = IMPORT + 1;
	
	/**
	* Method definitions, of type MethodDef.
	*/
	public static final int METHODDEF = CLASSDEF + 1;
	
	/**
	* Variable definitions, of type VarDef.
	*/
	public static final int VARDEF = METHODDEF + 1;
	
	/**
	* Blocks, of type Block.
	*/
	public static final int BLOCK = VARDEF + 1;
	
	/**
	* Do-while loops, of type DoLoop.
	*/
	public static final int DOLOOP = BLOCK + 1;
	
	/**
	* While-loops, of type WhileLoop.
	*/
	public static final int WHILELOOP = DOLOOP + 1;
	
	/**
	* For-loops, of type ForLoop.
	*/
	public static final int FORLOOP = WHILELOOP + 1;

	/**
	* Labelled statements, of type Labelled.
	*/
	public static final int LABELLED = FORLOOP + 1;
	
	/**
	* Switch statements, of type Switch.
	*/
	public static final int SWITCH = LABELLED + 1;
	
	/**
	* Case parts in switch statements, of type Case.
	*/
	public static final int CASE = SWITCH + 1;
	
	/**
	* Synchronized statements, of type Synchonized.
	*/
	public static final int SYNCHRONIZED = CASE + 1;
	
	/**
	* Try statements, of type Try.
	*/
	public static final int TRY = SYNCHRONIZED + 1;
	
	/**
	* Catch clauses in try statements, of type Catch.
	*/
	public static final int CATCH = TRY + 1;
	
	/**
	* Conditional expressions, of type Conditional.
	*/
	public static final int CONDEXPR = CATCH + 1;
	
	/**
	* Conditional statements, of type Conditional.
	*/
	public static final int CONDSTAT = CONDEXPR + 1;
	
	/**
	* Expression statements, of type Exec.
	*/
	public static final int EXEC = CONDSTAT + 1;

	/**
	* Break statements, of type Break.
	*/
	public static final int BREAK = EXEC + 1;
	
	/**
	* Continue statements, of type Continue.		// 20
	*/
	public static final int CONTINUE = BREAK + 1;
	
	/**
	* Return statements, of type Return.
	*/
	public static final int RETURN = CONTINUE + 1;
	
	/**
	* Throw statements, of type Throw.
	*/
	public static final int THROW = RETURN + 1;
	
	/**
	* Method invocation expressions, of type Apply.
	*/
	public static final int APPLY = THROW + 1;
	
	/**
	* Class instance creation expressions, of type NewClass.
	*/
	public static final int NEWCLASS = APPLY + 1;
	
	/**
	* Array creation expressions, of type NewArray.
	*/
	public static final int NEWARRAY = NEWCLASS + 1;
	
	/**
	* Assignment expressions, of type Assign.
	*/
	public static final int ASSIGN = NEWARRAY + 1;
	
	/**
	* Type cast expressions, of type TypeCast.
	*/
	public static final int TYPECAST = ASSIGN + 1;
	
	/**
	* Type test expressions, of type TypeTest.
	*/
	public static final int TYPETEST = TYPECAST + 1;

	/**
	* Indexed array expressions, of type Indexed.
	*/
	public static final int INDEXED = TYPETEST + 1;
	
	/**
	* Selections, of type Select.
	*/
	public static final int SELECT = INDEXED + 1;
	
	/**
	* Simple identifiers, of type Ident.
	*/
	public static final int IDENT = SELECT + 1;
	
	/**
	* Literals, of type Literal.
	*/
	public static final int LITERAL = IDENT + 1;
	
	/**
	* Basic type identifiers, of type TypeIdent.
	*/
	public static final int TYPEIDENT = LITERAL + 1;
	
	/**
	* Array types, of type TypeArray.
	*/
	public static final int TYPEARRAY = TYPEIDENT + 1;
	
	/**
	* Parameterized types, of type TypeApply.
	*/
	public static final int TYPEAPPLY = TYPEARRAY + 1;
	
	/**
	* Formal type parameters, of type TypeParameter.
	*/
	public static final int TYPEPARAMETER = TYPEAPPLY + 1;
	
	/**
	* Error trees, of type Erroneous.						// 37
	*/
	public static final int ERRONEOUS = TYPEPARAMETER + 1;		
	
	public static final int VECTOR = ERRONEOUS + 1;
	public static final int MATRIX = VECTOR + 1;
	public static final int TYPEMATRIX = MATRIX + 1;
	public static final int COMMAND = TYPEMATRIX + 1;
	public static final int GLOBALREF = COMMAND + 1;
	public static final int COLLECTION = GLOBALREF + 1;
	public static final int FORIN = COLLECTION + 1;
	public static final int ALIAS = FORIN + 1;

	/**
	 * Unary operators, of type Operation.
	 */
	public static final int POS = ALIAS + 1;				// 45
	public static final int NEG = POS + 1;
	public static final int NOT = NEG + 1;
	public static final int COMPL = NOT + 1;
	public static final int PREINC = COMPL + 1;
	public static final int PREDEC = PREINC + 1;
	public static final int POSTINC = PREDEC + 1;
	public static final int POSTDEC = POSTINC + 1;
	
	/**
	 * Binary operators, of type Operation.
	 */
	public static final int OR = POSTDEC + 1;
	public static final int AND = OR + 1;
	public static final int BITOR = AND + 1;
	public static final int BITXOR = BITOR + 1;
	public static final int BITAND = BITXOR + 1;
	public static final int EQ = BITAND + 1;
	public static final int NE = EQ + 1;
	public static final int LT = NE + 1;
	public static final int GT = LT + 1;
	public static final int LE = GT + 1;
	public static final int GE = LE + 1;
	public static final int SL = GE + 1;
	public static final int SR = SL + 1;
	public static final int USR = SR + 1;
	public static final int PLUS = USR + 1;
	public static final int MINUS = PLUS + 1;
	public static final int MUL = MINUS + 1;
	public static final int DIV = MUL + 1;
	public static final int MOD = DIV + 1;
	
	/**
	 * Assignment operators, of type Assignop.
	 */
	public static final int BITOR_ASG = MOD + 1;
	public static final int BITXOR_ASG = BITOR_ASG + 1;
	public static final int BITAND_ASG = BITXOR_ASG + 1;
	public static final int SL_ASG = SL + BITOR_ASG - BITOR;
	public static final int SR_ASG = SL_ASG + 1;
	public static final int USR_ASG = SR_ASG + 1;
	public static final int PLUS_ASG = USR_ASG + 1;
	public static final int MINUS_ASG = PLUS_ASG + 1;
	public static final int MUL_ASG = MINUS_ASG + 1;
	public static final int DIV_ASG = MUL_ASG + 1;
	public static final int MOD_ASG = DIV_ASG + 1;
	
	/**
	* The offset between assignment operators and normal operators.
	*/
	public static final int ASGOffset = BITOR_ASG - BITOR;
}
